 /*
 * sim-outorder.c - sample out-of-order issue perf simulator implementation
 *
 * This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin as a part of the Multiscalar Research Project.
 *  
 * The tool suite is currently maintained by Doug Burger and Todd M. Austin.
 * 
 * Copyright (C) 1994, 1995, 1996, 1997 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 *
 * $Id: sim-outorder.c,v 1.16 1999/02/03 18:11:45 huangj Exp $
 *
 * $Log: sim-outorder.c,v $
 * Revision 1.16  1999/02/03  18:11:45  huangj
 * relaxed writeback requirement.
 * added cmu_out_buffer.
 * reorder the events in run().
 *
 * Revision 1.15  1998/09/20  22:21:33  huangj
 * Now loads are issued only when the data is ready in the memory
 * buffer, if it requires the data from parent thread
 * However, since ruu_dispatch is kind of independent of ruu_issue and
 * lsq_refresh, we need to keep previous try-retry protocols.
 *
 * Revision 1.14  1998/07/13 20:04:37  huangj
 * *** empty log message ***
 *
 * Revision 1.13  1998/03/02  20:01:16  huangj
 * Bug-fixed release.
 *
 * Revision 1.12  1998/02/16 07:33:28  huangj
 * Support for forwarding mb entry via CMU
 *
 * Revision 1.11  1997/12/18 21:07:56  huangj
 * single process version
 *
 * Revision 1.10  1997/12/05 08:04:50  huangj
 * *** empty log message ***
 *
 * Revision 1.9  1997/12/01 20:48:12  huangj
 * Multiple thread_abort problem fixed.
 *
 * Revision 1.8  1997/12/01 04:42:36  huangj
 * Using the same number of processes from beginning to end.
 *
 * Revision 1.7  1997/11/26 22:12:34  huangj
 * Needs to search for free CMU everytime a thread is started.
 *
 * Revision 1.6  1997/11/18  20:39:29  huangj
 * Sync without semaphore.
 *
 * Revision 1.5  1997/11/13  23:32:01  huangj
 * File length reduced. MB write fixed to write in
 * right-aligned manner.
 * Hash table added for MB access. See st_utils.c
 *
 * Revision 1.4  1997/11/09  23:56:06  huangj
 * MB_WRITE fixed.
 *
 * Revision 1.3  1997/11/05  17:27:09  huangj
 * Presumably bugs are fixed. However, SGI sema ops are not
 * reliable. And the sync could fail.
 *
 * Revision 1.2  1997/11/01 22:54:38  huangj
 * Added half-word and byte support for ALTS and RLTS.
 * Bug fixed for process_st_msg_que.
 *
 * Revision 1.1  1997/10/12 03:49:05  huangj
 * Initial revision
 *
 * Revision 1.5  1997/09/20 16:56:04  huangj
 * *** empty log message ***
 *
 * Revision 1.4  1997/09/20 15:55:50  huangj
 * Works with printf!
 *
 * Revision 1.3  1997/09/18 23:06:39  huangj
 * First Working version!
 *
 * Revision 1.2  1997/09/18 05:16:00  huangj
 * STORE_TS fixed.
 *
 * Revision 1.4  1997/04/16  22:10:23  taustin
 * added -commit:width support (from kskadron)
 * fixed "bad l2 D-cache parms" fatal string
 *
 * Revision 1.3  1997/03/11  17:17:06  taustin
 * updated copyright
 * `-pcstat' option support added
 * long/int tweaks made for ALPHA target support
 * better defaults defined for caches/TLBs
 * "mstate" command supported added for DLite!
 * supported added for non-GNU C compilers
 * buglet fixed in speculative trace generation
 * multi-level cache hierarchy now supported
 * two-level predictor supported added
 * I/D-TLB supported added
 * many comments added
 * options package supported added
 * stats package support added
 * resource configuration options extended
 * pipetrace support added
 * DLite! support added
 * writeback throttling now supported
 * decode and issue B/W now decoupled
 * new and improved (and more precise) memory scheduler added
 * cruft for TLB paper removed
 *
 * Revision 1.1  1996/12/05  18:52:32  taustin
 * Initial revision
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <signal.h>

#define  EXT

#include "misc.h"
#include "ss.h"
#include "regs.h"
#include "memory.h"
#include "cache.h"
#include "loader.h"
#include "syscall.h"
#include "bpred.h"
#include "resource.h"
#include "bitmap.h"
#include "options.h"
#include "stats.h"
#include "ptrace.h"
#include "dlite.h"
#include "sim.h"

#ifdef SUPER_TR
SS_WORD_TYPE saved_regs[SS_NUM_REGS], saved_rhi, saved_rlo;
union regs_FP saved_fpr;
int saved_fcc;
#endif

/*
 * This file implements a very detailed out-of-order issue superscalar
 * processor with a two-level memory system and speculative execution support.
 * This simulator is a performance simulator, tracking the latency of all
 * pipeline operations.
 */


/*
 * simulator options
 */

/* pipeline trace range and output filename */
static int ptrace_nelt = 0;
static char *ptrace_opts[2];

/* instruction fetch queue size (in insts) */
static int ruu_ifq_size;

/* extra branch mis-prediction latency */
static int ruu_branch_penalty;

/* speed of front-end of machine relative to execution core */
static int fetch_speed;

/* branch predictor type {nottaken|taken|perfect|bimod|2lev} */
static char *pred_type;

/* bimodal predictor config (<table_size>) */
static int bimod_nelt = 1;
static int bimod_config[1] =
  { /* bimod tbl size */2048 };

/* 2-level predictor config (<l1size> <l2size> <hist_size> <xor>) */
static int twolev_nelt = 4;
static int twolev_config[4] =
  { /* l1size */1, /* l2size */1024, /* hist */8, /* xor */FALSE};

/* combining predictor config (<meta_table_size> */
static int comb_nelt = 1;
static int comb_config[1] =
  { /* meta_table_size */1024 };

/* return address stack (RAS) size */
static int ras_size = 8;

/* BTB predictor config (<num_sets> <associativity>) */
static int btb_nelt = 2;
static int btb_config[2] =
  { /* nsets */512, /* assoc */4 };

/* instruction decode B/W (insts/cycle) */
static int ruu_decode_width;

/* instruction issue B/W (insts/cycle) */
static int ruu_issue_width;

/* run pipeline with in-order issue */
static int ruu_inorder_issue;

/* issue instructions down wrong execution paths */
static int ruu_include_spec = TRUE;

/* instruction commit B/W (insts/cycle) */
static int ruu_commit_width;

/* register update unit (RUU) size */
static int RUU_size = 8;

/* load/store queue (LSQ) size */
static int LSQ_size = 4;

/* l1 data cache config, i.e., {<config>|none} */
static char *cache_dl1_opt;

/* l1 data cache hit latency (in cycles) */
static int cache_dl1_lat;

/* l2 data cache config, i.e., {<config>|none} */
static char *cache_dl2_opt;

/* l2 data cache hit latency (in cycles) */
static int cache_dl2_lat;

/* l1 instruction cache config, i.e., {<config>|dl1|dl2|none} */
static char *cache_il1_opt;

/* l1 instruction cache hit latency (in cycles) */
static int cache_il1_lat;

/* l2 instruction cache config, i.e., {<config>|dl1|dl2|none} */
static char *cache_il2_opt;

/* l2 instruction cache hit latency (in cycles) */
static int cache_il2_lat;

/* flush caches on system calls */
static int flush_on_syscalls;

/* convert 64-bit inst addresses to 32-bit inst equivalents */
static int compress_icache_addrs;

/* memory access latency (<first_chunk> <inter_chunk>) */
static int mem_nelt = 2;
static int mem_lat[2] =
  { /* lat to first chunk */18, /* lat between remaining chunks */2 };

/* memory access bus width (in bytes) */
static int mem_bus_width;

/* instruction TLB config, i.e., {<config>|none} */
static char *itlb_opt;

/* data TLB config, i.e., {<config>|none} */
static char *dtlb_opt;

/* inst/data TLB miss latency (in cycles) */
static int tlb_miss_lat;

/* total number of integer ALU's available */
static int res_ialu;

/* total number of integer multiplier/dividers available */
static int res_imult;

/* total number of memory system ports available (to CPU) */
static int res_memport;

/* total number of floating point ALU's available */
static int res_fpalu;

/* total number of floating point multiplier/dividers available */
static int res_fpmult;

/* text-based stat profiles */
#define MAX_PCSTAT_VARS 8
static int pcstat_nelt = 0;
static char *pcstat_vars[MAX_PCSTAT_VARS];

/* operate in backward-compatible bugs mode (for testing only) */
static int bugcompat_mode;

/*
 * functional unit resource configuration
 */

/* resource pool indices, NOTE: update these if you change FU_CONFIG */
#define FU_IALU_INDEX			0
#define FU_IMULT_INDEX			1
#define FU_MEMPORT_INDEX		2
#define FU_FPALU_INDEX			3
#define FU_FPMULT_INDEX			4

/* resource pool definition, NOTE: update FU_*_INDEX defs if you change this */
struct res_desc fu_config[] = {
  {
    "integer-ALU",
    4,
    0,
    {
      { IntALU, 1, 1 }
    }
  },
  {
    "integer-MULT/DIV",
    1,
    0,
    {
      { IntMULT, 3, 1 },
      { IntDIV, 20, 19 }
    }
  },
  {
    "memory-port",
    2,
    0,
    {
      { RdPort, 1, 1 },
      { WrPort, 1, 1 }
    }
  },
  {
    "FP-adder",
    4,
    0,
    {
      { FloatADD, 2, 1 },
      { FloatCMP, 2, 1 },
      { FloatCVT, 2, 1 }
    }
  },
  {
    "FP-MULT/DIV",
    1,
    0,
    {
      { FloatMULT, 4, 1 },
      { FloatDIV, 12, 12 },
      { FloatSQRT, 24, 24 }
    }
  },
};


/*
 * simulator stats
 */

/* total number of instructions committed */
SS_COUNTER_TYPE sim_num_insn = 0;

/* total number of instructions executed */
static SS_COUNTER_TYPE sim_total_insn = 0;

/* total number of memory references committed */
static SS_COUNTER_TYPE sim_num_refs = 0;

/* total number of memory references executed */
static SS_COUNTER_TYPE sim_total_refs = 0;

/* total number of loads committed */
static SS_COUNTER_TYPE sim_num_loads = 0;

/* total number of loads executed */
static SS_COUNTER_TYPE sim_total_loads = 0;

/* total number of branches committed */
static SS_COUNTER_TYPE sim_num_branches = 0;

/* total number of branches executed */
static SS_COUNTER_TYPE sim_total_branches = 0;

/* cycle counter */
SS_TIME_TYPE sim_cycle = 0;


/*
 * simulator state variables
 */

/* instruction sequence counter, used to assign unique id's to insts */
static unsigned int inst_seq = 0;

/* pipetrace instruction sequence counter */
static unsigned int ptrace_seq = 0;

/* speculation mode, non-zero when mis-speculating, i.e., executing
   instructions down the wrong path, thus state recovery will eventually have
   to occur that resets processor register and memory state back to the last
   precise state */
int spec_mode[ST_NUM_TU];

/* cycles until fetch issue resumes */
static unsigned ruu_fetch_issue_delay[ST_NUM_TU];

/* perfect prediction enabled */
static int pred_perfect;

/* speculative bpred-update enabled */
static char *bpred_spec_opt;
static enum { spec_ID, spec_WB, spec_CT } bpred_spec_update[ST_NUM_TU];

/* level 1 instruction cache, entry level instruction cache */
static struct cache *cache_il1;

/* level 1 instruction cache */
static struct cache *cache_il2;

/* level 1 data cache, entry level data cache */
static struct cache *cache_dl1;

/* level 2 data cache */
static struct cache *cache_dl2;

/* instruction TLB */
static struct cache *itlb;

/* data TLB */
static struct cache *dtlb;

/* branch predictor */
static struct bpred *pred;

/* functional unit resource pool */
static struct res_pool *fu_pool[ST_NUM_TU];

/* text-based stat profiles */
static struct stat_stat_t *pcstat_stats[MAX_PCSTAT_VARS];
static SS_COUNTER_TYPE pcstat_lastvals[MAX_PCSTAT_VARS];
static struct stat_stat_t *pcstat_sdists[MAX_PCSTAT_VARS];

/* wedge all stat values into a SS_COUNTER_TYPE */
#define STATVAL(STAT)							\
  ((STAT)->sc == sc_int							\
   ? (SS_COUNTER_TYPE)*((STAT)->variant.for_int.var)			\
   : ((STAT)->sc == sc_uint						\
      ? (SS_COUNTER_TYPE)*((STAT)->variant.for_uint.var)		\
      : ((STAT)->sc == sc_counter					\
	 ? *((STAT)->variant.for_counter.var)				\
	 : (panic("bad stat class"), 0))))


/* memory access latency, assumed to not cross a page boundary */
static unsigned int			/* total latency of access */
mem_access_latency(int blk_sz)		/* block size accessed */
{
  int chunks = (blk_sz + (mem_bus_width - 1)) / mem_bus_width;

  assert(chunks > 0);

  return (/* first chunk latency */mem_lat[0] +
	  (/* remainder chunk latency */mem_lat[1] * (chunks - 1)));
}


/*
 * cache miss handlers
 */

/* l1 data cache l1 block miss handler function */
static unsigned int			/* latency of block access */
dl1_access_fn(enum mem_cmd cmd,		/* access cmd, Read or Write */
	      SS_ADDR_TYPE baddr,	/* block address to access */
	      int bsize,		/* size of block to access */
	      struct cache_blk *blk,	/* ptr to block in upper level */
	      SS_TIME_TYPE now)		/* time of access */
{
  unsigned int lat;

  if (cache_dl2)
    {
      /* access next level of data cache hierarchy */
      lat = cache_access(cache_dl2, cmd, baddr, NULL, bsize,
			 /* now */now, /* pudata */NULL, /* repl addr */NULL);
      if (cmd == Read)
	return lat;
      else
	{
	  /* FIXME: unlimited write buffers */
	  return 0;
	}
    }
  else
    {
      /* access main memory */
      if (cmd == Read)
	return mem_access_latency(bsize);
      else
	{
	  /* FIXME: unlimited write buffers */
	  return 0;
	}
    }
}

/* l2 data cache block miss handler function */
static unsigned int			/* latency of block access */
dl2_access_fn(enum mem_cmd cmd,		/* access cmd, Read or Write */
	      SS_ADDR_TYPE baddr,	/* block address to access */
	      int bsize,		/* size of block to access */
	      struct cache_blk *blk,	/* ptr to block in upper level */
	      SS_TIME_TYPE now)		/* time of access */
{
  /* this is a miss to the lowest level, so access main memory */
  if (cmd == Read)
    return mem_access_latency(bsize);
  else
    {
      /* FIXME: unlimited write buffers */
      return 0;
    }
}

/* l1 inst cache l1 block miss handler function */
static unsigned int			/* latency of block access */
il1_access_fn(enum mem_cmd cmd,		/* access cmd, Read or Write */
	      SS_ADDR_TYPE baddr,	/* block address to access */
	      int bsize,		/* size of block to access */
	      struct cache_blk *blk,	/* ptr to block in upper level */
	      SS_TIME_TYPE now)		/* time of access */
{
  unsigned int lat;

if (cache_il2)
    {
      /* access next level of inst cache hierarchy */
      lat = cache_access(cache_il2, cmd, baddr, NULL, bsize,
			 /* now */now, /* pudata */NULL, /* repl addr */NULL);
      if (cmd == Read)
	return lat;
      else
	panic("writes to instruction memory not supported");
    }
  else
    {
      /* access main memory */
      if (cmd == Read)
	return mem_access_latency(bsize);
      else
	panic("writes to instruction memory not supported");
    }
}

/* l2 inst cache block miss handler function */
static unsigned int			/* latency of block access */
il2_access_fn(enum mem_cmd cmd,		/* access cmd, Read or Write */
	      SS_ADDR_TYPE baddr,	/* block address to access */
	      int bsize,		/* size of block to access */
	      struct cache_blk *blk,	/* ptr to block in upper level */
	      SS_TIME_TYPE now)		/* time of access */
{
  /* this is a miss to the lowest level, so access main memory */
  if (cmd == Read)
    return mem_access_latency(bsize);
  else
    panic("writes to instruction memory not supported");
}


/*
 * TLB miss handlers
 */

/* inst cache block miss handler function */
static unsigned int			/* latency of block access */
itlb_access_fn(enum mem_cmd cmd,	/* access cmd, Read or Write */
	       SS_ADDR_TYPE baddr,	/* block address to access */
	       int bsize,		/* size of block to access */
	       struct cache_blk *blk,	/* ptr to block in upper level */
	       SS_TIME_TYPE now)	/* time of access */
{
  SS_ADDR_TYPE *phy_page_ptr = (SS_ADDR_TYPE *)blk->user_data;

  /* no real memory access, however, should have user data space attached */
  assert(phy_page_ptr);

  /* fake translation, for now... */
  *phy_page_ptr = 0;

  /* return tlb miss latency */
  return tlb_miss_lat;
}

/* data cache block miss handler function */
static unsigned int			/* latency of block access */
dtlb_access_fn(enum mem_cmd cmd,	/* access cmd, Read or Write */
	       SS_ADDR_TYPE baddr,	/* block address to access */
	       int bsize,		/* size of block to access */
	       struct cache_blk *blk,	/* ptr to block in upper level */
	       SS_TIME_TYPE now)	/* time of access */
{
  SS_ADDR_TYPE *phy_page_ptr = (SS_ADDR_TYPE *)blk->user_data;

  /* no real memory access, however, should have user data space attached */
  assert(phy_page_ptr);

  /* fake translation, for now... */
  *phy_page_ptr = 0;

  /* return tlb miss latency */
  return tlb_miss_lat;
}


/* register simulator-specific options */
void
sim_reg_options(struct opt_odb_t *odb)
{
  opt_reg_header(odb, 
"sim-outorder: This simulator implements a very detailed out-of-order issue\n"
"superscalar processor with a two-level memory system and speculative\n"
"execution support.  This simulator is a performance simulator, tracking the\n"
"latency of all pipeline operations.\n"
		 );

  /* trace options */

  opt_reg_string_list(odb, "-ptrace",
	      "generate pipetrace, i.e., <fname|stdout|stderr> <range>",
	      ptrace_opts, /* arr_sz */2, &ptrace_nelt, /* default */NULL,
	      /* !print */FALSE, /* format */NULL, /* !accrue */FALSE);

  opt_reg_note(odb,
"  Pipetrace range arguments are formatted as follows:\n"
"\n"
"    {{@|#}<start>}:{{@|#|+}<end>}\n"
"\n"
"  Both ends of the range are optional, if neither are specified, the entire\n"
"  execution is traced.  Ranges that start with a `@' designate an address\n"
"  range to be traced, those that start with an `#' designate a cycle count\n"
"  range.  All other range values represent an instruction count range.  The\n"
"  second argument, if specified with a `+', indicates a value relative\n"
"  to the first argument, e.g., 1000:+100 == 1000:1100.  Program symbols may\n"
"  be used in all contexts.\n"
"\n"
"    Examples:   -ptrace FOO.trc #0:#1000\n"
"                -ptrace BAR.trc @2000:\n"
"                -ptrace BLAH.trc :1500\n"
"                -ptrace UXXE.trc :\n"
"                -ptrace FOOBAR.trc @main:+278\n"
	       );

  /* ifetch options */

  opt_reg_int(odb, "-fetch:ifqsize", "instruction fetch queue size (in insts)",
	      &ruu_ifq_size, /* default */4,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-fetch:mplat", "extra branch mis-prediction latency",
	      &ruu_branch_penalty, /* default */3,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-fetch:speed",
	      "speed of front-end of machine relative to execution core",
	      &fetch_speed, /* default */1,
	      /* print */TRUE, /* format */NULL);

  /* branch predictor options */

  opt_reg_note(odb,
"  Branch predictor configuration examples for 2-level predictor:\n"
"    Configurations:   N, M, W, X\n"
"      N   # entries in first level (# of shift register(s))\n"
"      W   width of shift register(s)\n"
"      M   # entries in 2nd level (# of counters, or other FSM)\n"
"      X   (yes-1/no-0) xor history and address for 2nd level index\n"
"    Sample predictors:\n"
"      GAg     : 1, W, 2^W, 0\n"
"      GAp     : 1, W, M (M > 2^W), 0\n"
"      PAg     : N, W, 2^W, 0\n"
"      PAp     : N, W, M (M == 2^(N+W)), 0\n"
"      gshare  : 1, W, 2^W, 1\n"
"  Predictor `comb' combines a bimodal and a 2-level predictor.\n"
               );

  opt_reg_string(odb, "-bpred",
		 "branch predictor type {nottaken|taken|perfect|bimod|2lev|comb}",
                 &pred_type, /* default */"bimod",
                 /* print */TRUE, /* format */NULL);

  opt_reg_int_list(odb, "-bpred:bimod",
		   "bimodal predictor config (<table size>)",
		   bimod_config, bimod_nelt, &bimod_nelt,
		   /* default */bimod_config,
		   /* print */TRUE, /* format */NULL, /* !accrue */FALSE);

  opt_reg_int_list(odb, "-bpred:2lev",
                   "2-level predictor config "
		   "(<l1size> <l2size> <hist_size> <xor>)",
                   twolev_config, twolev_nelt, &twolev_nelt,
		   /* default */twolev_config,
                   /* print */TRUE, /* format */NULL, /* !accrue */FALSE);

  opt_reg_int_list(odb, "-bpred:comb",
		   "combining predictor config (<meta_table_size>)",
		   comb_config, comb_nelt, &comb_nelt,
		   /* default */comb_config,
		   /* print */TRUE, /* format */NULL, /* !accrue */FALSE);

  opt_reg_int(odb, "-bpred:ras",
              "return address stack size (0 for no return stack)",
              &ras_size, /* default */ras_size,
              /* print */TRUE, /* format */NULL);

  opt_reg_int_list(odb, "-bpred:btb",
		   "BTB config (<num_sets> <associativity>)",
		   btb_config, btb_nelt, &btb_nelt,
		   /* default */btb_config,
		   /* print */TRUE, /* format */NULL, /* !accrue */FALSE);

  opt_reg_string(odb, "-bpred:spec_update",
		 "speculative predictors update in {ID|WB} (default non-spec)",
		 &bpred_spec_opt, /* default */NULL,
		 /* print */TRUE, /* format */NULL);

  /* decode options */

  opt_reg_int(odb, "-decode:width",
	      "instruction decode B/W (insts/cycle)",
	      &ruu_decode_width, /* default */4,
	      /* print */TRUE, /* format */NULL);

  /* issue options */

  opt_reg_int(odb, "-issue:width",
	      "instruction issue B/W (insts/cycle)",
	      &ruu_issue_width, /* default */4,
	      /* print */TRUE, /* format */NULL);

  opt_reg_flag(odb, "-issue:inorder", "run pipeline with in-order issue",
	       &ruu_inorder_issue, /* default */FALSE,
	       /* print */TRUE, /* format */NULL);

  opt_reg_flag(odb, "-issue:wrongpath",
	       "issue instructions down wrong execution paths",
	       &ruu_include_spec, /* default */TRUE,
	       /* print */TRUE, /* format */NULL);

  /* commit options */

  opt_reg_int(odb, "-commit:width",
	      "instruction commit B/W (insts/cycle)",
	      &ruu_commit_width, /* default */4,
	      /* print */TRUE, /* format */NULL);

  /* register scheduler options */

  opt_reg_int(odb, "-ruu:size",
	      "register update unit (RUU) size",
	      &RUU_size, /* default */16,
	      /* print */TRUE, /* format */NULL);

  /* memory scheduler options  */

  opt_reg_int(odb, "-lsq:size",
	      "load/store queue (LSQ) size",
	      &LSQ_size, /* default */8,
	      /* print */TRUE, /* format */NULL);

  /* cache options */

  opt_reg_string(odb, "-cache:dl1",
		 "l1 data cache config, i.e., {<config>|none}",
		 &cache_dl1_opt, "dl1:128:32:4:l",
		 /* print */TRUE, NULL);

  opt_reg_note(odb,
"  The cache config parameter <config> has the following format:\n"
"\n"
"    <name>:<nsets>:<bsize>:<assoc>:<repl>\n"
"\n"
"    <name>   - name of the cache being defined\n"
"    <nsets>  - number of sets in the cache\n"
"    <bsize>  - block size of the cache\n"
"    <assoc>  - associativity of the cache\n"
"    <repl>   - block replacement strategy, 'l'-LRU, 'f'-FIFO, 'r'-random\n"
"\n"
"    Examples:   -cache:dl1 dl1:4096:32:1:l\n"
"                -dtlb dtlb:128:4096:32:r\n"
	       );

  opt_reg_int(odb, "-cache:dl1lat",
	      "l1 data cache hit latency (in cycles)",
	      &cache_dl1_lat, /* default */1,
	      /* print */TRUE, /* format */NULL);

  opt_reg_string(odb, "-cache:dl2",
		 "l2 data cache config, i.e., {<config>|none}",
		 &cache_dl2_opt, "ul2:1024:64:4:l",
		 /* print */TRUE, NULL);

  opt_reg_int(odb, "-cache:dl2lat",
	      "l2 data cache hit latency (in cycles)",
	      &cache_dl2_lat, /* default */6,
	      /* print */TRUE, /* format */NULL);

  opt_reg_string(odb, "-cache:il1",
		 "l1 inst cache config, i.e., {<config>|dl1|dl2|none}",
		 &cache_il1_opt, "il1:512:32:1:l",
		 /* print */TRUE, NULL);

  opt_reg_note(odb,
"  Cache levels can be unified by pointing a level of the instruction cache\n"
"  hierarchy at the data cache hiearchy using the \"dl1\" and \"dl2\" cache\n"
"  configuration arguments.  Most sensible combinations are supported, e.g.,\n"
"\n"
"    A unified l2 cache (il2 is pointed at dl2):\n"
"      -cache:il1 il1:128:64:1:l -cache:il2 dl2\n"
"      -cache:dl1 dl1:256:32:1:l -cache:dl2 ul2:1024:64:2:l\n"
"\n"
"    Or, a fully unified cache hierarchy (il1 pointed at dl1):\n"
"      -cache:il1 dl1\n"
"      -cache:dl1 ul1:256:32:1:l -cache:dl2 ul2:1024:64:2:l\n"
	       );

  opt_reg_int(odb, "-cache:il1lat",
	      "l1 instruction cache hit latency (in cycles)",
	      &cache_il1_lat, /* default */1,
	      /* print */TRUE, /* format */NULL);

  opt_reg_string(odb, "-cache:il2",
		 "l2 instruction cache config, i.e., {<config>|dl2|none}",
		 &cache_il2_opt, "dl2",
		 /* print */TRUE, NULL);

  opt_reg_int(odb, "-cache:il2lat",
	      "l2 instruction cache hit latency (in cycles)",
	      &cache_il2_lat, /* default */6,
	      /* print */TRUE, /* format */NULL);

  opt_reg_flag(odb, "-cache:flush", "flush caches on system calls",
	       &flush_on_syscalls, /* default */FALSE, /* print */TRUE, NULL);

  opt_reg_flag(odb, "-cache:icompress",
	       "convert 64-bit inst addresses to 32-bit inst equivalents",
	       &compress_icache_addrs, /* default */FALSE,
	       /* print */TRUE, NULL);

  /* mem options */
  opt_reg_int_list(odb, "-mem:lat",
		   "memory access latency (<first_chunk> <inter_chunk>)",
		   mem_lat, mem_nelt, &mem_nelt, mem_lat,
		   /* print */TRUE, /* format */NULL, /* !accrue */FALSE);

  opt_reg_int(odb, "-mem:width", "memory access bus width (in bytes)",
	      &mem_bus_width, /* default */8,
	      /* print */TRUE, /* format */NULL);

  /* TLB options */

  opt_reg_string(odb, "-tlb:itlb",
		 "instruction TLB config, i.e., {<config>|none}",
		 &itlb_opt, "itlb:16:4096:4:l", /* print */TRUE, NULL);

  opt_reg_string(odb, "-tlb:dtlb",
		 "data TLB config, i.e., {<config>|none}",
		 &dtlb_opt, "dtlb:32:4096:4:l", /* print */TRUE, NULL);

  opt_reg_int(odb, "-tlb:lat",
	      "inst/data TLB miss latency (in cycles)",
	      &tlb_miss_lat, /* default */30,
	      /* print */TRUE, /* format */NULL);

  /* resource configuration */

  opt_reg_int(odb, "-res:ialu",
	      "total number of integer ALU's available",
	      &res_ialu, /* default */fu_config[FU_IALU_INDEX].quantity,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-res:imult",
	      "total number of integer multiplier/dividers available",
	      &res_imult, /* default */fu_config[FU_IMULT_INDEX].quantity,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-res:memport",
	      "total number of memory system ports available (to CPU)",
	      &res_memport, /* default */fu_config[FU_MEMPORT_INDEX].quantity,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-res:fpalu",
	      "total number of floating point ALU's available",
	      &res_fpalu, /* default */fu_config[FU_FPALU_INDEX].quantity,
	      /* print */TRUE, /* format */NULL);

  opt_reg_int(odb, "-res:fpmult",
	      "total number of floating point multiplier/dividers available",
	      &res_fpmult, /* default */fu_config[FU_FPMULT_INDEX].quantity,
	      /* print */TRUE, /* format */NULL);

  opt_reg_string_list(odb, "-pcstat",
		      "profile stat(s) against text addr's (mult uses ok)",
		      pcstat_vars, MAX_PCSTAT_VARS, &pcstat_nelt, NULL,
		      /* !print */FALSE, /* format */NULL, /* accrue */TRUE);

  opt_reg_flag(odb, "-bugcompat",
	       "operate in backward-compatible bugs mode (for testing only)",
	       &bugcompat_mode, /* default */FALSE, /* print */TRUE, NULL);
}

/* check simulator-specific option values */
void
sim_check_options(struct opt_odb_t *odb,        /* options database */
		  int argc, char **argv)        /* command line arguments */
{
  char name[128], c;
  int nsets, bsize, assoc;

  if (ruu_ifq_size < 1 || (ruu_ifq_size & (ruu_ifq_size - 1)) != 0)
    fatal("inst fetch queue size must be positive > 0 and a power of two");

  if (ruu_branch_penalty < 1)
    fatal("mis-prediction penalty must be at least 1 cycle");

  if (fetch_speed < 1)
    fatal("front-end speed must be positive and non-zero");

  if (!mystricmp(pred_type, "perfect"))
    {
      /* perfect predictor */
      pred = NULL;
      pred_perfect = TRUE;
    }
  else if (!mystricmp(pred_type, "taken"))
    {
      /* static predictor, not taken */
      pred = bpred_create(BPredTaken, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    }
  else if (!mystricmp(pred_type, "nottaken"))
    {
      /* static predictor, taken */
      pred = bpred_create(BPredNotTaken, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    }
  else if (!mystricmp(pred_type, "bimod"))
    {
      /* bimodal predictor, bpred_create() checks BTB_SIZE */
      if (bimod_nelt != 1)
	fatal("bad bimod predictor config (<table_size>)");
      if (btb_nelt != 2)
	fatal("bad btb config (<num_sets> <associativity>)");

      /* bimodal predictor, bpred_create() checks BTB_SIZE */
      pred = bpred_create(BPred2bit,
			  /* bimod table size */bimod_config[0],
			  /* 2lev l1 size */0,
			  /* 2lev l2 size */0,
			  /* meta table size */0,
			  /* history reg size */0,
			  /* history xor address */0,
			  /* btb sets */btb_config[0],
			  /* btb assoc */btb_config[1],
			  /* ret-addr stack size */ras_size);
    }
  else if (!mystricmp(pred_type, "2lev"))
    {
      /* 2-level adaptive predictor, bpred_create() checks args */
      if (twolev_nelt != 4)
	fatal("bad 2-level pred config (<l1size> <l2size> <hist_size> <xor>)");
      if (btb_nelt != 2)
	fatal("bad btb config (<num_sets> <associativity>)");

      pred = bpred_create(BPred2Level,
			  /* bimod table size */0,
			  /* 2lev l1 size */twolev_config[0],
			  /* 2lev l2 size */twolev_config[1],
			  /* meta table size */0,
			  /* history reg size */twolev_config[2],
			  /* history xor address */twolev_config[3],
			  /* btb sets */btb_config[0],
			  /* btb assoc */btb_config[1],
			  /* ret-addr stack size */ras_size);
    }
  else if (!mystricmp(pred_type, "comb"))
    {
      /* combining predictor, bpred_create() checks args */
      if (twolev_nelt != 4)
	fatal("bad 2-level pred config (<l1size> <l2size> <hist_size> <xor>)");
      if (bimod_nelt != 1)
	fatal("bad bimod predictor config (<table_size>)");
      if (comb_nelt != 1)
	fatal("bad combining predictor config (<meta_table_size>)");
      if (btb_nelt != 2)
	fatal("bad btb config (<num_sets> <associativity>)");

      pred = bpred_create(BPredComb,
			  /* bimod table size */bimod_config[0],
			  /* l1 size */twolev_config[0],
			  /* l2 size */twolev_config[1],
			  /* meta table size */comb_config[0],
			  /* history reg size */twolev_config[2],
			  /* history xor address */twolev_config[3],
			  /* btb sets */btb_config[0],
			  /* btb assoc */btb_config[1],
			  /* ret-addr stack size */ras_size);
    }
  else
    fatal("cannot parse predictor type `%s'", pred_type);

  if (!bpred_spec_opt)
    bpred_spec_update[TRI] = spec_CT;
  else if (!mystricmp(bpred_spec_opt, "ID"))
    bpred_spec_update[TRI] = spec_ID;
  else if (!mystricmp(bpred_spec_opt, "WB"))
    bpred_spec_update[TRI] = spec_WB;
  else
    fatal("bad speculative update stage specifier, use {ID|WB}");

  if (ruu_decode_width < 1 || (ruu_decode_width & (ruu_decode_width-1)) != 0)
    fatal("issue width must be positive non-zero and a power of two");

  if (ruu_issue_width < 1 || (ruu_issue_width & (ruu_issue_width-1)) != 0)
    fatal("issue width must be positive non-zero and a power of two");

  if (ruu_commit_width < 1)
    fatal("commit width must be positive non-zero");

  if (RUU_size < 2 || (RUU_size & (RUU_size-1)) != 0)
    fatal("RUU size must be a positive number > 1 and a power of two");

  if (LSQ_size < 2 || (LSQ_size & (LSQ_size-1)) != 0)
    fatal("LSQ size must be a positive number > 1 and a power of two");

  /* use a level 1 D-cache? */
  if (!mystricmp(cache_dl1_opt, "none"))
    {
      cache_dl1 = NULL;

      /* the level 2 D-cache cannot be defined */
      if (strcmp(cache_dl2_opt, "none"))
	fatal("the l1 data cache must defined if the l2 cache is defined");
      cache_dl2 = NULL;
    }
  else /* dl1 is defined */
    {
      if (sscanf(cache_dl1_opt, "%[^:]:%d:%d:%d:%c",
		 name, &nsets, &bsize, &assoc, &c) != 5)
	fatal("bad l1 D-cache parms: <name>:<nsets>:<bsize>:<assoc>:<repl>");
      cache_dl1 = cache_create(name, nsets, bsize, /* balloc */FALSE,
			       /* usize */0, assoc, cache_char2policy(c),
			       dl1_access_fn, /* hit lat */cache_dl1_lat);

      /* is the level 2 D-cache defined? */
      if (!mystricmp(cache_dl2_opt, "none"))
	cache_dl2 = NULL;
      else
	{
	  if (sscanf(cache_dl2_opt, "%[^:]:%d:%d:%d:%c",
		     name, &nsets, &bsize, &assoc, &c) != 5)
	    fatal("bad l2 D-cache parms: "
		  "<name>:<nsets>:<bsize>:<assoc>:<repl>");
	  cache_dl2 = cache_create(name, nsets, bsize, /* balloc */FALSE,
				   /* usize */0, assoc, cache_char2policy(c),
				   dl2_access_fn, /* hit lat */cache_dl2_lat);
	}
    }

  /* use a level 1 I-cache? */
  if (!mystricmp(cache_il1_opt, "none"))
    {
      cache_il1 = NULL;

      /* the level 2 I-cache cannot be defined */
      if (strcmp(cache_il2_opt, "none"))
	fatal("the l1 inst cache must defined if the l2 cache is defined");
      cache_il2 = NULL;
    }
  else if (!mystricmp(cache_il1_opt, "dl1"))
    {
      if (!cache_dl1)
	fatal("I-cache l1 cannot access D-cache l1 as it's undefined");
      cache_il1 = cache_dl1;

      /* the level 2 I-cache cannot be defined */
      if (strcmp(cache_il2_opt, "none"))
	fatal("the l1 inst cache must defined if the l2 cache is defined");
      cache_il2 = NULL;
    }
  else if (!mystricmp(cache_il1_opt, "dl2"))
    {
      if (!cache_dl2)
	fatal("I-cache l1 cannot access D-cache l2 as it's undefined");
      cache_il1 = cache_dl2;

      /* the level 2 I-cache cannot be defined */
      if (strcmp(cache_il2_opt, "none"))
	fatal("the l1 inst cache must defined if the l2 cache is defined");
      cache_il2 = NULL;
    }
  else /* il1 is defined */
    {
      if (sscanf(cache_il1_opt, "%[^:]:%d:%d:%d:%c",
		 name, &nsets, &bsize, &assoc, &c) != 5)
	fatal("bad l1 I-cache parms: <name>:<nsets>:<bsize>:<assoc>:<repl>");
      cache_il1 = cache_create(name, nsets, bsize, /* balloc */FALSE,
			       /* usize */0, assoc, cache_char2policy(c),
			       il1_access_fn, /* hit lat */cache_il1_lat);

      /* is the level 2 D-cache defined? */
      if (!mystricmp(cache_il2_opt, "none"))
	cache_il2 = NULL;
      else if (!mystricmp(cache_il2_opt, "dl2"))
	{
	  if (!cache_dl2)
	    fatal("I-cache l2 cannot access D-cache l2 as it's undefined");
	  cache_il2 = cache_dl2;
	}
      else
	{
	  if (sscanf(cache_il2_opt, "%[^:]:%d:%d:%d:%c",
		     name, &nsets, &bsize, &assoc, &c) != 5)
	    fatal("bad l2 I-cache parms: "
		  "<name>:<nsets>:<bsize>:<assoc>:<repl>");
	  cache_il2 = cache_create(name, nsets, bsize, /* balloc */FALSE,
				   /* usize */0, assoc, cache_char2policy(c),
				   il2_access_fn, /* hit lat */cache_il2_lat);
	}
    }

  /* use an I-TLB? */
  if (!mystricmp(itlb_opt, "none"))
    itlb = NULL;
  else
    {
      if (sscanf(itlb_opt, "%[^:]:%d:%d:%d:%c",
		 name, &nsets, &bsize, &assoc, &c) != 5)
	fatal("bad TLB parms: <name>:<nsets>:<page_size>:<assoc>:<repl>");
      itlb = cache_create(name, nsets, bsize, /* balloc */FALSE,
			  /* usize */sizeof(SS_ADDR_TYPE), assoc,
			  cache_char2policy(c), itlb_access_fn,
			  /* hit latency */1);
    }

  /* use a D-TLB? */
  if (!mystricmp(dtlb_opt, "none"))
    dtlb = NULL;
  else
    {
      if (sscanf(dtlb_opt, "%[^:]:%d:%d:%d:%c",
		 name, &nsets, &bsize, &assoc, &c) != 5)
	fatal("bad TLB parms: <name>:<nsets>:<page_size>:<assoc>:<repl>");
      dtlb = cache_create(name, nsets, bsize, /* balloc */FALSE,
			  /* usize */sizeof(SS_ADDR_TYPE), assoc,
			  cache_char2policy(c), dtlb_access_fn,
			  /* hit latency */1);
    }

  if (cache_dl1_lat < 1)
    fatal("l1 data cache latency must be greater than zero");

  if (cache_dl2_lat < 1)
    fatal("l2 data cache latency must be greater than zero");

  if (cache_il1_lat < 1)
    fatal("l1 instruction cache latency must be greater than zero");

  if (cache_il2_lat < 1)
    fatal("l2 instruction cache latency must be greater than zero");

  if (mem_nelt != 2)
    fatal("bad memory access latency (<first_chunk> <inter_chunk>)");

  if (mem_lat[0] < 1 || mem_lat[1] < 1)
    fatal("all memory access latencies must be greater than zero");

  if (mem_bus_width < 1 || (mem_bus_width & (mem_bus_width-1)) != 0)
    fatal("memory bus width must be positive non-zero and a power of two");

  if (tlb_miss_lat < 1)
    fatal("TLB miss latency must be greater than zero");

  if (res_ialu < 1)
    fatal("number of integer ALU's must be greater than zero");
  if (res_ialu > MAX_INSTS_PER_CLASS)
    fatal("number of integer ALU's must be <= MAX_INSTS_PER_CLASS");
  fu_config[FU_IALU_INDEX].quantity = res_ialu;
  
  if (res_imult < 1)
    fatal("number of integer multiplier/dividers must be greater than zero");
  if (res_imult > MAX_INSTS_PER_CLASS)
    fatal("number of integer mult/div's must be <= MAX_INSTS_PER_CLASS");
  fu_config[FU_IMULT_INDEX].quantity = res_imult;
  
  if (res_memport < 1)
    fatal("number of memory system ports must be greater than zero");
  if (res_memport > MAX_INSTS_PER_CLASS)
    fatal("number of memory system ports must be <= MAX_INSTS_PER_CLASS");
  fu_config[FU_MEMPORT_INDEX].quantity = res_memport;
  
  if (res_fpalu < 1)
    fatal("number of floating point ALU's must be greater than zero");
  if (res_fpalu > MAX_INSTS_PER_CLASS)
    fatal("number of floating point ALU's must be <= MAX_INSTS_PER_CLASS");
  fu_config[FU_FPALU_INDEX].quantity = res_fpalu;
  
  if (res_fpmult < 1)
    fatal("number of floating point multiplier/dividers must be > zero");
  if (res_fpmult > MAX_INSTS_PER_CLASS)
    fatal("number of FP mult/div's must be <= MAX_INSTS_PER_CLASS");
  fu_config[FU_FPMULT_INDEX].quantity = res_fpmult;
}

/* print simulator-specific configuration information */
void
sim_aux_config(FILE *stream)            /* output stream */
{
  /* nada */
}

/* register simulator-specific statistics */
void
sim_reg_stats(struct stat_sdb_t *sdb)   /* stats database */
{
  int i;

  /* register baseline stats */
  stat_reg_counter(sdb, "sim_num_insn",
		   "total number of instructions committed",
		   &sim_num_insn, 0, NULL);
  stat_reg_counter(sdb, "sim_num_refs",
		   "total number of loads and stores committed",
		   &sim_num_refs, 0, NULL);
  stat_reg_counter(sdb, "sim_num_loads",
		   "total number of loads committed",
		   &sim_num_loads, 0, NULL);
  stat_reg_formula(sdb, "sim_num_stores",
		   "total number of stores committed",
		   "sim_num_refs - sim_num_loads", NULL);
  stat_reg_counter(sdb, "sim_num_branches",
		   "total number of branches committed",
		   &sim_num_branches, /* initial value */0, /* format */NULL);
  stat_reg_int(sdb, "sim_elapsed_time",
	       "total simulation time in seconds",
	       (int *)&sim_elapsed_time, 0, NULL);
  stat_reg_formula(sdb, "sim_inst_rate",
		   "simulation speed (in insts/sec)",
		   "sim_num_insn / sim_elapsed_time", NULL);

  stat_reg_counter(sdb, "sim_total_insn",
		   "total number of instructions executed",
		   &sim_total_insn, 0, NULL);
  stat_reg_counter(sdb, "sim_total_refs",
		   "total number of loads and stores executed",
		   &sim_total_refs, 0, NULL);
  stat_reg_counter(sdb, "sim_total_loads",
		   "total number of loads executed",
		   &sim_total_loads, 0, NULL);
  stat_reg_formula(sdb, "sim_total_stores",
		   "total number of stores executed",
		   "sim_total_refs - sim_total_loads", NULL);
  stat_reg_counter(sdb, "sim_total_branches",
		   "total number of branches executed",
		   &sim_total_branches, /* initial value */0, /* format */NULL);

  /* register performance stats */
  stat_reg_counter(sdb, "sim_cycle",
		   "total simulation time in cycles",
		   &sim_cycle, /* initial value */0, /* format */NULL);
  stat_reg_formula(sdb, "sim_IPC",
		   "instructions per cycle",
		   "sim_num_insn / sim_cycle", /* format */NULL);
  stat_reg_formula(sdb, "sim_CPI",
		   "cycles per instruction",
		   "sim_cycle / sim_num_insn", /* format */NULL);
  stat_reg_formula(sdb, "sim_exec_BW",
		   "total instructions (mis-spec + committed) per cycle",
		   "sim_total_insn / sim_cycle", /* format */NULL);
  stat_reg_formula(sdb, "sim_IPB",
		   "instruction per branch",
		   "sim_num_insn / sim_num_branches", /* format */NULL);

  /* register predictor stats */
  if (pred)
    bpred_reg_stats(pred, sdb);

  /* register cache stats */
  if (cache_il1
      && (cache_il1 != cache_dl1 && cache_il1 != cache_dl2))
    cache_reg_stats(cache_il1, sdb);
  if (cache_il2
      && (cache_il2 != cache_dl1 && cache_il2 != cache_dl2))
    cache_reg_stats(cache_il2, sdb);
  if (cache_dl1)
    cache_reg_stats(cache_dl1, sdb);
  if (cache_dl2)
    cache_reg_stats(cache_dl2, sdb);
  if (itlb)
    cache_reg_stats(itlb, sdb);
  if (dtlb)
    cache_reg_stats(dtlb, sdb);

  for (i=0; i<pcstat_nelt; i++)
    {
      char buf[512], buf1[512];
      struct stat_stat_t *stat;

      /* track the named statistical variable by text address */

      /* find it... */
      stat = stat_find_stat(sdb, pcstat_vars[i]);
      if (!stat)
	fatal("cannot locate any statistic named `%s'", pcstat_vars[i]);

      /* stat must be an integral type */
      if (stat->sc != sc_int && stat->sc != sc_uint && stat->sc != sc_counter)
	fatal("`-pcstat' statistical variable `%s' is not an integral type",
	      stat->name);

      /* register this stat */
      pcstat_stats[i] = stat;
      pcstat_lastvals[i] = STATVAL(stat);

      /* declare the sparce text distribution */
      sprintf(buf, "%s_by_pc", stat->name);
      sprintf(buf1, "%s (by text address)", stat->desc);
      pcstat_sdists[i] = stat_reg_sdist(sdb, buf, buf1,
					/* initial value */0,
					/* print format */(PF_COUNT|PF_PDF),
					/* format */"0x%lx %lu %.2f",
					/* print fn */NULL);
    }
}

/* forward declarations */
static void ruu_init(void);
static void lsq_init(void);
static void rslink_init(int nlinks);
static void eventq_init(void);
static void readyq_init(void);
static void cv_init(void);
static void tracer_init(void);
static void fetch_init(void);

/* default register state accessor, used by DLite */
static char *					/* err str, NULL for no err */
simoo_reg_obj(enum dlite_access_t at,		/* access type */
	      enum dlite_reg_t rt,		/* reg bank to probe */
	      int reg,				/* register number */
	      union dlite_reg_val_t *val);	/* input, output */

/* default memory state accessor, used by DLite */
static char *					/* err str, NULL for no err */
simoo_mem_obj(enum dlite_access_t at,		/* access type */
	      SS_ADDR_TYPE addr,		/* address to access */
	      char *p,				/* input/output buffer */
	      int nbytes);			/* size of access */

/* default machine state accessor, used by DLite */
static char *					/* err str, NULL for no err */
simoo_mstate_obj(FILE *stream,			/* output stream */
		 char *cmd);			/* optional command string */

/* total RS links allocated at program start */
#define MAX_RS_LINKS                    4096

/* initialize the simulator */
void
sim_init(void)
{
  sim_num_insn = 0;
  sim_num_refs = 0;

  /* initialize here, so symbols can be loaded */
  if (ptrace_nelt == 2)
    {
      /* generate a pipeline trace */
      ptrace_open(/* fname */ptrace_opts[0], /* range */ptrace_opts[1]);
    }
  else if (ptrace_nelt == 0)
    {
      /* no pipetracing */;
    }
  else
    fatal("bad pipetrace args, use: <fname|stdout|stderr> <range>");

  /* decode all instructions */
  {
    SS_ADDR_TYPE addr;
    SS_INST_TYPE inst;

    if (OP_MAX > 255)
      fatal("cannot do fast decoding, too many opcodes");

    debug("sim: decoding text segment...");
    for (addr=ld_text_base;
	 addr < (ld_text_base+ld_text_size);
	 addr += SS_INST_SIZE)
      {
	inst = __UNCHK_MEM_ACCESS(SS_INST_TYPE, addr);
	inst.a = (inst.a & ~0xff) | (unsigned int)SS_OP_ENUM(SS_OPCODE(inst));
	__UNCHK_MEM_ACCESS(SS_INST_TYPE, addr) = inst;
      }
  }

  /* initialize the simulation engine */
  fu_pool[TRI] = res_create_pool("fu-pool", fu_config, N_ELT(fu_config));
  rslink_init(MAX_RS_LINKS);
  tracer_init();
  fetch_init();
  cv_init();
  eventq_init();
  readyq_init();
  ruu_init();
  lsq_init();

  /* initialize the DLite debugger */
  dlite_init(simoo_reg_obj, simoo_mem_obj, simoo_mstate_obj);
}

/* dump simulator-specific auxiliary simulator statistics */
void
sim_aux_stats(FILE *stream)             /* output stream */
{
  /* nada */
}

/* un-initialize the simulator */
void
sim_uninit(void)
{
  if (ptrace_nelt > 0)
    ptrace_close();
}


/*
 * processor core definitions and declarations
 */

/* inst tag type, used to tag an operation instance in the RUU */
typedef unsigned int INST_TAG_TYPE;

/* inst sequence type, used to order instructions in the ready list, if
   this rolls over the ready list order temporarily will get messed up,
   but execution will continue and complete correctly */
typedef unsigned int INST_SEQ_TYPE;


/* total input dependencies possible */
#define MAX_IDEPS               3

/* total output dependencies possible */
#define MAX_ODEPS               2

/* a register update unit (RUU) station, this record is contained in the
   processors RUU, which serves as a collection of ordered reservations
   stations.  The reservation stations capture register results and await
   the time when all operands are ready, at which time the instruction is
   issued to the functional units; the RUU is an order circular queue, in which
   instructions are inserted in fetch (program) order, results are stored in
   the RUU buffers, and later when an RUU entry is the oldest entry in the
   machines, it and its instruction's value is retired to the architectural
   register file in program order, NOTE: the RUU and LSQ share the same
   structure, this is useful because loads and stores are split into two
   operations: an effective address add and a load/store, the add is inserted
   into the RUU and the load/store inserted into the LSQ, allowing the add
   to wake up the load/store when effective address computation has finished */
struct RUU_station {
  /* inst info */
  SS_INST_TYPE IR;                      /* instruction bits */
  enum ss_opcode op;                    /* decoded instruction opcode */
  SS_ADDR_TYPE PC, next_PC, pred_PC;    /* inst PC, next PC, predicted PC */
  int in_LSQ;                           /* non-zero if op is in LSQ */
  int ea_comp;                          /* non-zero if op is an addr comp */
  int recover_inst;                     /* start of mis-speculation? */
  int stack_recover_idx;		/* non-speculative TOS for RSB pred */
  struct bpred_update dir_update;	/* bpred direction update info */
  int spec_mode;                        /* non-zero if issued in spec_mode */
  SS_ADDR_TYPE addr;                    /* effective address for ld/st's */
  INST_TAG_TYPE tag;                    /* RUU slot tag, increment to
					   squash operation */
  INST_SEQ_TYPE seq;                    /* instruction sequence, used to
					   sort the ready list and tag inst */
  unsigned int ptrace_seq;		/* pipetrace sequence number */

  /* instruction status */
  int queued;                           /* operands ready and queued */
  int issued;                           /* operation is/was executing */
  int completed;                        /* operation has completed execution */

  /* output operand dependency list, these lists are used to
     limit the number of associative searches into the RUU when
     instructions complete and need to wake up dependent insts */
  int onames[MAX_ODEPS];                /* output logical names (NA=unused) */
  struct RS_link *odep_list[MAX_ODEPS]; /* chains to consuming operations */

  /* input dependent links, the output chains rooted above use these
     fields to mark input operands as ready, when all these fields have
     been set non-zero, the RUU operation has all of its register
     operands, it may commence execution as soon as all of its memory
     operands are known to be read (see lsq_refresh() for details on
     enforcing memory dependencies) */
  int idep_ready[MAX_IDEPS];            /* input operand ready? */
};

/* non-zero if all register operands are ready, update with MAX_IDEPS */
#define OPERANDS_READY(RS)                                              \
  ((RS)->idep_ready[0] && (RS)->idep_ready[1] && (RS)->idep_ready[2])

/* register update unit, combination of reservation stations and reorder
   buffer device, organized as a circular queue */
static struct RUU_station *RUU[ST_NUM_TU];         /* register update unit */
/* RUU head and tail pointers */
static int RUU_head[ST_NUM_TU], RUU_tail[ST_NUM_TU]; 
static int RUU_num[ST_NUM_TU];           /* num entries currently in RUU */

/* allocate and initialize register update unit (RUU) */
static void
ruu_init(void)
{
  int i; 
  char *pi;

  if (RUU[TRI] == NULL) {
      RUU[TRI] = calloc(RUU_size, sizeof(struct RUU_station));

      if (!RUU[TRI])
          fatal("out of virtual memory");
  }

  RUU_num[TRI] = 0;
  RUU_head[TRI] = RUU_tail[TRI] = 0;
  
  pi = (char *) RUU[TRI];
  for (i =0 ; i<RUU_size * sizeof(struct RUU_station); i++)
	pi[i] = 0;
  
}

/* dump the contents of the RUU */
static void
ruu_dumpent(struct RUU_station *rs,		/* ptr to RUU station */
	    int index,				/* entry index */
	    FILE *stream,			/* output stream */
	    int header)				/* print header? */
{
  if (header)
    fprintf(stream, "idx: %2d: opcode: %s, inst: `",
	    index, SS_OP_NAME(rs->op));
  else
    fprintf(stream, "       opcode: %s, inst: `",
	    SS_OP_NAME(rs->op));
  ss_print_insn(rs->IR, rs->PC, stream);
  fprintf(stream, "'\n");
  fprintf(stream, "         PC: 0x%08x, NPC: 0x%08x (pred_PC: 0x%08x)\n",
	  rs->PC, rs->next_PC, rs->pred_PC);
  fprintf(stream, "         in_LSQ: %s, ea_comp: %s, recover_inst: %s\n",
	  rs->in_LSQ ? "t" : "f",
	  rs->ea_comp ? "t" : "f",
	  rs->recover_inst ? "t" : "f");
  fprintf(stream, "         spec_mode: %s, addr: 0x%08x, tag: 0x%08x\n",
	  rs->spec_mode ? "t" : "f", rs->addr, rs->tag);
  fprintf(stream, "         seq: 0x%08x, ptrace_seq: 0x%08x\n",
	  rs->seq, rs->ptrace_seq);
  fprintf(stream, "         queued: %s, issued: %s, completed: %s\n",
	  rs->queued ? "t" : "f",
	  rs->issued ? "t" : "f",
	  rs->completed ? "t" : "f");
  fprintf(stream, "         operands ready: %s\n",
	  OPERANDS_READY(rs) ? "t" : "f");
}

/* dump the contents of the RUU */
static void
ruu_dump(FILE *stream)				/* output stream */
{
  int num, head;
  struct RUU_station *rs;

  fprintf(stream, "** RUU state **\n");
  fprintf(stream, "RUU_head: %d, RUU_tail: %d\n", RUU_head, RUU_tail);
  fprintf(stream, "RUU_num: %d\n", RUU_num);

  num = RUU_num[TRI];
  head = RUU_head[TRI];
  while (num)
    {
      rs = &RUU[TRI][head];
      ruu_dumpent(rs, rs - RUU[TRI], stream, /* header */TRUE);
      head = (head + 1) % RUU_size;
      num--;
    }
}

/*
 * load/store queue (LSQ): holds loads and stores in program order, indicating
 * status of load/store access:
 *
 *   - issued: address computation complete, memory access in progress
 *   - completed: memory access has completed, stored value available
 *   - squashed: memory access was squashed, ignore this entry
 *
 * loads may execute when:
 *   1) register operands are ready, and
 *   2) memory operands are ready (no earlier unresolved store)
 *
 * loads are serviced by:
 *   1) previous store at same address in LSQ (hit latency), or
 *   2) data cache (hit latency + miss latency)
 *
 * stores may execute when:
 *   1) register operands are ready
 *
 * stores are serviced by:
 *   1) depositing store value into the load/store queue
 *   2) writing store value to the store buffer (plus tag check) at commit
 *   3) writing store buffer entry to data cache when cache is free
 *
 * NOTE: the load/store queue can bypass a store value to a load in the same
 *   cycle the store executes (using a bypass network), thus stores complete
 *   in effective zero time after their effective address is known
 */
static struct RUU_station *LSQ[ST_NUM_TU];         /* load/store queue */
  /* LSQ head and tail pointers */
static int LSQ_head[ST_NUM_TU], LSQ_tail[ST_NUM_TU];        
static int LSQ_num[ST_NUM_TU];         /* num entries currently in LSQ */

/*
 * input dependencies for stores in the LSQ:
 *   idep #0 - operand input (value that is store'd)
 *   idep #1 - effective address input (address of store operation)
 */
#define STORE_OP_INDEX                  0
#define STORE_ADDR_INDEX                1

#define STORE_OP_READY(RS)              ((RS)->idep_ready[STORE_OP_INDEX])
#define STORE_ADDR_READY(RS)            ((RS)->idep_ready[STORE_ADDR_INDEX])

/* allocate and initialize the load/store queue (LSQ) */
static void
lsq_init(void)
{ 
  int i;
  char *pi;

  if (LSQ[TRI] == NULL) {
      LSQ[TRI] = calloc(LSQ_size, sizeof(struct RUU_station));

      if (!LSQ)
          fatal("out of virtual memory");
  }

  LSQ_num[TRI] = 0;
  LSQ_head[TRI] = LSQ_tail[TRI] = 0;

  pi = (char *) LSQ[TRI];
  for (i =0 ; i<LSQ_size * sizeof(struct RUU_station); i++)
	pi[i] = 0;

}

/* dump the contents of the RUU */
static void
lsq_dump(FILE *stream)				/* output stream */
{
  int num, head;
  struct RUU_station *rs;

  fprintf(stream, "** LSQ state **\n");
  fprintf(stream, "LSQ_head: %d, LSQ_tail: %d\n", LSQ_head[TRI], LSQ_tail[TRI]);
  fprintf(stream, "LSQ_num: %d\n", LSQ_num[TRI]);

  num = LSQ_num[TRI];
  head = LSQ_head[TRI];
  while (num)
    {
      rs = &LSQ[TRI][head];
      ruu_dumpent(rs, rs - LSQ[TRI], stream, /* header */TRUE);
      head = (head + 1) % LSQ_size;
      num--;
    }
}


/*
 * RS_LINK defs and decls
 */

/* a reservation station link: this structure links elements of a RUU
   reservation station list; used for ready instruction queue, event queue, and
   output dependency lists; each RS_LINK node contains a pointer to the RUU
   entry it references along with an instance tag, the RS_LINK is only valid if
   the instruction instance tag matches the instruction RUU entry instance tag;
   this strategy allows entries in the RUU can be squashed and reused without
   updating the lists that point to it, which significantly improves the
   performance of (all to frequent) squash events */
struct RS_link {
  struct RS_link *next;                 /* next entry in list */
  struct RUU_station *rs;               /* referenced RUU resv station */
  INST_TAG_TYPE tag;                    /* inst instance sequence number */
  union {
    SS_TIME_TYPE when;                  /* time stamp of entry (for eventq) */
    INST_SEQ_TYPE seq;                  /* inst sequence */
    int opnum;                          /* input/output operand number */
  } x;
};

/* RS link free list, grab RS_LINKs from here, when needed */
static struct RS_link *rslink_free_list[ST_NUM_TU];

/* NULL value for an RS link */
#define RSLINK_NULL_DATA		{ NULL, NULL, 0 }
static struct RS_link RSLINK_NULL = RSLINK_NULL_DATA;

/* NULL value for an RS link */
#define RSLINK_INIT(RSL, RS)						\
  ((RSL).next = NULL, (RSL).rs = (RS), (RSL).tag = (RS)->tag)

/* non-zero if RS link is NULL */
#define RSLINK_IS_NULL(LINK)            ((LINK)->rs == NULL)

/* non-zero if RS link is to a valid (non-squashed) entry */
#define RSLINK_VALID(LINK)              ((LINK)->tag == (LINK)->rs->tag)

/* extra RUU reservation station pointer */
#define RSLINK_RS(LINK)                 ((LINK)->rs)

/* get a new RS link record */
#define RSLINK_NEW(DST, RS)						\
  { struct RS_link *n_link;						\
    if (!rslink_free_list[TRI])						\
      panic("out of rs links");						\
    n_link = rslink_free_list[TRI];					\
    rslink_free_list[TRI] = rslink_free_list[TRI]->next;		\
    n_link->next = NULL;						\
    n_link->rs = (RS); n_link->tag = n_link->rs->tag;			\
    (DST) = n_link;							\
  }

/* free an RS link record */
#define RSLINK_FREE(LINK)						\
  {  struct RS_link *f_link = (LINK);					\
     f_link->rs = NULL; f_link->tag = 0;				\
     f_link->next = rslink_free_list[TRI];				\
     rslink_free_list[TRI] = f_link;					\
  }

/* FIXME: could this be faster!!! */
/* free an RS link list */
#define RSLINK_FREE_LIST(LINK)						\
  {  struct RS_link *fl_link, *fl_link_next;				\
     for (fl_link=(LINK); fl_link; fl_link=fl_link_next)		\
       {								\
	 fl_link_next = fl_link->next;					\
	 RSLINK_FREE(fl_link);						\
       }								\
  }

/* initialize the free RS_LINK pool */
static void
rslink_init(int nlinks)			/* total number of RS_LINK available */
{
  int i; 
  char *pi;
  struct RS_link *link;
  static struct RS_link *rsl[ST_NUM_TU];

  if (rsl[TRI] == NULL) {
      rsl[TRI] = (struct RS_link *)calloc( nlinks, sizeof (struct RS_link));
      if (!rsl[TRI])
          fatal("out of virtual memory");

      rslink_free_list[TRI] = NULL;
      for (i=0; i<nlinks; i++)
        {
          link = rsl[TRI] + i;
          link->next = rslink_free_list[TRI];
          rslink_free_list[TRI] = link;
        }
  } else {
      pi = (char *) rsl[TRI];
      for (i=0; i<sizeof (struct RS_link) * nlinks; i++)
	   pi[i] = 0;
	/* rebuild the link */
      rslink_free_list[TRI] = NULL;
      for (i=0; i<nlinks; i++)
        {
          link = rsl[TRI] + i;
          link->next = rslink_free_list[TRI];
          rslink_free_list[TRI] = link;
        }
  }
}

/* service all functional unit release events, this function is called
   once per cycle, and it used to step the BUSY timers attached to each
   functional unit in the function unit resource pool, as long as a functional
   unit's BUSY count is > 0, it cannot be issued an operation */
static void
ruu_release_fu(void)
{
  int i;

  /* walk all resource units, decrement busy counts by one */
  for (i=0; i<fu_pool[TRI]->num_resources; i++)
    {
      /* resource is released when BUSY hits zero */
      if (fu_pool[TRI]->resources[i].busy > 0)
	fu_pool[TRI]->resources[i].busy--;
    }

  /* we don't need to release the MB_Write ports, since they automatically
     free themselves when port.busy <= sim_cycle */
}


/*
 * the execution unit event queue implementation follows, the event queue
 * indicates which instruction will complete next, the writeback handler
 * drains this queue
 */

/* pending event queue, sorted from soonest to latest event (in time), NOTE:
   RS_LINK nodes are used for the event queue list so that it need not be
   updated during squash events */
static struct RS_link *event_queue[ST_NUM_TU];

/* initialize the event queue structures */
static void
eventq_init(void)
{
  event_queue[TRI] = NULL;
}

/* dump the contents of the event queue */
static void
eventq_dump(FILE *stream)			/* output stream */
{
  struct RS_link *ev;

  fprintf(stream, "** event queue state **\n");

  for (ev = event_queue[TRI]; ev != NULL; ev = ev->next)
    {
      /* is event still valid? */
      if (RSLINK_VALID(ev))
	{
	  struct RUU_station *rs = RSLINK_RS(ev);

	  fprintf(stream, "idx: %2d: @ %.0f\n",
		  rs - (rs->in_LSQ ? LSQ[TRI] : RUU[TRI]), (double)ev->x.when);
	  ruu_dumpent(rs, rs - (rs->in_LSQ ? LSQ[TRI] : RUU[TRI]),
		      stream, /* !header */FALSE);
	}
    }
}

/* insert an event for RS into the event queue, event queue is sorted from
   earliest to latest event, event and associated side-effects will be
   apparent at the start of cycle WHEN */
static void
eventq_queue_event(struct RUU_station *rs, SS_TIME_TYPE when)
{
  struct RS_link *prev, *ev, *new_ev;

  if (rs->completed)
    panic("event completed");

  if (when <= sim_cycle)
    panic("event occurred in the past");

  /* get a free event record */
  RSLINK_NEW(new_ev, rs);
  new_ev->x.when = when;

  /* locate insertion point */
  for (prev=NULL, ev=event_queue[TRI];
       ev && ev->x.when < when;
       prev=ev, ev=ev->next);

  if (prev)
    {
      /* insert middle or end */
      new_ev->next = prev->next;
      prev->next = new_ev;
    }
  else
    {
      /* insert at beginning */
      new_ev->next = event_queue[TRI];
      event_queue[TRI] = new_ev;
    }
}

/* return the next event that has already occurred, returns NULL when no
   remaining events or all remaining events are in the future */
static struct RUU_station *
eventq_next_event(void)
{
  struct RS_link *ev;

  if (event_queue[TRI] && event_queue[TRI]->x.when <= sim_cycle)
    {
      /* unlink and return first event on priority list */
      ev = event_queue[TRI];
      event_queue[TRI] = event_queue[TRI]->next;

      /* event still valid? */
      if (RSLINK_VALID(ev))
	{
	  struct RUU_station *rs = RSLINK_RS(ev);

	  /* reclaim event record */
	  RSLINK_FREE(ev);

	  /* event is valid, return resv station */
	  return rs;
	}
      else
	{
	  /* reclaim event record */
	  RSLINK_FREE(ev);

	  /* receiving inst was squashed, return next event */
	  return eventq_next_event();
	}
    }
  else
    {
      /* no event or no event is ready */
      return NULL;
    }
}


/*
 * the ready instruction queue implementation follows, the ready instruction
 * queue indicates which instruction have all of there *register* dependencies
 * satisfied, instruction will issue when 1) all memory dependencies for
 * the instruction have been satisfied (see lsq_refresh() for details on how
 * this is accomplished) and 2) resources are available; ready queue is fully
 * constructed each cycle before any operation is issued from it -- this
 * ensures that instruction issue priorities are properly observed; NOTE:
 * RS_LINK nodes are used for the event queue list so that it need not be
 * updated during squash events
 */

/* the ready instruction queue */
static struct RS_link *ready_queue[ST_NUM_TU];

/* initialize the event queue structures */
static void
readyq_init(void)
{
  ready_queue[TRI] = NULL;
}

/* dump the contents of the ready queue */
static void
readyq_dump(FILE *stream)			/* output stream */
{
  struct RS_link *link;

  fprintf(stream, "** ready queue state **\n");

  for (link = ready_queue[TRI]; link != NULL; link = link->next)
    {
      /* is entry still valid? */
      if (RSLINK_VALID(link))
	{
	  struct RUU_station *rs = RSLINK_RS(link);

	  ruu_dumpent(rs, rs - (rs->in_LSQ ? LSQ[TRI] : RUU[TRI]),
		      stream, /* header */TRUE);
	}
    }
}

/* insert ready node into the ready list using ready instruction scheduling
   policy; currently the following scheduling policy is enforced:

     memory and long latency operands, and branch instructions first

   then

     all other instructions, oldest instructions first

  this policy works well because branches pass through the machine quicker
  which works to reduce branch misprediction latencies, and very long latency
  instructions (such loads and multiplies) get priority since they are very
  likely on the program's critical path */
static void
readyq_enqueue(struct RUU_station *rs)		/* RS to enqueue */
{
  struct RS_link *prev, *node, *new_node;

  /* node is now queued */
  if (rs->queued)
    panic("node is already queued");
  rs->queued = TRUE;

  /* get a free ready list node */
  RSLINK_NEW(new_node, rs);
  new_node->x.seq = rs->seq;

  /* locate insertion point */
  if (rs->in_LSQ || SS_OP_FLAGS(rs->op) & (F_LONGLAT|F_CTRL))
    {
      /* insert loads/stores and long latency ops at the head of the queue */
      prev = NULL;
      node = ready_queue[TRI];
    }
  else
    {
      /* otherwise insert in program order (earliest seq first) */
      for (prev=NULL, node=ready_queue[TRI];
	   node && node->x.seq < rs->seq;
	   prev=node, node=node->next);
    }

  if (prev)
    {
      /* insert middle or end */
      new_node->next = prev->next;
      prev->next = new_node;
    }
  else
    {
      /* insert at beginning */
      new_node->next = ready_queue[TRI];
      ready_queue[TRI] = new_node;
    }
}


/*
 * the create vector maps a logical register to a creator in the RUU (and
 * specific output operand) or the architected register file (if RS_link
 * is NULL)
 */

/* an entry in the create vector */
struct CV_link {
  struct RUU_station *rs;               /* creator's reservation station */
  int odep_num;                         /* specific output operand */
};

/* a NULL create vector entry */
static struct CV_link CVLINK_NULL = { NULL, 0 };

/* get a new create vector link */
#define CVLINK_INIT(CV, RS,ONUM)	((CV).rs = (RS), (CV).odep_num = (ONUM))

/* size of the create vector (one entry per architected register) */
#define CV_BMAP_SZ              (BITMAP_SIZE(SS_TOTAL_REGS))

/* the create vector, NOTE: speculative copy on write storage provided
   for fast recovery during wrong path execute (see tracer_recover() for
   details on this process */
static BITMAP_TYPE(SS_TOTAL_REGS, use_spec_cv[ST_NUM_TU]);
static struct CV_link create_vector[ST_NUM_TU][SS_TOTAL_REGS];
static struct CV_link spec_create_vector[ST_NUM_TU][SS_TOTAL_REGS];

/* these arrays shadow the create vector an indicate when a register was
   last created */
static SS_TIME_TYPE create_vector_rt[ST_NUM_TU][SS_TOTAL_REGS];
static SS_TIME_TYPE spec_create_vector_rt[ST_NUM_TU][SS_TOTAL_REGS];

/* read a create vector entry */
#define CREATE_VECTOR(N)      (BITMAP_SET_P(use_spec_cv[TRI], CV_BMAP_SZ, (N))\
				 ? spec_create_vector[TRI][N]                \
				 : create_vector[TRI][N])

/* read a create vector timestamp entry */
#define CREATE_VECTOR_RT(N)     (BITMAP_SET_P(use_spec_cv[TRI],CV_BMAP_SZ,(N))\
				 ? spec_create_vector_rt[TRI][N]             \
				 : create_vector_r[TRI]t[N])

/* set a create vector entry */
#define SET_CREATE_VECTOR(N, L) (spec_mode[TRI]                              \
			    ? (BITMAP_SET(use_spec_cv[TRI], CV_BMAP_SZ, (N)),\
				    spec_create_vector[TRI][N] = (L))        \
				 : (create_vector[TRI][N] = (L)))

/* initialize the create vector */
static void
cv_init(void)
{
  int i;

  /* initially all registers are valid in the architected register file,
     i.e., the create vector entry is CVLINK_NULL */
  for (i=0; i<SS_TOTAL_REGS; i++)
    {
      create_vector[TRI][i] = CVLINK_NULL;
      create_vector_rt[TRI][i] = 0;
      spec_create_vector[TRI][i] = CVLINK_NULL;
      spec_create_vector_rt[TRI][i] = 0;
    }

  /* all create vector entries are non-speculative */
  BITMAP_CLEAR_MAP(use_spec_cv[TRI], CV_BMAP_SZ);
}

/* dependency index names */
static char *dep_names[SS_TOTAL_REGS] = {
  "n/a", "$r1", "$r2", "$r3", "$r4", "$r5", "$r6", "$r7", "$r8",  "$r9",
  "$r10", "$r11", "$r12", "$r13", "$r14", "$r15", "$r16", "$r17", "$r18",
  "$r19", "$r20", "$r21", "$r22", "$r23", "$r24", "$r25", "$r26", "$r27",
  "$r28",  "$r29", "$r30", "$r31",
  "$f0", "$f1", "$f2", "$f3", "$f4", "$f5", "$f6", "$f7", "$f8",  "$f9",
  "$f10", "$f11", "$f12", "$f13", "$f14", "$f15", "$f16", "$f17", "$f18",
  "$f19", "$f20", "$f21", "$f22", "$f23", "$f24", "$f25", "$f26", "$f27",
  "$f28",  "$f29", "$f30", "$f31",
  "$hi", "$lo", "$fcc", "$tmp",
  "n/a", "n/a"
};

/* dump the contents of the create vector */
static void
cv_dump(FILE *stream)				/* output stream */
{
  int i;
  struct CV_link ent;

  fprintf(stream, "** create vector state **\n");

  for (i=0; i<SS_TOTAL_REGS; i++)
    {
      ent = CREATE_VECTOR(i);
      if (!ent.rs)
	fprintf(stream, "[%4s]: from architected reg file\n", dep_names[i]);
      else
	fprintf(stream, "[%4s]: from %s, idx: %d\n",
	      dep_names[i], (ent.rs->in_LSQ ? "LSQ" : "RUU"),
		ent.rs - (ent.rs->in_LSQ ? LSQ[TRI] : RUU[TRI]));
    }
}


/*
 *  RUU_COMMIT() - instruction retirement pipeline stage
 */

/* this function commits the results of the oldest completed entries from the
   RUU and LSQ to the architected reg file, stores in the LSQ will commit
   their store data to the data cache at this point as well */
static void
ruu_commit(void)
{
  int i, lat, events, committed = 0;

  /* all values must be retired to the architected reg file in program order */
  while (RUU_num[TRI] > 0 && committed < ruu_commit_width)
    {
      struct RUU_station *rs = &(RUU[TRI][RUU_head[TRI]]);

      if (!rs->completed)
	{
	  /* at least RUU entry must be complete */
	  break;
	}

      /* default commit events */
      events = 0;

      /* load/stores must retire load/store queue entry as well */
      if (RUU[TRI][RUU_head[TRI]].ea_comp)
	{
	  /* load/store, retire head of LSQ as well */
	  if (LSQ_num[TRI] <= 0 || !LSQ[TRI][LSQ_head[TRI]].in_LSQ)
	    panic("RUU out of sync with LSQ");

	  /* load/store operation must be complete */
	  if (!LSQ[TRI][LSQ_head[TRI]].completed)
	    {
	      /* load/store operation is not yet complete */
	      break;
	    }

	  if ((SS_OP_FLAGS(LSQ[TRI][LSQ_head[TRI]].op) & (F_MEM|F_STORE))
	      == (F_MEM|F_STORE))
	    {
	      struct res_template *fu;

	      if (head_thread_id != pid) {
	          int i;
		  i = get_mb_port();
		  if (i<0) /* no ports left, can not continue to commit */
		     break;
		  mb_port[i].busy = sim_cycle + MB_LATENCY;
		  lat = MB_LATENCY;
	      } else {

	        /* stores must retire their store value to the cache at commit,
		 try to get a store port (functional unit allocation) */
	        fu = res_get(fu_pool[TRI],
				SS_OP_FUCLASS(LSQ[TRI][LSQ_head[TRI]].op));
	        if (fu)
		{
		  /* reserve the functional unit */
		  if (fu->master->busy)
		    panic("functional unit already in use");

		  /* schedule functional unit release event */
		  fu->master->busy = fu->issuelat;

		  /* go to the data cache */
		  if (cache_dl1)
		    {
		      /* commit store value to D-cache */
		      lat =
			cache_access(cache_dl1, Write, 
				     (LSQ[TRI][LSQ_head[TRI]].addr&~3),
				     NULL, 4, sim_cycle, NULL, NULL);
		      if (lat > cache_dl1_lat)
			events |= PEV_CACHEMISS;
		    }

		  /* all loads and stores must access D-TLB */
		  if (dtlb)
		    {
		      /* access the D-TLB */
		      lat =
			cache_access(dtlb, Read, 
				     (LSQ[TRI][LSQ_head[TRI]].addr & ~3),
				     NULL, 4, sim_cycle, NULL, NULL);
		      if (lat > 1)
			events |= PEV_TLBMISS;
		    }
		}
	        else
		{
		  /* no store ports left, cannot continue to commit insts */
		  break;
		}
	      } /* end if else not head_thread */
	    }

	  /* invalidate load/store operation instance */
	  LSQ[TRI][LSQ_head[TRI]].tag++;

	  /* indicate to pipeline trace that this instruction retired */
	  ptrace_newstage(LSQ[TRI][LSQ_head[TRI]].ptrace_seq, PST_COMMIT, events);
	  ptrace_endinst(LSQ[TRI][LSQ_head[TRI]].ptrace_seq);

	  /* commit head of LSQ as well */
	  LSQ_head[TRI] = (LSQ_head[TRI] + 1) % LSQ_size;
	  LSQ_num[TRI]--;
	}

      if (pred
	  && bpred_spec_update[TRI] == spec_CT
	  && (SS_OP_FLAGS(rs->op) & F_CTRL))
	{
	  SS_INST_TYPE inst = rs->IR;

	  bpred_update(pred,
		       /* branch address */rs->PC,
		       /* actual target address */rs->next_PC,
                       /* taken? */rs->next_PC != (rs->PC +
                                                   sizeof(SS_INST_TYPE)),
                       /* pred taken? */rs->pred_PC != (rs->PC +
                                                        sizeof(SS_INST_TYPE)),
                       /* correct pred? */rs->pred_PC == rs->next_PC,
                       /* opcode */rs->op,
		       /* jump through R31? */(RS) == 31,
                       /* dir predictor update pointer */&rs->dir_update);
	}

      /* invalidate RUU operation instance */
      RUU[TRI][RUU_head[TRI]].tag++;

      /* indicate to pipeline trace that this instruction retired */
      ptrace_newstage(RUU[TRI][RUU_head[TRI]].ptrace_seq, PST_COMMIT, events);
      ptrace_endinst(RUU[TRI][RUU_head[TRI]].ptrace_seq);

      /* commit head entry of RUU */
      RUU_head[TRI] = (RUU_head[TRI] + 1) % RUU_size;
      RUU_num[TRI]--;

      /* one more instruction committed to architected state */
      committed++;

      for (i=0; i<MAX_ODEPS; i++)
	{
	  if (rs->odep_list[i])
	    panic ("retired instruction has odeps\n");
        }
    }
}


/*
 *  RUU_RECOVER() - squash mispredicted microarchitecture state
 */

/* recover processor microarchitecture state back to point of the
   mis-predicted branch at RUU[BRANCH_INDEX] */
static void
ruu_recover(int branch_index)			/* index of mis-pred branch */
{
  int i, RUU_index = RUU_tail[TRI], LSQ_index = LSQ_tail[TRI];
  int RUU_prev_tail = RUU_tail[TRI], LSQ_prev_tail = LSQ_tail[TRI];

  /* recover from the tail of the RUU towards the head until the branch index
     is reached, this direction ensures that the LSQ can be synchronized with
     the RUU */

  /* go to first element to squash */
  RUU_index = (RUU_index + (RUU_size-1)) % RUU_size;
  LSQ_index = (LSQ_index + (LSQ_size-1)) % LSQ_size;

  /* traverse to older insts until the mispredicted branch is encountered */
  while (RUU_index != branch_index)
    {
      /* the RUU should not drain since the mispredicted branch will remain */
      if (!RUU_num[TRI])
	panic("empty RUU");

      /* should meet up with the tail first */
      if (RUU_index == RUU_head[TRI])
	panic("RUU head and tail broken");

      /* is this operation an effective addr calc for a load or store? */
      if (RUU[TRI][RUU_index].ea_comp)
	{
	  /* should be at least one load or store in the LSQ */
	  if (!LSQ_num[TRI])
	    panic("RUU and LSQ out of sync");

	  /* recover any resources consumed by the load or store operation */
	  for (i=0; i<MAX_ODEPS; i++)
	    {
	      RSLINK_FREE_LIST(LSQ[TRI][LSQ_index].odep_list[i]);
	      /* blow away the consuming op list */
	      LSQ[TRI][LSQ_index].odep_list[i] = NULL;
	    }
      
	  /* squash this LSQ entry */
	  LSQ[TRI][LSQ_index].tag++;

	  /* indicate in pipetrace that this instruction was squashed */
	  ptrace_endinst(LSQ[TRI][LSQ_index].ptrace_seq);

	  /* go to next earlier LSQ slot */
	  LSQ_prev_tail = LSQ_index;
	  LSQ_index = (LSQ_index + (LSQ_size-1)) % LSQ_size;
	  LSQ_num[TRI]--;
	}

      /* recover any resources used by this RUU operation */
      for (i=0; i<MAX_ODEPS; i++)
	{
	  RSLINK_FREE_LIST(RUU[TRI][RUU_index].odep_list[i]);
	  /* blow away the consuming op list */
	  RUU[TRI][RUU_index].odep_list[i] = NULL;
	}
      
      /* squash this RUU entry */
      RUU[TRI][RUU_index].tag++;

      /* indicate in pipetrace that this instruction was squashed */
      ptrace_endinst(RUU[TRI][RUU_index].ptrace_seq);

      /* go to next earlier slot in the RUU */
      RUU_prev_tail = RUU_index;
      RUU_index = (RUU_index + (RUU_size-1)) % RUU_size;
      RUU_num[TRI]--;
    }

  /* reset head/tail pointers to point to the mis-predicted branch */
  RUU_tail[TRI] = RUU_prev_tail;
  LSQ_tail[TRI] = LSQ_prev_tail;

  /* revert create vector back to last precise create vector state, NOTE:
     this is accomplished by resetting all the copied-on-write bits in the
     USE_SPEC_CV bit vector */
  BITMAP_CLEAR_MAP(use_spec_cv[TRI], CV_BMAP_SZ);

  /* FIXME: could reset functional units at squash time */
}


/*
 *  RUU_WRITEBACK() - instruction result writeback pipeline stage
 */

/* forward declarations */
static void tracer_recover(void);

/* writeback completed operation results from the functional units to RUU,
   at this point, the output dependency chains of completing instructions
   are also walked to determine if any dependent instruction now has all
   of its register operands, if so the (nearly) ready instruction is inserted
   into the ready instruction queue */
static void
ruu_writeback(void)
{
  int i;
  struct RUU_station *rs;

  /* service all completed events */
  while ((rs = eventq_next_event()))
    {
      /* RS has completed execution and (possibly) produced a result */
      if (!OPERANDS_READY(rs) || rs->queued || !rs->issued || rs->completed)
	panic("inst completed and !ready, !issued, or completed");

      /* operation has completed */
      rs->completed = TRUE;

      /* does this operation reveal a mis-predicted branch? */
      if (rs->recover_inst)
	{
	  if (rs->in_LSQ)
	    panic("mis-predicted load or store?!?!?");

	  /* recover processor state and reinit fetch to correct path */
	  ruu_recover(rs - RUU[TRI]);
	  tracer_recover();
	  bpred_recover(pred, rs->PC, rs->stack_recover_idx);

	  /* stall fetch until I-fetch and I-decode recover */
	  ruu_fetch_issue_delay[TRI] = ruu_branch_penalty;

	  /* continue writeback of the branch/control instruction */
	}

      /* if we speculatively update branch-predictor, do it here */
      if (pred
	  && bpred_spec_update[TRI] == spec_WB
	  && !rs->in_LSQ
	  && (SS_OP_FLAGS(rs->op) & F_CTRL))
	{
	  SS_INST_TYPE inst = rs->IR;

	  bpred_update(pred,
		       /* branch address */rs->PC,
		       /* actual target address */rs->next_PC,
		       /* taken? */rs->next_PC != (rs->PC +
						   sizeof(SS_INST_TYPE)),
		       /* pred taken? */rs->pred_PC != (rs->PC +
							sizeof(SS_INST_TYPE)),
		       /* correct pred? */rs->pred_PC == rs->next_PC,
		       /* opcode */rs->op,
		       /* jump through R31? */(RS) == 31,
		       /* dir predictor update pointer */&rs->dir_update);
	}

      /* entered writeback stage, indicate in pipe trace */
      ptrace_newstage(rs->ptrace_seq, PST_WRITEBACK,
		      rs->recover_inst ? PEV_MPDETECT : 0);

      /* broadcast results to consuming operations, this is more efficiently
         accomplished by walking the output dependency chains of the
	 completed instruction */
      for (i=0; i<MAX_ODEPS; i++)
	{
	  if (rs->onames[i] != NA)
	    {
	      struct CV_link link;
	      struct RS_link *olink, *olink_next;

	      if (rs->spec_mode)
		{
		  /* update the speculative create vector, future operations
		     get value from later creator or architected reg file */
		  link = spec_create_vector[TRI][rs->onames[i]];
		  if (/* !NULL */link.rs
		      && /* refs RS */(link.rs == rs && link.odep_num == i))
		    {
		      /* the result can now be read from a physical register,
			 indicate this as so */
		      spec_create_vector[TRI][rs->onames[i]] = CVLINK_NULL;
		      spec_create_vector_rt[TRI][rs->onames[i]] = sim_cycle;
		    }
		  /* else, creator invalidated or there is another creator */
		}
	      else
		{
		  /* update the non-speculative create vector, future
		     operations get value from later creator or architected
		     reg file */
		  link = create_vector[TRI][rs->onames[i]];
		  if (/* !NULL */link.rs
		      && /* refs RS */(link.rs == rs && link.odep_num == i))
		    {
		      /* the result can now be read from a physical register,
			 indicate this as so */
		      create_vector[TRI][rs->onames[i]] = CVLINK_NULL;
		      create_vector_rt[TRI][rs->onames[i]] = sim_cycle;
		    }
		  /* else, creator invalidated or there is another creator */
		}

	      /* walk output list, queue up ready operations */
	      for (olink=rs->odep_list[i]; olink; olink=olink_next)
		{
		  if (RSLINK_VALID(olink))
		    {
		      if (olink->rs->idep_ready[olink->x.opnum])
			panic("output dependence already satisfied");

		      /* input is now ready */
		      olink->rs->idep_ready[olink->x.opnum] = TRUE;

		      /* are all the register operands of target ready? */
		      if (OPERANDS_READY(olink->rs))
			{
			  /* yes! enqueue instruction as ready, NOTE: stores
			     complete at dispatch, so no need to enqueue
			     them */
			  if (!olink->rs->in_LSQ
			      || ((SS_OP_FLAGS(olink->rs->op)&(F_MEM|F_STORE))
				  == (F_MEM|F_STORE)))
			    readyq_enqueue(olink->rs);
			  /* else, ld op, issued when no mem conflict */
			}
		    }

		  /* grab link to next element prior to free */
		  olink_next = olink->next;

		  /* free dependence link element */
		  RSLINK_FREE(olink);
		}
	      /* blow away the consuming op list */
	      rs->odep_list[i] = NULL;

	    } /* if not NA output */

	} /* for all outputs */

   } /* for all writeback events */

}


/*
 *  LSQ_REFRESH() - memory access dependence checker/scheduler
 */

/* this function locates ready instructions whose memory dependencies have
   been satisfied, this is accomplished by walking the LSQ for loads, looking
   for blocking memory dependency condition (e.g., earlier store with an
   unknown address) */
#define MAX_STD_UNKNOWNS		64
static void
lsq_refresh(void)
{
  int i, j, index, n_std_unknowns;
  SS_ADDR_TYPE std_unknowns[MAX_STD_UNKNOWNS];

  /* scan entire queue for ready loads: scan from oldest instruction
     (head) until we reach the tail or an unresolved store, after which no
     other instruction will become ready */
  for (i=0, index=LSQ_head[TRI], n_std_unknowns=0;
       i < LSQ_num[TRI];
       i++, index=(index + 1) % LSQ_size)
    {
      /* terminate search for ready loads after first unresolved store,
	 as no later load could be resolved in its presence */
      if (/* store? */
	  (SS_OP_FLAGS(LSQ[TRI][index].op) & (F_MEM|F_STORE)) == (F_MEM|F_STORE))
	{
	  if (!STORE_ADDR_READY(&LSQ[TRI][index]))
	    {
	      /* FIXME: a later STD + STD known could hide the STA unknown */
	      /* sta unknown, blocks all later loads, stop search */
	      break;
	    }
	  else if (!OPERANDS_READY(&LSQ[TRI][index]))
	    {
	      /* sta known, but std unknown, may block a later store, record
		 this address for later referral, we use an array here because
		 for most simulations the number of entries to search will be
		 very small */
	      if (n_std_unknowns == MAX_STD_UNKNOWNS)
		fatal("STD unknown array overflow, increase MAX_STD_UNKNOWNS");
	      std_unknowns[n_std_unknowns++] = LSQ[TRI][index].addr;
	    }
	  else /* STORE_ADDR_READY() && OPERANDS_READY() */
	    {
	      /* a later STD known hides an earlier STD unknown */
	      for (j=0; j<n_std_unknowns; j++)
		{
		  if (std_unknowns[j]== /* STA/STD known */LSQ[TRI][index].addr)
		    std_unknowns[j] = /* bogus addr */0;
		}
	    }
	}

      if (/* load? */
	  ((SS_OP_FLAGS(LSQ[TRI][index].op) & (F_MEM|F_LOAD)) == (F_MEM|F_LOAD))
	  && /* queued? */!LSQ[TRI][index].queued
	  && /* waiting? */!LSQ[TRI][index].issued
	  && /* completed? */!LSQ[TRI][index].completed
	  && /* regs ready? */OPERANDS_READY(&LSQ[TRI][index]))
	{ int p;
	  /* no STA unknown conflict (because we got to this check), check for
	     a STD unknown conflict */
	  for (j=0; j<n_std_unknowns; j++)
	    {
	      /* found a relevant STD unknown? */
	      if (std_unknowns[j] == LSQ[TRI][index].addr)
		break;
	    }
	  p = get_mb_port();
	  if (p>=0)
	     mb_port[p].busy = sim_cycle+1;
	  if (j == n_std_unknowns && p>=0 && 
		(!WAIT_TSAG_DONE || pid == head_thread_id) &&
		in_mb(LSQ[TRI][index].addr) != ADDR_VALID)
	    {
	      /* no STA or STD unknown conflicts, put load on ready queue */
	      readyq_enqueue(&LSQ[TRI][index]);
	    }
	}
    }
}


/*
 *  RUU_ISSUE() - issue instructions to functional units
 */

/* attempt to issue all operations in the ready queue; insts in the ready
   instruction queue have all register dependencies satisfied, this function
   must then 1) ensure the instructions memory dependencies have been satisfied
   (see lsq_refresh() for details on this process) and 2) a function unit
   is available in this cycle to commence execution of the operation; if all
   goes well, the function unit is allocated, a writeback event is scheduled,
   and the instruction begins execution */
static void
ruu_issue(void)
{
  int i, load_lat, tlb_lat, n_issued;
  struct RS_link *node, *next_node;
  struct res_template *fu;

  /* FIXME: could be a little more efficient when scanning the ready queue */

  /* copy and then blow away the ready list, NOTE: the ready list is
     always totally reclaimed each cycle, and instructions that are not
     issue are explicitly reinserted into the ready instruction queue,
     this management strategy ensures that the ready instruction queue
     is always properly sorted */
  node = ready_queue[TRI];
  ready_queue[TRI] = NULL;

  /* visit all ready instructions (i.e., insts whose register input
     dependencies have been satisfied, stop issue when no more instructions
     are available or issue bandwidth is exhausted */
  for (n_issued=0;
       node && n_issued < ruu_issue_width;
       node = next_node)
    {
      next_node = node->next;

      /* still valid? */
      if (RSLINK_VALID(node))
	{
	  struct RUU_station *rs = RSLINK_RS(node);

	  /* issue operation, both reg and mem deps have been satisfied */
	  if (!OPERANDS_READY(rs) || !rs->queued
	      || rs->issued || rs->completed)
	    panic("issued inst !ready, issued, or completed");

	  /* node is now un-queued */
	  rs->queued = FALSE;

	  if (rs->in_LSQ
	      && ((SS_OP_FLAGS(rs->op) & (F_MEM|F_STORE)) == (F_MEM|F_STORE)))
	    {
	      /* stores complete in effectively zero time, result is
		 written into the load/store queue, the actual store into
		 the memory system occurs when the instruction is retired
		 (see ruu_commit()) */
	      rs->issued = TRUE;
	      rs->completed = TRUE;

	      if (rs->onames[0] || rs->onames[1])
		panic("store creates result");

	      if (rs->recover_inst)
		panic("mis-predicted store");

	      /* entered execute stage, indicate in pipe trace */
	      ptrace_newstage(rs->ptrace_seq, PST_WRITEBACK, 0);

	      /* one more inst issued */
	      n_issued++;
	    }
	  else
	    {
	      /* issue the instruction to a functional unit */
	      if (SS_OP_FUCLASS(rs->op) != NA)
		{
		  fu = res_get(fu_pool[TRI], SS_OP_FUCLASS(rs->op));
		  if (fu)
		    {
		      /* got one! issue inst to functional unit */
		      rs->issued = TRUE;

		      /* reserve the functional unit */
		      if (fu->master->busy)
			panic("functional unit already in use");

		      /* schedule functional unit release event */
		      fu->master->busy = fu->issuelat;

		      /* schedule a result writeback event */
		      if (rs->in_LSQ
			  && ((SS_OP_FLAGS(rs->op) & (F_MEM|F_LOAD))
			      == (F_MEM|F_LOAD)))
			{
			  int events = 0, j;

			  /* for loads, determine cache access latency:
			     first scan LSQ to see if a store forward is
			     possible, if not, access the data cache */
			  load_lat = 0;
			  i = (rs - LSQ[TRI]);
			  if (i != LSQ_head[TRI])
			    {
			      for (;;)
				{
				  /* go to next earlier LSQ entry */
				  i = (i + (LSQ_size-1)) % LSQ_size;

				  /* FIXME: not dealing with partials! */
				  if ((SS_OP_FLAGS(LSQ[TRI][i].op) & F_STORE)
				      && (LSQ[TRI][i].addr == rs->addr))
				    {
				      /* hit in the LSQ */
				      load_lat = 1;
				      break;
				    }

				  /* scan finished? */
				  if (i == LSQ_head[TRI])
				    break;
				}
			    }

#ifdef SUPER_TR 
			  if (!In_Critical_Region && !load_lat && 
			       pid != head_thread_id && (j=in_mb(rs->addr))) {
				/* load from Memory Buffer, no need to
				   access cache or TLB */
				if (j == DATA_VALID)
				    load_lat = MB_LATENCY;
			  }
#endif

			  /* was the value store forwared from the LSQ? */
			  if (!load_lat)
			    {
			      /* no! go to the data cache */
			      if (cache_dl1
				  /* valid address? */
				  && (rs->addr >= ld_data_base
				      && rs->addr < ld_stack_base))
				{
				  /* access the cache if non-faulting */
				  load_lat =
				    cache_access(cache_dl1, Read,
						 (rs->addr & ~3), NULL, 4,
						 sim_cycle, NULL, NULL);
				  if (load_lat > cache_dl1_lat)
				    events |= PEV_CACHEMISS;
				}
			      else
				{
				  /* no caches defined, just use op latency */
				  load_lat = fu->oplat;
				}
			    }

			  /* all loads and stores must access D-TLB */
			  if (dtlb
			      /* valid address? */
			      && (rs->addr >= ld_data_base
				  && rs->addr < ld_stack_base))
			    {
			      /* access the D-DLB, NOTE: this code will
				 initiate speculative TLB misses */
			      tlb_lat =
				cache_access(dtlb, Read, (rs->addr & ~3),
					     NULL, 4, sim_cycle, NULL, NULL);
			      if (tlb_lat > 1)
				events |= PEV_TLBMISS;

			      /* D-cache/D-TLB accesses occur in parallel */
			      load_lat = MAX(tlb_lat, load_lat);
			    }

			  /* use computed cache access latency */
			  eventq_queue_event(rs, sim_cycle + load_lat);

			  /* entered execute stage, indicate in pipe trace */
			  ptrace_newstage(rs->ptrace_seq, PST_EXECUTE,
					  ((rs->ea_comp ? PEV_AGEN : 0)
					   | events));
			}
		      else /* !load && !store */
			{
			  /* use deterministic functional unit latency */
			  eventq_queue_event(rs, sim_cycle + fu->oplat);

			  /* entered execute stage, indicate in pipe trace */
			  ptrace_newstage(rs->ptrace_seq, PST_EXECUTE, 
					  rs->ea_comp ? PEV_AGEN : 0);
			}

		      /* one more inst issued */
		      n_issued++;
		    }
		  else /* no functional unit */
		    {
		      /* insufficient functional unit resources, put operation
			 back onto the ready list, we'll try to issue it
			 again next cycle */
		      readyq_enqueue(rs);
		    }
		}
	      else /* does not require a functional unit! */
		{
		  /* FIXME: need better solution for these */
		  /* the instruction does not need a functional unit */
		  rs->issued = TRUE;

		  /* schedule a result event */
		  eventq_queue_event(rs, sim_cycle + 1);

		  /* entered execute stage, indicate in pipe trace */
		  ptrace_newstage(rs->ptrace_seq, PST_EXECUTE,
				  rs->ea_comp ? PEV_AGEN : 0);

		  /* one more inst issued */
		  n_issued++;
		  if (rs->op == WTSAGD || rs->op == GLOCK)
			n_issued = ruu_issue_width;	
		}
	    } /* !store */

	}
      /* else, RUU entry was squashed */

      /* reclaim ready list entry, NOTE: this is done whether or not the
         instruction issued, since the instruction was once again reinserted
         into the ready queue if it did not issue, this ensures that the ready
         queue is always properly sorted */
      RSLINK_FREE(node);
    }

  /* put any instruction not issued back into the ready queue, go through
     normal channels to ensure instruction stay ordered correctly */
  for (; node; node = next_node)
    {
      next_node = node->next;

      /* still valid? */
      if (RSLINK_VALID(node))
        {
	  struct RUU_station *rs = RSLINK_RS(node);

          /* node is now un-queued */
          rs->queued = FALSE;

	  /* not issued, put operation back onto the ready list, we'll try to
	     issue it again next cycle */
          readyq_enqueue(rs);
        }
      /* else, RUU entry was squashed */

      /* reclaim ready list entry, NOTE: this is done whether or not the
         instruction issued, since the instruction was once again reinserted
         into the ready queue if it did not issue, this ensures that the ready
         queue is always properly sorted */
      RSLINK_FREE(node);
    }
}


/*
 * routines for generating on-the-fly instruction traces with support
 * for control and data misspeculation modeling
 */

/* integer register file */
#define R_BMAP_SZ       (BITMAP_SIZE(SS_NUM_REGS))
static BITMAP_TYPE(SS_NUM_REGS, use_spec_R[ST_NUM_TU]);
static SS_WORD_TYPE spec_regs_R[ST_NUM_TU][SS_NUM_REGS];

/* floating point register file */
#define F_BMAP_SZ       (BITMAP_SIZE(SS_NUM_REGS))
static BITMAP_TYPE(SS_NUM_REGS, use_spec_F[ST_NUM_TU]);
static union regs_FP spec_regs_F[ST_NUM_TU];

 /* miscellaneous registers */
static int use_spec_HI[ST_NUM_TU];
static SS_WORD_TYPE spec_regs_HI[ST_NUM_TU];
static int use_spec_LO[ST_NUM_TU];
static SS_WORD_TYPE spec_regs_LO[ST_NUM_TU];
static int use_spec_FCC[ST_NUM_TU];
static int spec_regs_FCC[ST_NUM_TU];

/* dump speculative register state */
static void
rspec_dump(FILE *stream)			/* output stream */
{
  int i;

  fprintf(stream, "** speculative register contents **\n");

  fprintf(stream, "spec_mode: %s\n", spec_mode[TRI] ? "t" : "f");

  /* dump speculative integer regs */
  for (i=0; i<SS_NUM_REGS; i++)
    {
      if (BITMAP_SET_P(use_spec_R[TRI], R_BMAP_SZ, i))
	{
	  /* speculative state */
	  fprintf(stream, "[%4s]: %12d/0x%08x\n", dep_names[i],
		  spec_regs_R[TRI][i], spec_regs_R[TRI][i]);
	}
    }

  /* dump speculative FP regs */
  for (i=0; i<SS_NUM_REGS; i++)
    {
      if (BITMAP_SET_P(use_spec_F[TRI], F_BMAP_SZ, i))
	{
	  /* speculative state */
	  fprintf(stream,
		  "[%4s]: %12d/0x%08x/%f ([%4s] as double: %f)\n\n",
		  dep_names[i+32],
		  spec_regs_F[TRI].l[i], spec_regs_F[TRI].l[i], 
		  spec_regs_F[TRI].f[i],
		  dep_names[i+32],
		  spec_regs_F[TRI].d[i >> 1]);
	}
    }

  /* dump speculative miscellaneous regs */
  if (use_spec_HI[TRI])
    fprintf(stream, "[ $hi]: %12d/0x%08x\n", spec_regs_HI[TRI], spec_regs_HI[TRI]);
  if (use_spec_LO[TRI])
    fprintf(stream, "[ $lo]: %12d/0x%08x\n", spec_regs_LO[TRI], spec_regs_LO[TRI]);
  if (use_spec_FCC[TRI])
    fprintf(stream, "[$fcc]: 0x%08x\n", spec_regs_FCC[TRI]);
}


/* speculative memory hash table size, NOTE: this must be a power-of-two */
#define STORE_HASH_SIZE		32

/* speculative memory hash table definition, accesses go through this hash
   table when accessing memory in speculative mode, the hash table flush the
   table when recovering from mispredicted branches */
struct spec_mem_ent {
  struct spec_mem_ent *next;		/* ptr to next hash table bucket */
  SS_ADDR_TYPE addr;			/* virtual address of spec state */
  unsigned int data[2];			/* spec buffer, up to 8 bytes */
};

/* speculative memory hash table */
static struct spec_mem_ent *store_htable[ST_NUM_TU][STORE_HASH_SIZE];

/* speculative memory hash table bucket free list */
static struct spec_mem_ent *bucket_free_list[ST_NUM_TU];


/* program counter */
static SS_ADDR_TYPE pred_PC[ST_NUM_TU];
static SS_ADDR_TYPE recover_PC[ST_NUM_TU];

/* fetch unit next fetch address */
static SS_ADDR_TYPE fetch_regs_PC[ST_NUM_TU];
static SS_ADDR_TYPE fetch_pred_PC[ST_NUM_TU];

/* IFETCH -> DISPATCH instruction queue definition */
struct fetch_rec {
  SS_INST_TYPE IR;                      /* inst register */
  SS_ADDR_TYPE regs_PC, pred_PC;        /* current PC, predicted next PC */
  struct bpred_update dir_update;	/* bpred direction update info */
  int stack_recover_idx;		/* branch predictor RSB index */
  unsigned int ptrace_seq;		/* print trace sequence id */
};
static struct fetch_rec *fetch_data[ST_NUM_TU];	/* IFETCH -> DISPATCH inst queue */
static int fetch_num[ST_NUM_TU];			/* num entries in IF -> DIS queue */
static int fetch_tail[ST_NUM_TU], fetch_head[ST_NUM_TU];	/* head and tail pointers of queue */

/* recover instruction trace generator state to precise state state immediately
   before the first mis-predicted branch; this is accomplished by resetting
   all register value copied-on-write bitmasks are reset, and the speculative
   memory hash table is cleared */
static void
tracer_recover(void)
{
  int i;
  struct spec_mem_ent *ent, *ent_next;

  /* better be in mis-speculative trace generation mode */
  if (!spec_mode[TRI])
    panic("cannot recover unless in speculative mode");

  /* reset to non-speculative trace generation mode */
  spec_mode[TRI] = FALSE;

  /* reset copied-on-write register bitmasks back to non-speculative state */
  BITMAP_CLEAR_MAP(use_spec_R[TRI], R_BMAP_SZ);
  BITMAP_CLEAR_MAP(use_spec_F[TRI], F_BMAP_SZ);
  use_spec_HI[TRI] = FALSE;
  use_spec_LO[TRI] = FALSE;
  use_spec_FCC[TRI] = FALSE;

  /* reset memory state back to non-speculative state */
  /* FIXME: could version stamps be used here?!?!? */
  for (i=0; i<STORE_HASH_SIZE; i++)
    {
      /* release all hash table buckets */
      for (ent=store_htable[TRI][i]; ent; ent=ent_next)
	{
	  ent_next = ent->next;
	  ent->next = bucket_free_list[TRI];
	  bucket_free_list[TRI] = ent;
	}
      store_htable[TRI][i] = NULL;
    }

  /* if pipetracing, indicate squash of instructions in the inst fetch queue */
  if (ptrace_active)
    {
      while (fetch_num[TRI] != 0)
	{
	  /* squash the next instruction from the IFETCH -> DISPATCH queue */
	  ptrace_endinst(fetch_data[TRI][fetch_head[TRI]].ptrace_seq);

	  /* consume instruction from IFETCH -> DISPATCH queue */
	  fetch_head[TRI] = (fetch_head[TRI]+1) & (ruu_ifq_size - 1);
	  fetch_num[TRI]--;
	}
    }

  /* reset IFETCH state */
  fetch_num[TRI] = 0;
  fetch_tail[TRI] = fetch_head[TRI] = 0;
  fetch_pred_PC[TRI] = fetch_regs_PC[TRI] = recover_PC[TRI];
}

/* initialize the speculative instruction state generator state */
static void
tracer_init(void)
{
  int i;

  /* initially in non-speculative mode */
  spec_mode[TRI] = FALSE;

  /* register state is from non-speculative state buffers */
  BITMAP_CLEAR_MAP(use_spec_R[TRI], R_BMAP_SZ);
  BITMAP_CLEAR_MAP(use_spec_F[TRI], F_BMAP_SZ);
  use_spec_HI[TRI] = FALSE;
  use_spec_LO[TRI] = FALSE;
  use_spec_FCC[TRI] = FALSE;

  /* memory state is from non-speculative memory pages */
  for (i=0; i<STORE_HASH_SIZE; i++)
    store_htable[TRI][i] = NULL;
}


/* speculative memory hash table address hash function */
#define HASH_ADDR(ADDR)							\
  ((((ADDR) >> 24)^((ADDR) >> 16)^((ADDR) >> 8)^(ADDR)) & (STORE_HASH_SIZE-1))

/* this functional provides a layer of mis-speculated state over the
   non-speculative memory state, when in mis-speculation trace generation mode,
   the simulator will call this function to access memory, instead of the
   non-speculative memory access interfaces defined in memory.h; when storage
   is written, an entry is allocated in the speculative memory hash table,
   future reads and writes while in mis-speculative trace generation mode will
   access this buffer instead of non-speculative memory state; when the trace
   generator transitions back to non-speculative trace generation mode,
   tracer_recover() clears this table */
static void
spec_mem_access(enum mem_cmd cmd,		/* Read or Write access cmd */
		SS_ADDR_TYPE addr,		/* virtual address of access */
		void *p,			/* input/output buffer */
		int nbytes)			/* number of bytes to access */
{
  int index;
  struct spec_mem_ent *ent, *prev;

  /* FIXME: partially overlapping writes are not combined... */
  /* FIXME: partially overlapping reads are not handled correctly... */

  /* check alignments, even speculative this test should always pass */
  if ((nbytes & (nbytes-1)) != 0 || (addr & (nbytes-1)) != 0)
    {
      /* no can do */
      return;
    }

  /* check permissions */
  if (!((addr >= ld_text_base && addr < (ld_text_base+ld_text_size)
	 && cmd == Read)
	|| (addr >= ld_data_base && addr < ld_stack_base)))
    {
      /* no can do */
      return;
    }

  /* has this memory state been copied on mis-speculative write? */
  index = HASH_ADDR(addr);
  for (prev=NULL,ent=store_htable[TRI][index]; ent; prev=ent,ent=ent->next)
    {
      if (ent->addr == addr)
	{
	  /* reorder chains to speed access into hash table */
	  if (prev != NULL)
	    {
	      /* not at head of list, relink the hash table entry at front */
	      prev->next = ent->next;
              ent->next = store_htable[TRI][index];
              store_htable[TRI][index] = ent;
	    }
	  break;
	}
    }

  /* no, if it is a write, allocate a hash table entry to hold the data */
  if (!ent && cmd == Write)
    {
      /* try to get an entry from the free list, if available */
      if (!bucket_free_list[TRI])
	{
	  /* otherwise, call calloc() to get the needed storage */
	  bucket_free_list[TRI] = calloc(1, sizeof(struct spec_mem_ent));
	  if (!bucket_free_list[TRI])
	    fatal("out of virtual memory");
	}
      ent = bucket_free_list[TRI];
      bucket_free_list[TRI] = bucket_free_list[TRI]->next;

      if (!bugcompat_mode)
	{
	  /* insert into hash table */
	  ent->next = store_htable[TRI][index];
	  store_htable[TRI][index] = ent;
	  ent->addr = addr;
	  ent->data[0] = 0; ent->data[1] = 0;
	}
    }

  /* handle the read or write to speculative or non-speculative storage */
  switch (nbytes)
    {
    case 1:
      if (cmd == Read)
	{
	  if (ent)
	    {
	      /* read from mis-speculated state buffer */
	      *((unsigned char *)p) = *((unsigned char *)(&ent->data[0]));
	    }
	  else
	    {
	      /* read from non-speculative memory state */
	      *((unsigned char *)p) = MEM_READ_BYTE(addr);
	    }
	}
      else
	{
	  /* always write into mis-speculated state buffer */
	  *((unsigned char *)(&ent->data[0])) = *((unsigned char *)p);
	}
      break;
    case 2:
      if (cmd == Read)
	{
	  if (ent)
	    {
	      /* read from mis-speculated state buffer */
	      *((unsigned short *)p) = *((unsigned short *)(&ent->data[0]));
	    }
	  else
	    {
	      /* read from non-speculative memory state */
	      *((unsigned short *)p) = MEM_READ_HALF(addr);
	    }
	}
      else
	{
	  /* always write into mis-speculated state buffer */
	  *((unsigned short *)&ent->data[0]) = *((unsigned short *)p);
	}
      break;
    case 4:
      if (cmd == Read)
	{
	  if (ent)
	    {
	      /* read from mis-speculated state buffer */
	      *((unsigned int *)p) = *((unsigned int *)&ent->data[0]);
	    }
	  else
	    {
	      /* read from non-speculative memory state */
	      *((unsigned int *)p) = MEM_READ_WORD(addr);
	    }
	}
      else
	{
	  /* always write into mis-speculated state buffer */
	  *((unsigned int *)&ent->data[0]) = *((unsigned int *)p);
	}
      break;
    case 8:
      if (cmd == Read)
	{
	  if (ent)
	    {
	      /* read from mis-speculated state buffer */
	      *((unsigned int *)p) = *((unsigned int *)&ent->data[0]);
	      *(((unsigned int *)p)+1) = *((unsigned int *)&ent->data[1]);
	    }
	  else
	    {
	      /* read from non-speculative memory state */
	      *((unsigned int *)p) = MEM_READ_WORD(addr);
	      *(((unsigned int *)p)+1) =
		MEM_READ_WORD(addr + sizeof(SS_WORD_TYPE));
	    }
	}
      else
	{
	  /* always write into mis-speculated state buffer */
	  *((unsigned int *)&ent->data[0]) = *((unsigned int *)p);
	}
      break;
    default:
      panic("access size not supported in mis-speculative mode");
    }
}

/* dump speculative memory state */
static void
mspec_dump(FILE *stream)			/* output stream */
{
  int i;
  struct spec_mem_ent *ent;

  fprintf(stream, "** speculative memory contents **\n");

  fprintf(stream, "spec_mode: %s\n", spec_mode[TRI] ? "t" : "f");

  for (i=0; i<STORE_HASH_SIZE; i++)
    {
      /* dump contents of all hash table buckets */
      for (ent=store_htable[TRI][i]; ent; ent=ent->next)
	{
	  fprintf(stream, "[0x%08x]: %12.0f/0x%08x:%08x\n",
		  ent->addr, (double)(*((double *)ent->data)),
		  *((unsigned int *)&ent->data[0]),
		  *(((unsigned int *)&ent->data[0]) + 1));
	}
    }
}

/* default memory state accessor, used by DLite */
static char *					/* err str, NULL for no err */
simoo_mem_obj(enum dlite_access_t at,		/* access type */
	      SS_ADDR_TYPE addr,		/* address to access */
	      char *p,				/* input/output buffer */
	      int nbytes)			/* size of access */
{
  char *errstr;
  enum mem_cmd cmd;

  if (at == at_read)
    cmd = Read;
  else if (at == at_write)
    cmd = Write;
  else
    panic("bogus access type");

  errstr = mem_valid(cmd, addr, nbytes, /* declare */FALSE);
  if (errstr)
    return errstr;

  /* else, no error, access memory */
  if (spec_mode[TRI])
    spec_mem_access(cmd, addr, p, nbytes);
  else
    mem_access(cmd, addr, p, nbytes);

  /* no error */
  return NULL;
}


/*
 *  RUU_DISPATCH() - decode instructions and allocate RUU and LSQ resources
 */

/* link RS onto the output chain number of whichever operation will next
   create the architected register value IDEP_NAME */
static INLINE void
ruu_link_idep(struct RUU_station *rs,		/* rs station to link */
	      int idep_num,			/* input dependence number */
	      int idep_name)			/* input register name */
{
  struct CV_link head;
  struct RS_link *link;

  /* any dependence? */
  if (idep_name == NA)
    {
      /* no input dependence for this input slot, mark operand as ready */
      rs->idep_ready[idep_num] = TRUE;
      return;
    }

  /* locate creator of operand */
  head = CREATE_VECTOR(idep_name);

  /* any creator? */
  if (!head.rs)
    {
      /* no active creator, use value available in architected reg file,
         indicate the operand is ready for use */
      rs->idep_ready[idep_num] = TRUE;
      return;
    }
  /* else, creator operation will make this value sometime in the future */

  /* indicate value will be created sometime in the future, i.e., operand
     is not yet ready for use */
  rs->idep_ready[idep_num] = FALSE;

  /* link onto creator's output list of dependant operand */
  RSLINK_NEW(link, rs); link->x.opnum = idep_num;
  link->next = head.rs->odep_list[head.odep_num];
  head.rs->odep_list[head.odep_num] = link;
}

/* make RS the creator of architected register ODEP_NAME */
static INLINE void
ruu_install_odep(struct RUU_station *rs,	/* creating RUU station */
		 int odep_num,			/* output operand number */
		 int odep_name)			/* output register name */
{
  struct CV_link cv;

  /* any dependence? */
  if (odep_name == NA)
    {
      /* no value created */
      rs->onames[odep_num] = NA;
      return;
    }
  /* else, create a RS_NULL terminated output chain in create vector */

  /* record output name, used to update create vector at completion */
  rs->onames[odep_num] = odep_name;

  /* initialize output chain to empty list */
  rs->odep_list[odep_num] = NULL;

  /* indicate this operation is latest creator of ODEP_NAME */
  CVLINK_INIT(cv, rs, odep_num);
  SET_CREATE_VECTOR(odep_name, cv);
}


/*
 * configure the instruction decode engine
 */

/* general register dependence decoders */
#define DGPR(N)			(N)
#define DCGPR(N)		(SS_COMP_OP != SS_COMP_NOP ? (N) : 0)
#define DGPR_D(N)		((N)&~1)

/* floating point register dependence decoders */
#define DFPR_L(N)		(((N)+32)&~1)
#define DFPR_F(N)		(((N)+32)&~1)
#define DFPR_D(N)		(((N)+32)&~1)

/* miscellaneous register dependence decoders */
#define DHI			(0+32+32)
#define DLO			(1+32+32)
#define DFCC			(2+32+32)
#define DTMP			(3+32+32)
#define DNA			(0)


/*
 * configure the execution engine
 */

/* next program counter */
#define SET_NPC(EXPR)           (next_PC = (EXPR))

/* target program counter */
#undef  SET_TPC
#define SET_TPC(EXPR)		(target_PC = (EXPR))

/* current program counter */
#define CPC                     (regs_PC[TRI])
#define SET_CPC(EXPR)           (regs_PC[TRI] = (EXPR))

/* general purpose register accessors, NOTE: speculative copy on write storage
   provided for fast recovery during wrong path execute (see tracer_recover()
   for details on this process */
#define GPR(N)                  (BITMAP_SET_P(use_spec_R[TRI], R_BMAP_SZ, (N))\
				 ? spec_regs_R[TRI][N]                       \
				 : regs_R[TRI][N])
#define SET_GPR(N,EXPR)         (spec_mode[TRI]                              \
				 ? (BITMAP_SET(use_spec_R[TRI], R_BMAP_SZ, (N)),\
				    spec_regs_R[TRI][N] = (EXPR))            \
				 : (regs_R[TRI][N] = (EXPR)))

/* floating point register accessors, NOTE: speculative copy on write storage
   provided for fast recovery during wrong path execute (see tracer_recover()
   for details on this process */
#define FPR_L(N)                (BITMAP_SET_P(use_spec_F[TRI], F_BMAP_SZ, ((N)&~1))\
				 ? spec_regs_F[TRI].l[(N)]                   \
				 : regs_F[TRI].l[(N)])
#define SET_FPR_L(N,EXPR)       (spec_mode[TRI]                              \
				 ? (BITMAP_SET(use_spec_F[TRI],F_BMAP_SZ,((N)&~1)),\
				    spec_regs_F[TRI].l[(N)] = (EXPR))        \
				 : (regs_F[TRI].l[(N)] = (EXPR)))
#define FPR_F(N)                (BITMAP_SET_P(use_spec_F[TRI], F_BMAP_SZ, ((N)&~1))\
				 ? spec_regs_F[TRI].f[(N)]                   \
				 : regs_F[TRI].f[(N)])
#define SET_FPR_F(N,EXPR)       (spec_mode[TRI]                              \
				 ? (BITMAP_SET(use_spec_F[TRI],F_BMAP_SZ,((N)&~1)),\
				    spec_regs_F[TRI].f[(N)] = (EXPR))        \
				 : (regs_F[TRI].f[(N)] = (EXPR)))
#define FPR_D(N)                (BITMAP_SET_P(use_spec_F[TRI], F_BMAP_SZ, ((N)&~1))\
				 ? spec_regs_F[TRI].d[(N) >> 1]              \
				 : regs_F[TRI].d[(N) >> 1])
#define SET_FPR_D(N,EXPR)       (spec_mode[TRI]                              \
				 ? (BITMAP_SET(use_spec_F[TRI],F_BMAP_SZ,((N)&~1)),\
				    spec_regs_F[TRI].d[(N) >> 1] = (EXPR))   \
				 : (regs_F[TRI].d[(N) >> 1] = (EXPR)))

/* miscellanous register accessors, NOTE: speculative copy on write storage
   provided for fast recovery during wrong path execute (see tracer_recover()
   for details on this process */
#define HI                      (use_spec_HI[TRI] ? spec_regs_HI[TRI] : regs_HI[TRI])
#define SET_HI(EXPR)            (spec_mode[TRI]                              \
				 ? (use_spec_HI[TRI]=TRUE,                   \
				    spec_regs_HI[TRI] = (EXPR))              \
				 : (regs_HI[TRI] = (EXPR)))
#define LO                      (use_spec_LO[TRI] ? spec_regs_LO[TRI] : regs_LO[TRI])
#define SET_LO(EXPR)            (spec_mode[TRI]                              \
				 ? (use_spec_LO[TRI]=TRUE,                   \
				    spec_regs_LO[TRI] = (EXPR))              \
				 : (regs_LO[TRI] = (EXPR)))
#define FCC                     (use_spec_FCC[TRI] ? spec_regs_FCC[TRI] : regs_FCC[TRI])
#define SET_FCC(EXPR)           (spec_mode[TRI]                              \
				 ? (use_spec_FCC[TRI]=TRUE,                  \
				    spec_regs_FCC[TRI] = (EXPR))             \
				 : (regs_FCC[TRI] = (EXPR)))

/* precise architected memory state accessor macros, NOTE: speculative copy on
   write storage provided for fast recovery during wrong path execute (see
   tracer_recover() for details on this process */
#define __READ_SPECMEM(DST_T, SRC_N, SRC)				\
  (addr = (SRC),							\
   (spec_mode[TRI]							\
    ? spec_mem_access(Read, addr, SYMCAT(&temp_,SRC_N),			\
		      sizeof(SYMCAT(temp_,SRC_N)))			\
    : mem_access(Read, addr, SYMCAT(&temp_,SRC_N),			\
		 sizeof(SYMCAT(temp_,SRC_N)))),				\
   ((unsigned int)((DST_T)SYMCAT(temp_,SRC_N))))

#define READ_WORD(SRC)							\
  __READ_SPECMEM(unsigned int, uint, (SRC))
#define READ_UNSIGNED_HALF(SRC)						\
  __READ_SPECMEM(unsigned int, ushort, (SRC))
#define READ_SIGNED_HALF(SRC)						\
  __READ_SPECMEM(signed int, short, (SRC))
#define READ_UNSIGNED_BYTE(SRC)						\
  __READ_SPECMEM(unsigned int, uchar, (SRC))
#define READ_SIGNED_BYTE(SRC)						\
  __READ_SPECMEM(signed int, char, (SRC))

#define __WRITE_SPECMEM(SRC, DST_T, DST_N, DST)				\
  (addr = (DST),							\
   SYMCAT(temp_,DST_N) = (DST_T)((unsigned int)(SRC)),			\
   (spec_mode[TRI]							\
    ? spec_mem_access(Write, addr, SYMCAT(&temp_,DST_N),		\
		      sizeof(SYMCAT(temp_,DST_N)))			\
    : mem_access(Write, addr, SYMCAT(&temp_,DST_N),			\
		 sizeof(SYMCAT(temp_,DST_N)))))

#define WRITE_WORD(SRC, DST)						\
  __WRITE_SPECMEM((SRC), unsigned int, uint, (DST))
#define WRITE_HALF(SRC, DST)						\
  __WRITE_SPECMEM((SRC), unsigned short, ushort, (DST))
#define WRITE_BYTE(SRC, DST)						\
  __WRITE_SPECMEM((SRC), unsigned char, uchar, (DST))

/* system call handler macro */
#define SYSCALL(INST)							\
  (/* only execute system calls in non-speculative mode */		\
   (spec_mode[TRI] ? panic("speculative syscall") : (void) 0),		\
   ss_syscall(mem_access, INST))

/* non-expression inst implementations */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,CLASS,O1,O2,I1,I2,I3,EXPR)
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)
#define CONNECT(OP)
#define IMPL
#include "ss.def"
#undef DEFINST
#undef DEFLINK
#undef CONNECT
#undef IMPL


/* default register state accessor, used by DLite */
static char *					/* err str, NULL for no err */
simoo_reg_obj(enum dlite_access_t at,		/* access type */
	      enum dlite_reg_t rt,		/* reg bank to probe */
	      int reg,				/* register number */
	      union dlite_reg_val_t *val)	/* input, output */
{
  if (reg < 0 || reg >= SS_NUM_REGS)
    return "register number out of range";

  if (at == at_read || at == at_write)
    {
      switch (rt)
	{
	case rt_gpr:
	  if (at == at_read)
	    val->as_word = GPR(reg);
	  else
	    SET_GPR(reg, val->as_word);
	  break;
	case rt_lpr:
	  if (at == at_read)
	    val->as_word = FPR_L(reg);
	  else
	    SET_FPR_L(reg, val->as_word);
	  break;
	case rt_fpr:
	  if (at == at_read)
	    val->as_float = FPR_F(reg);
	  else
	    SET_FPR_F(reg, val->as_float);
	  break;
	case rt_dpr:
	  /* 1/2 as many regs in this mode */
	  if (reg < 0 || reg >= SS_NUM_REGS/2)
	    return "register number out of range";

	  if (at == at_read)
	    val->as_double = FPR_D(reg * 2);
	  else
	    SET_FPR_D(reg * 2, val->as_double);
	  break;
	case rt_hi:
	  if (at == at_read)
	    val->as_word = HI;
	  else
	    SET_HI(val->as_word);
	  break;
	case rt_lo:
	  if (at == at_read)
	    val->as_word = LO;
	  else
	    SET_LO(val->as_word);
	  break;
	case rt_FCC:
	  if (at == at_read)
	    val->as_condition = FCC;
	  else
	    SET_FCC(val->as_condition);
	  break;
	case rt_PC:
	  if (at == at_read)
	    val->as_address = regs_PC[TRI];
	  else
	    regs_PC[TRI] = val->as_address;
	  break;
	default:
	  panic("bogus register bank");
	}
    }
  else
    panic("bogus access type");

  /* no error */
  return NULL;
}

/* non-zero for a valid address */
#define VALID_ADDR(ADDR) (SS_DATA_BASE <= (ADDR) && (ADDR) <= SS_STACK_BASE)

/* the last operation that ruu_dispatch() attempted to dispatch, for
   implementing in-order issue */
static struct RS_link last_op[ST_NUM_TU] = {RSLINK_NULL_DATA, RSLINK_NULL_DATA,
					    RSLINK_NULL_DATA, RSLINK_NULL_DATA,
					    RSLINK_NULL_DATA, RSLINK_NULL_DATA,
					    RSLINK_NULL_DATA, RSLINK_NULL_DATA,
					    RSLINK_NULL_DATA, RSLINK_NULL_DATA,
					    RSLINK_NULL_DATA, RSLINK_NULL_DATA,
					    RSLINK_NULL_DATA, RSLINK_NULL_DATA,
					    RSLINK_NULL_DATA, RSLINK_NULL_DATA};
/*
*/
/* dispatch instructions from the IFETCH -> DISPATCH queue: instructions are
   first decoded, then they allocated RUU (and LSQ for load/stores) resources
   and input and output dependence chains are updated accordingly */
static void
ruu_dispatch(void)
{
  int i;
  int n_dispatched;                     /* total insts dispatched */
  SS_INST_TYPE inst;                    /* actual instruction bits */
  enum ss_opcode op;                    /* decoded opcode enum */
  int out1, out2, in1, in2, in3;        /* output/input register names */
  SS_ADDR_TYPE next_PC, target_PC;	/* actual next/target PC address */
  SS_ADDR_TYPE addr;                    /* effective address, if load/store */
  struct RUU_station *rs;               /* RUU station being allocated */
  struct RUU_station *lsq;              /* LSQ station for ld/st's */
  struct bpred_update *dir_update_ptr;	/* branch predictor dir update ptr */
  int stack_recover_idx;		/* bpred retstack recovery index */
  unsigned int pseq;			/* pipetrace sequence number */
  int is_write;				/* store? */
  int made_check;			/* used to ensure DLite entry */
  int br_taken, br_pred_taken;		/* if br, taken?  predicted taken? */
  int fetch_redirected = FALSE;
  unsigned int temp_uint;		/* temp variable for spec mem access */
  signed short temp_short;		/* " ditto " */
  unsigned short temp_ushort;		/* " ditto " */
  signed char temp_char;		/* " ditto " */
  unsigned char temp_uchar;		/* " ditto " */
  
  made_check = FALSE;
  n_dispatched = 0;
  while (/* instruction decode B/W left? */
	 n_dispatched < (ruu_decode_width * fetch_speed)
	 /* RUU and LSQ not full? */
	 && RUU_num[TRI] < RUU_size && LSQ_num[TRI] < LSQ_size
	 /* insts still available from fetch unit? */
	 && fetch_num[TRI] != 0
	 /* on an acceptable trace path */
	 && (ruu_include_spec || !spec_mode[TRI]))
    {
      /* if issuing in-order, block until last op issues if inorder issue */
      if (ruu_inorder_issue
	  && (last_op[TRI].rs && RSLINK_VALID(&last_op[TRI])
	      && !OPERANDS_READY(last_op[TRI].rs)))
	{
	  /* stall until last operation is ready to issue */
	  break;
	}

      /* get the next instruction from the IFETCH -> DISPATCH queue */
      inst = fetch_data[TRI][fetch_head[TRI]].IR;
      regs_PC[TRI] = fetch_data[TRI][fetch_head[TRI]].regs_PC;
      pred_PC[TRI] = fetch_data[TRI][fetch_head[TRI]].pred_PC;
      dir_update_ptr = &(fetch_data[TRI][fetch_head[TRI]].dir_update);
      stack_recover_idx = fetch_data[TRI][fetch_head[TRI]].stack_recover_idx;
      pseq = fetch_data[TRI][fetch_head[TRI]].ptrace_seq;

      /* decode the inst */
      op = SS_OPCODE(inst);

      /* compute default next_PC */
      next_PC = regs_PC[TRI] + sizeof(SS_INST_TYPE);

#ifdef SUPER_TR
      /* save the GPR's *
      save_GPR(); */

      if (spec_mode[TRI] && (SS_OP_FLAGS(op) & F_ST))
	  break;

#endif

      /* drain RUU for TRAPs and system calls */
      if (SS_OP_FLAGS(op) & F_TRAP)
	{
	  if (RUU_num[TRI] != 0)
	    break;

	  /* else, syscall is only instruction in the machine, at this
	     point we should not be in (mis-)speculative mode */
	  if (spec_mode[TRI])
	    panic("drained and speculative");
	}

      /* maintain $r0 semantics (in spec and non-spec space) */
      regs_R[TRI][0] = 0; spec_regs_R[TRI][0] = 0;

      /* default effective address (none) and access */
      addr = 0; is_write = FALSE;

      /* more decoding and execution */
      switch (op)
	{

/* the divide-by-zero instruction check macro is redefined because there
   is no portable and reliable way to hide integer divide-by-zero faults */
#undef  DIV0
#define DIV0(N)		(spec_mode[TRI]					\
			 ? /* do nothing */(void)0			\
			 : (((N) == 0) ? IFAIL("divide by 0") : (void)0))

/* the divide operator semantics are also redefined because there
   are no portable and reliable way to hide integer divide-by-zero faults */
#undef  IDIV
#define IDIV(A, B)	(((B) == 0) ? 0 : ((A) / (B)))

#undef  IMOD
#define IMOD(A, B)	(((B) == 0) ? 0 : ((A) % (B)))

#undef  FDIV
#define FDIV(A, B)	(((B) == 0) ? 0 : ((A) / (B)))

#undef  FINT
#define FINT(A)		(isnan(A) ? 0 : ((int)(A)))

/* decode and execute the instruction */
/* the following macro wraps the instruction check failure declaration
   with a test to see if the trace generator is in non-speculative
   mode, if so the instruction fault is declared, otherwise, the error
   is shunted because instruction faults need to be masked on the
   mis-speculated instruction paths */
#undef  IFAIL
#define IFAIL(S)							\
  (spec_mode[TRI] ? /* ignore */(void)0 : /* declare */(void)fatal(S))

#define DEFINST(OP,MSK,NAME,OPFORM,RES,CLASS,O1,O2,I1,I2,I3,EXPR)	\
	case OP:							\
	  out1 = O1; out2 = O2;						\
	  in1 = I1; in2 = I2; in3 = I3;					\
	  EXPR;								\
	  break;
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT)					\
	case OP:							\
	  /* could speculatively decode a bogus inst */			\
	  op = NOP;							\
	  out1 = NA; out2 = NA;						\
	  in1 = NA; in2 = NA; in3 = NA;					\
	  /* no EXPR */							\
	  break;
#define CONNECT(OP)
#include "ss.def"
#undef DEFINST
#undef DEFLINK
#undef CONNECT
	default:
	  /* can speculatively decode a bogus inst */
	  op = NOP;
	  out1 = NA; out2 = NA;
	  in1 = NA; in2 = NA; in3 = NA;
	  /* no EXPR */
	}

#ifdef SUPER_TR  
      if (TU_BLOCKED /* && op != ESTR */) { /* if lock ws unsuccessfull */
	  /* restore_GPR();  */
	  break;
      } 
#endif
      /* operation sets next_PC */

      /* update memory access stats */
      if (SS_OP_FLAGS(op) & F_MEM)
	{
	  sim_total_refs++;
	  if (!spec_mode[TRI])
	    sim_num_refs++;

	  if (SS_OP_FLAGS(op) & F_STORE)
	    is_write = TRUE;
	  else
	    {
	      sim_total_loads++;
	      if (!spec_mode[TRI])
		sim_num_loads++;
	    }
	}

      br_taken = (next_PC != (regs_PC[TRI] + sizeof(SS_INST_TYPE)));
      br_pred_taken = (pred_PC[TRI] != (regs_PC[TRI] + sizeof(SS_INST_TYPE)));

      if ((pred_PC[TRI] != next_PC && pred_perfect)
	  || ((SS_OP_FLAGS(op) & (F_CTRL|F_DIRJMP)) == (F_CTRL|F_DIRJMP)
	      && target_PC != pred_PC[TRI] && br_pred_taken))
	{
	  /* Either 1) we're simulating perfect prediction and are in a
             mis-predict state and need to patch up, or 2) We're not simulating
             perfect prediction, we've predicted the branch taken, but our
             predicted target doesn't match the computed target (i.e.,
             mis-fetch).  Just update the PC values and do a fetch squash.
             This is just like calling fetch_squash() except we pre-anticipate
             the updates to the fetch values at the end of this function.  If
             case #2, also charge a mispredict penalty for redirecting fetch */
	  fetch_pred_PC[TRI] = fetch_regs_PC[TRI] = next_PC;
	  /* was: if (pred_perfect) */
#if 1
	  if (pred_perfect)
#endif
	    pred_PC[TRI] = next_PC;

	  fetch_head[TRI] = (ruu_ifq_size-1);
	  fetch_num[TRI] = 1;
	  fetch_tail[TRI] = 0;

	  if (!pred_perfect)
	    ruu_fetch_issue_delay[TRI] = ruu_branch_penalty;
#if 0
	  else
	    {
	      /* only 1 cycle bubble */
	      ruu_fetch_issue_delay[TRI] = 1;
	    }
#endif

	  fetch_redirected = TRUE;
	}

      /* is this a NOP */
      if (op != NOP)
	{
	  /* for load/stores:
	       idep #0     - store operand (value that is store'ed)
	       idep #1, #2 - eff addr computation inputs (addr of access)

	     resulting RUU/LSQ operation pair:
	       RUU (effective address computation operation):
		 idep #0, #1 - eff addr computation inputs (addr of access)
	       LSQ (memory access operation):
		 idep #0     - operand input (value that is store'd)
		 idep #1     - eff addr computation result (from RUU op)

	     effective address computation is transfered via the reserved
	     name DTMP
	   */

	  /* fill in RUU reservation station */
	  rs = &RUU[TRI][RUU_tail[TRI]];

	  rs->IR = inst;
	  rs->op = op;
	  rs->PC = regs_PC[TRI];
	  rs->next_PC = next_PC; rs->pred_PC = pred_PC[TRI];
	  rs->in_LSQ = FALSE;
	  rs->ea_comp = FALSE;
	  rs->recover_inst = FALSE;
          rs->dir_update = *dir_update_ptr;
	  rs->stack_recover_idx = stack_recover_idx;
	  rs->spec_mode = spec_mode[TRI];
	  rs->addr = 0;
	  /* rs->tag is already set */
	  rs->seq = ++inst_seq;
	  rs->queued = rs->issued = rs->completed = FALSE;
	  rs->ptrace_seq = pseq;

	  /* split ld/st's into two operations: eff addr comp + mem access */
	  if (SS_OP_FLAGS(op) & F_MEM)
	    {
	      /* convert RUU operation from ld/st to an add (eff addr comp) */
	      rs->op = ADD;
	      rs->ea_comp = TRUE;

	      /* fill in LSQ reservation station */
	      lsq = &LSQ[TRI][LSQ_tail[TRI]];

	      lsq->IR = inst;
	      lsq->op = op;
	      lsq->PC = regs_PC[TRI];
	      lsq->next_PC = next_PC; lsq->pred_PC = pred_PC[TRI];
	      lsq->in_LSQ = TRUE;
	      lsq->ea_comp = FALSE;
	      lsq->recover_inst = FALSE;
	      lsq->dir_update.pdir1 = lsq->dir_update.pdir2 = NULL;
	      lsq->dir_update.pmeta = NULL;
	      lsq->stack_recover_idx = 0;
	      lsq->spec_mode = spec_mode[TRI];
	      lsq->addr = addr;
	      /* lsq->tag is already set */
	      lsq->seq = ++inst_seq;
	      lsq->queued = lsq->issued = lsq->completed = FALSE;
	      lsq->ptrace_seq = ptrace_seq++;

	      /* pipetrace this uop */
	      ptrace_newuop(lsq->ptrace_seq, "internal ld/st", lsq->PC, 0);
	      ptrace_newstage(lsq->ptrace_seq, PST_DISPATCH, 0);

	      /* link eff addr computation onto operand's output chains */
	      ruu_link_idep(rs, /* idep_ready[] index */0, NA);
	      ruu_link_idep(rs, /* idep_ready[] index */1, in2);
	      ruu_link_idep(rs, /* idep_ready[] index */2, in3);

	      /* install output after inputs to prevent self reference */
	      ruu_install_odep(rs, /* odep_list[] index */0, DTMP);
	      ruu_install_odep(rs, /* odep_list[] index */1, NA);

	      /* link memory access onto output chain of eff addr operation */
	      ruu_link_idep(lsq,
			    /* idep_ready[] index */STORE_OP_INDEX/* 0 */,
			    in1);
	      ruu_link_idep(lsq,
			    /* idep_ready[] index */STORE_ADDR_INDEX/* 1 */,
			    DTMP);
	      ruu_link_idep(lsq, /* idep_ready[] index */2, NA);

	      /* install output after inputs to prevent self reference */
	      ruu_install_odep(lsq, /* odep_list[] index */0, out1);
	      ruu_install_odep(lsq, /* odep_list[] index */1, out2);

	      /* install operation in the RUU and LSQ */
	      n_dispatched++;
	      RUU_tail[TRI] = (RUU_tail[TRI] + 1) % RUU_size;
	      RUU_num[TRI]++;
	      LSQ_tail[TRI] = (LSQ_tail[TRI] + 1) % LSQ_size;
	      LSQ_num[TRI]++;

	      if (OPERANDS_READY(rs))
		{
		  /* eff addr computation ready, queue it on ready list */
		  readyq_enqueue(rs);
		}
	      /* issue may continue when the load/store is issued */
	      RSLINK_INIT(last_op[TRI], lsq);

	      /* issue stores only, loads are issued by lsq_refresh() */
	      if (((SS_OP_FLAGS(op) & (F_MEM|F_STORE)) == (F_MEM|F_STORE))
		  && OPERANDS_READY(lsq))
		{
		  /* panic("store immediately ready"); */
		  /* put operation on ready list, ruu_issue() issue it later */
		  readyq_enqueue(lsq);
		}
	    }
	  else /* !(SS_OP_FLAGS(op) & F_MEM) */
	    {
	      /* link onto producing operation */
	      ruu_link_idep(rs, /* idep_ready[] index */0, in1);
	      ruu_link_idep(rs, /* idep_ready[] index */1, in2);
	      ruu_link_idep(rs, /* idep_ready[] index */2, in3);

	      /* install output after inputs to prevent self reference */
	      ruu_install_odep(rs, /* odep_list[] index */0, out1);
	      ruu_install_odep(rs, /* odep_list[] index */1, out2);

	      /* install operation in the RUU */
	      n_dispatched++;
	      RUU_tail[TRI] = (RUU_tail[TRI] + 1) % RUU_size;
	      RUU_num[TRI]++;

	      /* issue op if all its reg operands are ready (no mem input) */
	      if (OPERANDS_READY(rs))
		{
		  /* put operation on ready list, ruu_issue() issue it later */
		  readyq_enqueue(rs);
		  /* issue may continue */
		  last_op[TRI] = RSLINK_NULL;
		}
	      else
		{
		  /* could not issue this inst, stall issue until we can */
		  RSLINK_INIT(last_op[TRI], rs);
		}
	    }
	}
      else
	{
	  /* this is a NOP, no need to update RUU/LSQ state */
	  rs = NULL;
	}

      /* one more instruction executed, speculative or otherwise */
      sim_total_insn++;
      if (SS_OP_FLAGS(op) & F_CTRL)
	sim_total_branches++;

      if (!spec_mode[TRI])
	{
	  /* one more non-speculative instruction executed */
	  sim_num_insn++;
	  NUM_INSN_EXED++;

	  /* if this is a branching instruction update BTB, i.e., only
	     non-speculative state is committed into the BTB */
	  if (SS_OP_FLAGS(op) & F_CTRL)
	    {
	      sim_num_branches++;
	      if (pred && bpred_spec_update[TRI] == spec_ID)
		{
		  bpred_update(pred,
			       /* branch address */regs_PC[TRI],
			       /* actual target address */next_PC,
			       /* taken? */next_PC != (regs_PC[TRI] +
						       sizeof(SS_INST_TYPE)),
			       /* pred taken? */pred_PC[TRI] != (regs_PC[TRI] +
							sizeof(SS_INST_TYPE)),
			       /* correct pred? */pred_PC[TRI] == next_PC,
			       /* opcode */op,
			       /* jump through R31? */(RS) == 31,
			       /* predictor update ptr */&rs->dir_update);
		}
	    }

	  /* is the trace generator trasitioning into mis-speculation mode? */
	  if (pred_PC[TRI] != next_PC && !fetch_redirected)
	    {
	      /* entering mis-speculation mode, indicate this and save PC */
	      spec_mode[TRI] = TRUE;
	      rs->recover_inst = TRUE;
	      recover_PC[TRI] = next_PC;
	    }
	}

      /* entered decode/allocate stage, indicate in pipe trace */
      ptrace_newstage(pseq, PST_DISPATCH,
		      (pred_PC[TRI] != next_PC) ? PEV_MPOCCURED : 0);
      if (op == NOP)
	{
	  /* end of the line */
	  ptrace_endinst(pseq);
	}

      /* update any stats tracked by PC */
      for (i=0; i<pcstat_nelt; i++)
	{
	  SS_COUNTER_TYPE newval;
	  int delta;

	  /* check if any tracked stats changed */
	  newval = STATVAL(pcstat_stats[i]);
	  delta = newval - pcstat_lastvals[i];
	  if (delta != 0)
	    {
	      stat_add_samples(pcstat_sdists[i], regs_PC[TRI], delta);
	      pcstat_lastvals[i] = newval;
	    }
	}

      /* consume instruction from IFETCH -> DISPATCH queue */
      fetch_head[TRI] = (fetch_head[TRI]+1) & (ruu_ifq_size - 1);
      fetch_num[TRI]--;

      /* check for DLite debugger entry condition */
      made_check = TRUE;
      if (dlite_check_break(pred_PC[TRI],
			    is_write ? ACCESS_WRITE : ACCESS_READ,
			    addr, sim_num_insn, sim_cycle))
	dlite_main(regs_PC[TRI], pred_PC[TRI], sim_cycle);

      if (parent_id == UNIT_FREE || TU_BLOCKED)
	  break;
    }

  /* need to enter DLite at least once per cycle */
  if (!made_check)
    {
      if (dlite_check_break(/* no next PC */0,
			    is_write ? ACCESS_WRITE : ACCESS_READ,
			    addr, sim_num_insn, sim_cycle))
	dlite_main(regs_PC[TRI], /* no next PC */0, sim_cycle);
    }
}


/*
 *  RUU_FETCH() - instruction fetch pipeline stage(s)
 */

/* initialize the instruction fetch pipeline stage */
static void
fetch_init(void)
{ 
  int i;
  char *pi;

  if (fetch_data[TRI] == NULL) {
      /* allocate the IFETCH -> DISPATCH instruction queue */
      fetch_data[TRI] =
            (struct fetch_rec *)calloc(ruu_ifq_size, sizeof(struct fetch_rec));

      if (!fetch_data[TRI])
          fatal("out of virtual memory");
  }

  fetch_num[TRI] = 0;
  fetch_tail[TRI] = fetch_head[TRI] = 0;

  pi = (char *) fetch_data[TRI];
  for (i =0 ; i< ruu_ifq_size * sizeof(struct fetch_rec); i++)
	pi[i] = 0;
}

/* dump contents of fetch stage registers and fetch queue */
void
fetch_dump(FILE *stream)			/* output stream */
{
  int num, head;

  fprintf(stream, "** fetch stage state **\n");

  fprintf(stream, "spec_mode: %s\n", spec_mode[TRI] ? "t" : "f");
  fprintf(stream, "pred_PC: 0x%08x, recover_PC: 0x%08x\n",
	  pred_PC[TRI], recover_PC[TRI]);
  fprintf(stream, "fetch_regs_PC: 0x%08x, fetch_pred_PC: 0x%08x\n",
	  fetch_regs_PC[TRI], fetch_pred_PC[TRI]);
  fprintf(stream, "\n");

  fprintf(stream, "** fetch queue contents **\n");
  fprintf(stream, "fetch_num: %d\n", fetch_num[TRI]);
  fprintf(stream, "fetch_head: %d, fetch_tail: %d\n",
	  fetch_head[TRI], fetch_tail[TRI]);

  num = fetch_num[TRI];
  head = fetch_head[TRI];
  while (num)
    {
      fprintf(stream, "idx: %2d: inst: `", head);
      ss_print_insn(fetch_data[TRI][head].IR, fetch_data[TRI][head].regs_PC, stream);
      fprintf(stream, "'\n");
      fprintf(stream, "         regs_PC: 0x%08x, pred_PC: 0x%08x\n",
	      fetch_data[TRI][head].regs_PC, fetch_data[TRI][head].pred_PC);
      head = (head + 1) & (ruu_ifq_size - 1);
      num--;
    }
}

static int last_inst_missed[ST_NUM_TU];
static int last_inst_tmissed[ST_NUM_TU];

/* fetch up as many instruction as one branch prediction and one cache line
   acess will support without overflowing the IFETCH -> DISPATCH QUEUE */
static void
ruu_fetch(void)
{
  int i, lat, tlb_lat, done = FALSE;
  SS_INST_TYPE inst;
  int stack_recover_idx;
  int branch_cnt;

  for (i=0, branch_cnt=0;
       /* fetch up to as many instruction as the DISPATCH stage can decode */
       i < (ruu_decode_width * fetch_speed)
       /* fetch until IFETCH -> DISPATCH queue fills */
       && fetch_num[TRI] < ruu_ifq_size
       /* and no IFETCH blocking condition encountered */
       && !done;
       i++)
    {
      /* fetch an instruction at the next predicted fetch address */
      fetch_regs_PC[TRI] = fetch_pred_PC[TRI];

      /* is this a bogus text address? (can happen on mis-spec path) */
      if (ld_text_base <= fetch_regs_PC[TRI]
	  && fetch_regs_PC[TRI]< (ld_text_base+ld_text_size)
	  && !(fetch_regs_PC[TRI] & (sizeof(SS_INST_TYPE)-1)))
	{
	  /* read instruction from memory */
	  mem_access(Read, fetch_regs_PC[TRI], &inst, sizeof(SS_INST_TYPE));

	  /* address is within program text, read instruction from memory */
	  lat = cache_il1_lat;
	  if (cache_il1)
	    {
	      /* access the I-cache */
	      lat =
		cache_access(cache_il1, Read, fetch_regs_PC[TRI],
			     NULL, sizeof(SS_INST_TYPE), sim_cycle,
			     NULL, NULL);
	      if (lat > cache_il1_lat)
		last_inst_missed[TRI] = TRUE;
	    }

	  if (itlb)
	    {
	      /* access the I-TLB, NOTE: this code will initiate
		 speculative TLB misses */
	      tlb_lat =
		cache_access(itlb, Read, fetch_regs_PC[TRI],
			     NULL, sizeof(SS_INST_TYPE), sim_cycle,
			     NULL, NULL);
	      if (tlb_lat > 1)
		last_inst_tmissed[TRI] = TRUE;

	      /* I-cache/I-TLB accesses occur in parallel */
	      lat = MAX(tlb_lat, lat);
	    }

	  /* I-cache/I-TLB miss? assumes I-cache hit >= I-TLB hit */
	  if (lat != cache_il1_lat)
	    {
	      /* I-cache miss, block fetch until it is resolved */
	      ruu_fetch_issue_delay[TRI] += lat - 1;
	      break;
	    }
	  /* else, I-cache/I-TLB hit */
	}
      else
	{
	  /* fetch PC is bogus, send a NOP down the pipeline */
	  inst = SS_NOP_INST;
	}

      /* have a valid inst, here */

      /* possibly use the BTB target */
      if (pred)
	{
	  /* get the next predicted fetch address; only use branch predictor
	     result for branches (assumes pre-decode bits); NOTE: returned
	     value may be 1 if bpred can only predict a direction */
	  if (SS_OP_FLAGS(SS_OPCODE(inst)) & F_CTRL)
	    fetch_pred_PC[TRI] =
	      bpred_lookup(pred,
			   /* branch address */fetch_regs_PC[TRI],
			   /* target address *//* FIXME: not computed */0,
			   /* opcode */SS_OPCODE(inst),
			   /* jump through r31? */(RS) == 31,
		       /* updt */&(fetch_data[TRI][fetch_tail[TRI]].dir_update),
			   /* RSB index */&stack_recover_idx);
	  else
	    fetch_pred_PC[TRI] = 0;

	  /* valid address returned from branch predictor? */
	  if (!fetch_pred_PC[TRI])
	    {
	      /* no predicted taken target, attempt not taken target */
	      fetch_pred_PC[TRI] = fetch_regs_PC[TRI] + sizeof(SS_INST_TYPE);
	    }
	  else
	    {
	      /* go with target, NOTE: discontinuous fetch, so terminate */
	      branch_cnt++;
	      if (branch_cnt >= fetch_speed)
		done = TRUE;
	    }
	}
      else
	{
	  /* no predictor, just default to predict not taken, and
	     continue fetching instructions linearly */
	  fetch_pred_PC[TRI] = fetch_regs_PC[TRI] + sizeof(SS_INST_TYPE);
	}

      /* commit this instruction to the IFETCH -> DISPATCH queue */
      fetch_data[TRI][fetch_tail[TRI]].IR = inst;
      fetch_data[TRI][fetch_tail[TRI]].regs_PC = fetch_regs_PC[TRI];
      fetch_data[TRI][fetch_tail[TRI]].pred_PC = fetch_pred_PC[TRI];
      fetch_data[TRI][fetch_tail[TRI]].stack_recover_idx = stack_recover_idx;
      fetch_data[TRI][fetch_tail[TRI]].ptrace_seq = ptrace_seq++;

      /* for pipe trace */
      ptrace_newinst(fetch_data[TRI][fetch_tail[TRI]].ptrace_seq,
		     inst, fetch_data[TRI][fetch_tail[TRI]].regs_PC,
		     0);
      ptrace_newstage(fetch_data[TRI][fetch_tail[TRI]].ptrace_seq,
		      PST_IFETCH,
		      ((last_inst_missed[TRI] ? PEV_CACHEMISS : 0)
		       | (last_inst_tmissed[TRI] ? PEV_TLBMISS : 0)));
      last_inst_missed[TRI] = FALSE;
      last_inst_tmissed[TRI] = FALSE;

      /* adjust instruction fetch queue */
      fetch_tail[TRI] = (fetch_tail[TRI] + 1) & (ruu_ifq_size - 1);
      fetch_num[TRI]++;
    }
}

/* default machine state accessor, used by DLite */
static char *					/* err str, NULL for no err */
simoo_mstate_obj(FILE *stream,			/* output stream */
		 char *cmd)			/* optional command string */
{
  if (!cmd || !strcmp(cmd, "help"))
    fprintf(stream,
"mstate commands:\n"
"\n"
"    mstate help   - show all machine-specific commands (this list)\n"
"    mstate stats  - dump all statistical variables\n"
"    mstate res    - dump current functional unit resource states\n"
"    mstate ruu    - dump contents of the register update unit\n"
"    mstate lsq    - dump contents of the load/store queue\n"
"    mstate eventq - dump contents of event queue\n"
"    mstate readyq - dump contents of ready instruction queue\n"
"    mstate cv     - dump contents of the register create vector\n"
"    mstate rspec  - dump contents of speculative regs\n"
"    mstate mspec  - dump contents of speculative memory\n"
"    mstate fetch  - dump contents of fetch stage registers and fetch queue\n"
"\n"
	    );
  else if (!strcmp(cmd, "stats"))
    {
      /* just dump intermediate stats */
      sim_print_stats(stream);
    }
  else if (!strcmp(cmd, "res"))
    {
      /* dump resource state */
      res_dump(fu_pool[TRI], stream);
    }
  else if (!strcmp(cmd, "ruu"))
    {
      /* dump RUU contents */
      ruu_dump(stream);
    }
  else if (!strcmp(cmd, "lsq"))
    {
      /* dump LSQ contents */
      lsq_dump(stream);
    }
  else if (!strcmp(cmd, "eventq"))
    {
      /* dump event queue contents */
      eventq_dump(stream);
    }
  else if (!strcmp(cmd, "readyq"))
    {
      /* dump event queue contents */
      readyq_dump(stream);
    }
  else if (!strcmp(cmd, "cv"))
    {
      /* dump event queue contents */
      cv_dump(stream);
    }
  else if (!strcmp(cmd, "rspec"))
    {
      /* dump event queue contents */
      rspec_dump(stream);
    }
  else if (!strcmp(cmd, "mspec"))
    {
      /* dump event queue contents */
      mspec_dump(stream);
    }
  else if (!strcmp(cmd, "fetch"))
    {
      /* dump event queue contents */
      fetch_dump(stream);
    }
  else
    return "unknown mstate command";

  /* no error */
  return NULL;
}

int save_GPR()
{ 
    int i;

    for (i=0; i < SS_NUM_REGS; i++) {
	saved_regs[i] = regs_R[TRI][i];
	saved_fpr.f[i] = regs_F[TRI].f[i];
    }

    saved_rhi = regs_HI[TRI];
    saved_rlo = regs_LO[TRI];
    saved_fcc = regs_FCC[TRI];
    return 0;
}

int restore_GPR()
{
    int i;

    for (i=0; i < SS_NUM_REGS; i++) {
	regs_R[TRI][i] = saved_regs[i];
	regs_F[TRI].f[i] = saved_fpr.f[i];
    }

    regs_HI[TRI] = saved_rhi;
    regs_LO[TRI] = saved_rlo;
    regs_FCC[TRI] = saved_fcc; 
    return 0;
}

int store_fork_state()
{
    int i;

    for (i=0; i < SS_NUM_REGS; i++) {
	ST_regs_R[i] = regs_R[TRI][i];
	ST_regs_F.f[i] = regs_F[TRI].f[i];
    }

    ST_regs_HI = regs_HI[TRI];
    ST_regs_LO = regs_LO[TRI];
    ST_regs_FCC = regs_FCC[TRI];
}
		
int set_fork_state()
{
    int i;

    for (i=0; i < SS_NUM_REGS; i++) {
	regs_R[TRI][i] = ST_regs_R[i];
	regs_F[TRI].f[i] = ST_regs_F.f[i];
    }

    regs_HI[TRI] = ST_regs_HI;
    regs_LO[TRI] = ST_regs_LO;
    regs_FCC[TRI] = ST_regs_FCC;
}
		

/* start simulation, program loaded, processor precise state initialized */
void
sim_main(void)
{
  int i, j;
  char name[10];

  fprintf(stderr, "sim: ** starting performance simulation **\n");

  /* ignore any floating point exceptions, they may occur on mis-speculated
     execution paths */
  signal(SIGFPE, SIG_IGN);

  /* set up program entry state */
  SET_CPC(ld_prog_entry-sizeof(SS_INST_TYPE));
  fetch_regs_PC[TRI] = regs_PC[TRI];
  fetch_pred_PC[TRI] = ld_prog_entry;
  FORK_LABEL = regs_PC[TRI];

  /* check for DLite debugger entry condition */
  if (dlite_check_break(regs_PC[TRI], /* no access */0, /* addr */0, 0, 0))
    dlite_main(regs_PC[TRI], regs_PC[TRI] + SS_INST_SIZE, sim_cycle);

#ifdef SUPER_TR
  for (i=1; i< ST_NUM_TU; i++) {
    	tu_init(i);
  }

  TRI = 0;
  parent_id = 0;
  start_pc = regs_PC[TRI];
  next_cmu = NULL;
  run();
}

int run (void)
{

  char flag;
  int  i; 
  
#endif

  /* main simulator loop, NOTE: the pipe stages are traverse in reverse order
     to eliminate this/next state synchronization and relaxation problems */
  for (;;)
    {

      if (parent_id == UNIT_FREE) /* Disabled */
	  goto next_unit;

	/* checking for errors */
      prelude();

#ifdef SUPER_TR

	/* local tu unfinished tasks, this may produce new messages 
	   so it has to come after the flush of messages to guarantee
	   the one cycle delay */
      process_task_que();

	/* flush CMU's output buffer to next thread's CMU. 
	   All messages are pushed out one cycle later than produced */
      process_cmu_out_buf();

	/* all messages processed are available next cycle, not now */
	/* process messages from previous threads */
      process_st_msg_q();

      if (ABORT_NOW) {
          thread_abort();
	  goto next_unit;
      }

      if (HFRK_NOW && num_active_threads < ST_NUM_TU && !fork_delay) { 
            HFRK_NOW = FALSE;
 	    st_fork_next();
      }

      if (TU_BLOCKED) 
	  goto check;
#endif

      /* commit entries from RUU/LSQ to architected register file */
      ruu_commit();

      /* service function unit release events */
      ruu_release_fu();

      /* ==> may have ready queue entries carried over from previous cycles */

      /* service result completions, also readies dependent operations */
      /* ==> inserts operations into ready queue --> register deps resolved */
      ruu_writeback();

          if (!bugcompat_mode)
	    {
	  /* try to locate memory operations that are ready to execute */
	  /* ==> inserts operations into ready queue --> mem deps resolved */
	      lsq_refresh();

	  /* issue operations ready to execute from a previous cycle */
	  /* <== drains ready queue <-- ready operations commence execution */
	      ruu_issue();
	    }

          /* decode and dispatch new operations */
          /*==>insert ops w/ no deps or all regs ready --> reg deps resolved */

          ruu_dispatch();

	  /* instructions executed in ruu_dispatch may signal the exit of
	     current thread unit, or block the current unit */
	  if (parent_id == UNIT_FREE)
	      goto next_unit;
	  if (TU_BLOCKED)
	      goto check;

          if (bugcompat_mode)
	    {
	      /* try to locate memory operations that are ready to execute */
	  /* ==> inserts operations into ready queue --> mem deps resolved */
	      lsq_refresh();

	      /* issue operations ready to execute from a previous cycle */
	  /* <== drains ready queue <-- ready operations commence execution */
	      ruu_issue();
	    }

         if (!ruu_fetch_issue_delay[TRI]) {
	    ruu_fetch();
         } else
	    ruu_fetch_issue_delay[TRI]--;

#ifdef SUPER_TR 

check:	

      if (fork_delay && num_active_threads < ST_NUM_TU)
	  fork_delay--;

	/* even head-thread checks cmu, in case blocked incorrectly */
      if (pid != head_thread_id || TU_BLOCKED)	
          check_cmu();

next_unit:
      advance();

#endif
    }
}

#ifdef SUPER_TR

/* write to memory all entries in memory buffer */
int writeback_stage (int i)
{
  int lat, j, tlb_lat, k; 
  static char flag[ST_NUM_TU];
  static int  prev_mb_entry_used;

  block_tu();

  if (!flag[TRI]) {

      if ((int)stage3_start == 0 && (int)stage2_start >0) { 
          stage3_start = sim_cycle - stage2_start;
          stage2_start = 0;
      } else if ((int)stage3_start > 0)
          stage3_start = sim_cycle - stage3_start;

      flag[TRI] = 1;

      if (dump_prof)
	  dump_local_profile (OP_STOP);	
  }

  
   /* wait until non-speculative, all incoming messages are processed
      and outgoing messages are sent.*/
  if (pid != head_thread_id || HFRK_NOW || forwarding
	/* || st_msg_q[st_msg_q_head].st_op != ENTRY_FREE ||
	cmu_out_buf[0].st_op != ENTRY_FREE */) {
	wb_delay++;
        task_enque(WRITE_BACK, i, 0, 0, 0, (unsigned int)sim_cycle+1);
        return 0;
  }

#ifdef DBG_ST
  fprintf (outfile, "Thread %d, performing writeback at %u.\n", pid, 
	  (unsigned int)sim_cycle);
#endif

  if (i == 0) {
      stage4_start = sim_cycle;
      prev_mb_entry_used = mb_entry_used;
  }

  /* We do not clear the message queue here, since it is still necessary
     to process the incoming messages. In case current
     thread release_ts'ed and the previous thread store_ts'ed, but this
     store_ts message is not forwarded yet. Forwarding this store_ts earlier
     will unblock the successor thread earlier for better performance. 
     Although syntaticcally we could stop processing incoming messages, and
     let the successor thread become head thread to automatically recover
     from the blocked state.
  */

  for (k=FALSE; i < mem_buf_size; i++) {

	/* Byte by Byte write back here, since we do not know 
	   how the memory buffer was allocated for. Memory buffer
	   is different from a regular cache in that it is NOT 
	   read-allocated, which means a partial access like a 
	   halfword or a byte does not cause the system to fill in
	   valid data for the unused bytes, hence, writing back the
	   unused bytes will polute the memory. We simulated
	   a write-combine queue, which writes a whole word out
	   by disabling some bytes in it, this is actually true with
	   some of the system buses nowadays. */

	if ((i & 0x3) == 0 && MB_WORD_VALID(shm_mem_buf[i]))
		mb_entry_used++;
	
	/* must be valid and locally written */
        if (MB_DATA_BITS(shm_mem_buf[i]) && 
	    (!MB_TSDVD(shm_mem_buf[i]) /* MB_WRITE does not use TSDVD */ ||
	    (MB_TSDVD(shm_mem_buf[i]) & 0x1) /* written by STTS*/ )) {

	    int addr;

	    addr = MB_TAG(shm_mem_buf[i]);

            MEM_WRITE_BYTE (addr, MB_CONT(shm_mem_buf[i]));
	    k = TRUE;
        }

	if (k && (i & 0x3) == 3) { /* reached word boundary */
	      
	      k = FALSE;
	      if ((j = get_mb_port()) < 0) {
        	  task_enque(WRITE_BACK, i+1, 0, 0, 0, (unsigned int)sim_cycle);
		  return 0;
  	      }

	      if (cache_dl1) {
		  /* commit value to D-cache */
	          lat=cache_access(cache_dl1, Write, shm_mem_buf[i-3].tag,
			     NULL, 4, sim_cycle, NULL, NULL);

	          /* access the D-TLB */
	          tlb_lat =
		      cache_access(dtlb, Read, shm_mem_buf[i-3].tag,
			     NULL, 4, sim_cycle, NULL, NULL);
	          lat = MAX(tlb_lat, lat);

	          mb_port[j].busy = sim_cycle + lat;
	      }
	      
	} /* end reached word boundary */

  } /* end for */

  flag[TRI] = 0;
  if (mb_entry_used - prev_mb_entry_used > max_mb_entry_used)
	max_mb_entry_used = mb_entry_used - prev_mb_entry_used;
  ST_DONE();
  return 0;
}

/* pick the first available shm segment for CMU and MB, the
   predecessor will search all the shm segments for matching successor
   pid when it does allocate_ts or store_ts */
int thread_init()
{
  int j;

	pid = tu_instance;
#ifdef DBG_ST
        fprintf (outfile, "Thread %d activated, parent %d parent_rid %d\n", pid,
		parent_id, parent_rid);
#endif

	FORK_LABEL = start_pc;
	SET_CPC(start_pc - sizeof(SS_INST_TYPE));
  	fetch_regs_PC[TRI] = regs_PC[TRI];
  	fetch_pred_PC[TRI] = regs_PC[TRI] + sizeof (SS_INST_TYPE);

        for (j = 0; j < cmu_size; j++) {
                shm_comm_unit[j].st_op = ENTRY_FREE;
		st_msg_q[j].st_op = ENTRY_FREE;
		cmu_out_buf[j].st_op = ENTRY_FREE;
	}
	NUM_INSN_EXED = 0;
	next_thread = 0;
	next_cmu = NULL;
	TSAG_PASSED = FALSE;
  	WAIT_TSAG_DONE = FALSE;
	FORK_ALLOWED = TRUE;
	ABORT_NOW = FALSE;
	HFRK_NOW = FALSE;
 	In_Critical_Region = FALSE;
 	st_msg_q_head = st_msg_q_tail = cmu_head = 0;
	task_que = NULL;
	block_tu(); 	/* wait until pred_memory_buffer is copied */
	ruu_fetch_issue_delay[TRI] = 0;
	set_conflict = 0;
	data_delay = 0;
	wb_delay = 0;
	life_start = sim_cycle;
	stage1_start = sim_cycle;
	lock_delay = 0;

	/* reset register masks */
        reg_masks[TRI][0] = reg_masks[TRI][1] = reg_masks[TRI][2] = 0;

	/* if parent has passed TSAG_STAGE before forking, no need to wait 
	   for TSAG_DONE flag */
	if (pid != head_thread_id) {
	    NEED_TO_WAIT = !p_tsag_passed;
     	}

	/* No need to clear the memory buffer, since the writing back of
	   the last thread that occupies this unit has already done so */

	/* reset all the ports */
	for (j=0; j<NUM_MB_PORTS; j++)
	   mb_port[j].busy = 0;
	for (j=0; j<NUM_RING_PORTS; j++)
	   ring_port[j].busy = 0;

       /* initialize the simulation engine */
  	fu_pool[TRI] = res_create_pool("fu-pool", fu_config, N_ELT(fu_config));
        tracer_init();
	ruu_init();
	lsq_init();
	fetch_init();
        cv_init();
  	rslink_init(MAX_RS_LINKS);
        eventq_init();
        readyq_init();
	return 0;
}

/* this function does the actual forking procedure */
int st_fork_next() 
{
  int i, j;

    /* only one process can fork at a time */
    /* search for available TU */
    for (i = (resource_index+1 < ST_NUM_TU ? resource_index + 1 : 0), j =0; 
		j < ST_NUM_TU - 1; j++) {
        if (tu_state[i][0]==UNIT_FREE) {
	    child_ri = i;
	    next_cmu = comm_unit[i];
            break;
  	}
	i = (i+1) & (ST_NUM_TU - 1);
    }
   
    tu_state[i][3] =  FORK_LABEL;
    tu_state[i][2] =  resource_index << 16;
    tu_state[i][2] |=  TSAG_PASSED;
    num_active_threads++;
    next_thread = ++tu_instance;
    tu_state[i][0] =  pid; /* mark ownership, activate TU */

#ifdef DBG_TR
    fprintf (outfile, "Thread %d forking. next_thread=%d\n", pid, next_thread);
#endif

    j = TRI;
    TRI = i;
    set_fork_state();
    thread_init();
    TRI = j;
    forwarding = TRUE;
    forward_mbuf(0);
}

/* called when ABORT_FUTURE instruction (abfutr) is executed. */
int kill_successor ()
{
  int i;
#ifdef DBG_ST
  fprintf (outfile, "Thread %d killing successor %d.\n", pid, 
			next_thread);
#endif
  HFRK_NOW = FALSE;
  FORK_ALLOWED = FALSE;

  if (next_thread > 0) {
	/* preemptive recursive kill */
      i = TRI;
      TRI = child_ri;
      thread_abort();
      TRI = i;
  }

  /* no child thread now, no need to continue forwarding */ 
  clear_cmu_out_buf();
  forwarding = FALSE;
  next_thread = 0;
  next_cmu = NULL;
   
  return 0;
}

/* return the ID of the first available memory buffer port, if any */
int get_mb_port()
{
  int i;

  for (i=0; i<st_num_mb_ports; i++)
     if (mb_port[i].busy <= sim_cycle)
        return i;

  return -1;

}

/* return the ID of the first available ring port, if any */
int get_ring_port()
{
  int i;

  for (i=0; i<st_num_ring_ports; i++)
     if (ring_port[i].busy <= sim_cycle)
        return i;

  return -1;
}

/* this is the local buffer of any unprocessed incoming messages */
int process_st_msg_q ()
{
    int i, j, p, cmu_entry_processed;

    cmu_entry_processed = 0;

    /* process messages up to the limit of CU_SPEED, or memory buffer ports */
    while (cmu_entry_processed < st_cmu_speed) {
        i=st_msg_q_head;
        if (st_msg_q[i].st_op != ENTRY_FREE) {
	    cmu_entry_processed++;
            switch ( st_msg_q[i].st_op) {
		case OP_TU_GO: /* start execution now */
#ifdef DBG_TR
			printf ("Thread %d, got TU_GO\n", pid);
#endif
			tu_continue();
			break;
                case OP_ALLOCATE_TS:
                        p = get_mb_port();
                        if (p<0)
                           return 0;
                        mb_port[p].busy = sim_cycle + MB_LATENCY;
			
                        j = st_msg_q[i].nbytes;
                        ALLOCATE_TS ( st_msg_q[i].addr, 
					  st_msg_q[i].disv, j,0);
                        break;
                case OP_STORE_TS:
                        p = get_mb_port();
                        if (p<0)
                           return 0;
                        mb_port[p].busy = sim_cycle + MB_LATENCY;
			
                        j = st_msg_q[i].nbytes;

                        STORE_TS (st_msg_q[i].addr, st_msg_q[i].data,
                                      j, st_msg_q[i].disv, 0);
                        break;
                case OP_TSAG_DONE:
#ifdef DBG_TR
			printf ("Thread %d, got TSAG_DONE\n", pid);
#endif
                        WAIT_TSAG_DONE = FALSE;
                        NEED_TO_WAIT = FALSE;
                        break;
                case OP_RELEASE_TS:
                        p = get_mb_port();
                        if (p<0)
                            return 0;
                        mb_port[p].busy = sim_cycle + MB_LATENCY;
			
                        j=st_msg_q[i].nbytes;
                        RELEASE_TS (st_msg_q[i].addr, 
					st_msg_q[i].disv, j, 0);
                        break;
/* propagation of abort request is not efficient, used abort signal line.
                case OP_ABORT:
                        ABORT_FUTURE();
                        break;
*/
                default:
                        fatal ("Unknown opcode %d in CMU !", st_msg_q[i].st_op);
           }
	   st_msg_q[i].st_op = ENTRY_FREE;
           st_msg_q_head++;
           st_msg_q_head %= cmu_size;
       } else
           break;

   }

   return 0;

}

#endif


int prelude () 
{
      /* RUU/LSQ sanity checks */
      if (RUU_num[TRI] < LSQ_num[TRI])
	panic("RUU_num < LSQ_num");
      if (((RUU_head[TRI] + RUU_num[TRI]) % RUU_size) != RUU_tail[TRI])
	panic("RUU_head/RUU_tail wedged");
      if (((LSQ_head[TRI] + LSQ_num[TRI]) % LSQ_size) != LSQ_tail[TRI])
	panic("LSQ_head/LSQ_tail wedged");

      /* check if pipetracing is still active */
      ptrace_check_active(regs_PC[TRI], sim_num_insn, sim_cycle);

      /* indicate new cycle in pipetrace */
      ptrace_newcycle(sim_cycle);

}

int advance ()
{

      if (TRI == ST_NUM_TU - 1) {
	  sim_cycle ++;
	  if (((unsigned int)sim_cycle & 0x3FFFF) == 1)
		fprintf(stderr, "%u cycles simulated ...\n", 
		(unsigned int)sim_cycle);
      }

      TRI = ((TRI + 1) & (ST_NUM_TU -1));
}
