/*
 * ss.c - simplescalar definition routines
 *
 * This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin as a part of the Multiscalar Research Project.
 *  
 * The tool suite is currently maintained by Doug Burger and Todd M. Austin.
 * 
 * Copyright (C) 1994, 1995, 1996, 1997 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 *
 * $Id: ss.c,v 1.16 1999/02/03 18:13:48 huangj Exp $
 *
 * $Log: ss.c,v $
 * Revision 1.16  1999/02/03  18:13:48  huangj
 * support bstr, lfrk. Backward compatible.
 *
 * Revision 1.15  1998/09/20  22:24:07  huangj
 * *** empty log message ***
 *
 * Revision 1.14  1998/07/13 20:05:48  huangj
 * *** empty log message ***
 *
 * Revision 1.13  1998/03/02  20:01:55  huangj
 * Bug-fixed release
 *
 * Revision 1.12  1998/02/16 07:34:11  huangj
 * *** empty log message ***
 *
 * Revision 1.11  1997/12/18 21:13:11  huangj
 * single process version
 *
 * Revision 1.9  1997/12/01 20:48:22  huangj
 * Multiple thread_abort problem fixed.
 *
 * Revision 1.8  1997/12/01 04:43:06  huangj
 * Constant number of processes for simulation.
 *
 * Revision 1.7  1997/11/26 22:14:35  huangj
 * *** empty log message ***
 *
 * Revision 1.5  1997/11/13 23:34:31  huangj
 * *** empty log message ***
 *
 * Revision 1.4  1997/11/09  23:57:10  huangj
 * RLTS bug fixed. Forward disv should use calling disv +1.
 *
 * Revision 1.3  1997/11/05  17:28:38  huangj
 * lock and unlock support added.
 *
 * Revision 1.2  1997/11/01 23:32:15  huangj
 * Added half-word and byte support for ALTS and RLTS.
 *
 * Revision 1.1  1997/10/12 03:49:05  huangj
 * Initial revision
 *
 * Revision 1.4  1997/09/20 15:56:36  huangj
 * *** empty log message ***
 *
 * Revision 1.3  1997/09/18 23:07:01  huangj
 * *** empty log message ***
 *
 * Revision 1.4  1997/03/11  01:40:38  taustin
 * updated copyrights
 * long/int tweaks made for ALPHA target support
 * supported added for non-GNU C compilers
 * ss_print_insn() can now tolerate bogus insts (for DLite!)
 *
 * Revision 1.3  1997/01/06  16:07:24  taustin
 * comments updated
 * functional unit definitions moved from ss.def
 *
 * Revision 1.1  1996/12/05  18:52:32  taustin
 * Initial revision
 *
 *
 */

#include "misc.h"
#include "ss.h"
#include "regs.h"
#include "memory.h"

#ifdef SUPER_TR
#include <signal.h>
#include <sys/wait.h>
#endif

extern SS_TIME_TYPE sim_cycle;
extern SS_COUNTER_TYPE sim_num_insn;
extern int spec_mode[ST_NUM_TU];

/* INC_DEC expression step tables, they map (operation, size) -> step value,
   and speed up pre/post-incr/desc handling */

/* force a nasty address */
#define XX		0x6bababab

/* before increment */
int ss_fore_tab[/* operand size */8][/* operation */5] = {
             /* NOP   POSTI POSTD  PREI   PRED */
/* byte */    {  0,    0,    0,     1,     -1,  },
/* half */    {  0,    0,    0,     2,     -2,  },
/* invalid */ {  XX,   XX,   XX,    XX,    XX,  },
/* word */    {  0,    0,    0,     4,     -4,  },
/* invalid */ {  XX,   XX,   XX,    XX,    XX,  },
/* invalid */ {  XX,   XX,   XX,    XX,    XX,  },
/* invalid */ {  XX,   XX,   XX,    XX,    XX,  },
/* dword */   {  0,    0,    0,     8,     -8,  },
};

/* after increment */
int ss_aft_tab[/* operand size */8][/* operation */5] = {
             /* NOP   POSTI POSTD  PREI   PRED */
/* byte */    {  0,    1,    -1,    0,     0,   },
/* half */    {  0,    2,    -2,    0,     0,   },
/* invalid */ {  XX,   XX,   XX,    XX,    XX,  },
/* word */    {  0,    4,    -4,    0,     0,   },
/* invalid */ {  XX,   XX,   XX,    XX,    XX,  },
/* invalid */ {  XX,   XX,   XX,    XX,    XX,  },
/* invalid */ {  XX,   XX,   XX,    XX,    XX,  },
/* dword */   {  0,    8,    -8,    0,     0,   },
};

/* lwl/lwr/swl/swr masks */
unsigned int ss_lr_masks[] = {
#ifdef BYTES_BIG_ENDIAN
  0x00000000,
  0x000000ff,
  0x0000ffff,
  0x00ffffff,
  0xffffffff,
#else
  0xffffffff,
  0x00ffffff,
  0x0000ffff,
  0x000000ff,
  0x00000000,
#endif
};

/* LWL/LWR implementation workspace */
SS_ADDR_TYPE ss_lr_temp;

/* temporary variables */
SS_ADDR_TYPE temp_bs, temp_rd;

/* opcode mask -> enum ss_opcodem, used by decoder (SS_OP_ENUM()) */
enum ss_opcode ss_mask2op[SS_MAX_MASK+1];

/* preferred nop instruction definition */
SS_INST_TYPE SS_NOP_INST = { NOP, 0 };

/* intialize the inst decoder, this function builds the ISA decode tables */
void
ss_init_decoder(void)
{
  /* FIXME: CONNECT defined? */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR)	\
  if (ss_mask2op[(MSK)]) fatal("doubly defined mask value");		\
  if ((MSK) > SS_MAX_MASK) fatal("mask value is too large");		\
  ss_mask2op[(MSK)]=(OP);
#include "ss.def"
#undef DEFINST
}

/* enum ss_opcode -> description string */
char *ss_op2name[OP_MAX] = {
  NULL, /* NA */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR) NAME,
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT) NAME,
#define CONNECT(OP)
#include "ss.def"
#undef DEFINST
#undef DEFLINK
#undef CONNECT
};

/* enum ss_opcode -> opcode operand format, used by disassembler */
char *ss_op2format[OP_MAX] = {
  NULL, /* NA */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR) OPFORM,
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT) NULL,
#define CONNECT(OP)
#include "ss.def"
#undef DEFINST
#undef DEFLINK
#undef CONNECT
};

/* enum ss_opcode -> enum ss_fu_class, used by performance simulators */
enum ss_fu_class ss_op2fu[OP_MAX] = {
  FUClass_NA, /* NA */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR) RES,
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT) FUClass_NA,
#define CONNECT(OP)
#include "ss.def"
#undef DEFINST
#undef DEFLINK
#undef CONNECT
};

/* enum ss_opcode -> opcode flags, used by simulators */
unsigned int ss_op2flags[OP_MAX] = {
  NA, /* NA */
#define DEFINST(OP,MSK,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR) FLAGS,
#define DEFLINK(OP,MSK,NAME,MASK,SHIFT) NA,
#define CONNECT(OP)
#include "ss.def"
#undef DEFINST
#undef DEFLINK
#undef CONNECT
};

/* enum ss_fu_class -> description string */
char *ss_fu2name[NUM_FU_CLASSES] = {
  NULL, /* NA */
  "fu-int-ALU",
  "fu-int-multiply",
  "fu-int-divide",
  "fu-FP-add/sub",
  "fu-FP-comparison",
  "fu-FP-conversion",
  "fu-FP-multiply",
  "fu-FP-divide",
  "fu-FP-sqrt",
  "rd-port",
  "wr-port"
};

/* disassemble a SimpleScalar instruction */
void
ss_print_insn(SS_INST_TYPE inst,	/* instruction to disassemble */
	      SS_ADDR_TYPE pc,		/* addr of inst, used for PC-rels */
	      FILE *stream)		/* output stream */
{
  enum ss_opcode op;

  /* use stderr as default output stream */
  if (!stream)
    stream = stderr;

  /* decode the instruction, assumes predecoded text segment */
  op = SS_OPCODE(inst); /* use: SS_OP_ENUM(SS_OPCODE(inst)); if not predec */

  /* disassemble the instruction */
  if (op >= OP_MAX)
    {
      /* bogus instruction */
      fprintf(stream, "<invalid inst: 0x%08x:%08x>", inst.a, inst.b);
    }
  else
    {
      char *s;

      fprintf(stream, "%-10s", SS_OP_NAME(op));

      s = SS_OP_FORMAT(op);
      while (*s) {
	switch (*s) {
	case 'd':
	  fprintf(stream, "r%d", RD);
	  break;
	case 's':
	  fprintf(stream, "r%d", RS);
	  break;
	case 't':
	  fprintf(stream, "r%d", RT);
	  break;
	case 'b':
	  fprintf(stream, "r%d", BS);
	  break;
	case 'D':
	  fprintf(stream, "f%d", FD);
	  break;
	case 'S':
	  fprintf(stream, "f%d", FS);
	  break;
	case 'T':
	  fprintf(stream, "f%d", FT);
	  break;
	case 'j':
	  fprintf(stream, "0x%x", (pc + 8 + (OFS << 2)));
	  break;
	case 'o':
	case 'i':
	  fprintf(stream, "%d", IMM);
	  break;
	case 'H':
	  fprintf(stream, "%d", SHAMT);
	  break;
	case 'u':
	  fprintf(stream, "%u", UIMM);
	  break;
	case 'U':
	  fprintf(stream, "0x%x", UIMM);
	  break;
	case 'J':
	  fprintf(stream, "0x%x", ((pc & 036000000000) | (TARG << 2)));
	  break;
	case 'B':
	  fprintf(stream, "0x%x", BCODE);
	  break;
	case ')':
	  /* handle pre- or post-inc/dec */
	  if (SS_COMP_OP == SS_COMP_NOP)
	    fprintf(stream, ")");
	  else if (SS_COMP_OP == SS_COMP_POST_INC)
	    fprintf(stream, ")+");
	  else if (SS_COMP_OP == SS_COMP_POST_DEC)
	    fprintf(stream, ")-");
	  else if (SS_COMP_OP == SS_COMP_PRE_INC)
	    fprintf(stream, ")^+");
	  else if (SS_COMP_OP == SS_COMP_PRE_DEC)
	    fprintf(stream, ")^-");
	  else if (SS_COMP_OP == SS_COMP_POST_DBL_INC)
	    fprintf(stream, ")++");
	  else if (SS_COMP_OP == SS_COMP_POST_DBL_DEC)
	    fprintf(stream, ")--");
	  else if (SS_COMP_OP == SS_COMP_PRE_DBL_INC)
	    fprintf(stream, ")^++");
	  else if (SS_COMP_OP == SS_COMP_PRE_DBL_DEC)
	    fprintf(stream, ")^--");
	  else
	    panic("bogus SS_COMP_OP");
	  break;
	default:
	  /* anything unrecognized, e.g., '.' is just passed through */
	  fputc(*s, stream);
	}
	s++;
      }
    }
}

#ifdef SUPER_TR

/* local issued commands should have zero as disv */
int ALLOCATE_TS (unsigned int address, DISV_DATA_TYPE disv /*distance vector*/, int nbytes, char code)
{

  int i, j, word_index, byte_index;
  struct MB_ENTRY *mb_entry1, *mb_entry2;

   /* this function allocates a memory buffer entry locally and forwards this
      address to successor threads. All threads should check every cycle
      for arrival of TS address or TS data in the communication unit */

   if (code == FWD_ONLY)
	goto forward;

   word_index =  address & 0xFFFFFFFC;
   byte_index =  address & 0x3;

   mb_entry1 = get_mb_entry(word_index);
   mb_entry2 = NON_NULL; /* a non-zero value */

   if (byte_index + nbytes > 4) /* non-aligned access */
	mb_entry2 = get_mb_entry (word_index + 4);

   if (mb_entry1 == NULL || mb_entry2 == NULL) {
         fprintf (outfile, "Thread %d MB full in alts, blocked\n", pid);
	 if (pid != head_thread_id) {
#ifdef DBG_ST
   fprintf (stderr, "Thread %d MB full in alts, blocked\n", pid);
#endif
	     task_enque(CHECK_HEAD, 0, 0, 0, 0, (int)sim_cycle+1);
	     block_tu();
	  } else
	     tu_continue();
	/* successor may have a full MB, too.
	   it won't hurt to forward, though.*/
	 goto forward;
   }

   /* update tag */
   mb_entry1->tag = word_index;
   mb_entry1->word_valid = TRUE;

   if (mb_entry2 != NON_NULL) {
       mb_entry2->tag = word_index+4;
       mb_entry2->word_valid = TRUE;
   }

   for (i=byte_index; i< byte_index + nbytes; i++) {
	struct MB_ENTRY *pme;

	pme =  (i<4 ? mb_entry1 : mb_entry2);

	/* scanning from MSB to LSB */
        pme[i&3].tsdva |= disv;
       	pme[i&3].addr_valid = 1;
        pme[i&3].tag = word_index + i;
	pme[i&3].access_mode = nbytes;
	pme[i&3].word_valid = 1;

 	/* update alias count for local calls */
	if (disv == 1) 
	    pme[i&3].alias_count++;

    }/* end for */


#ifdef DBG_ST
   fprintf (outfile, "Thread %d ALLOCATE_TS%d,add=%x, disv %d, ac %d, tsdva %x, dvd %x\n",
		pid, nbytes, mb_entry1[byte_index].tag, disv,
		mb_entry1[byte_index].alias_count,
		mb_entry1[byte_index].tsdva, mb_entry1[byte_index].tsdvd);
#endif

forward:

   if (next_thread == 0 || disv >= MAX_DISV)
        return 0;

   i = get_cmu_entry();
   if (i < 0) {
	task_enque (FWD_ALTS, address, disv, nbytes, FWD_ONLY, (int)sim_cycle+1);
	return 0;
   }

#ifdef DBG_ST
   fprintf (outfile, "Thread %d fwd ALTS%d, add=%x, %dth item written\n",
	 pid, nbytes, address, i);
#endif

   cmu_out_buf[i].nbytes = nbytes;
   cmu_out_buf[i].disv = disv << 1;
   cmu_out_buf[i].st_op =  OP_ALLOCATE_TS;
   cmu_out_buf[i].addr = address;

   /* only dump the locally generated message */
   if (dump_prof && disv == 1)
	dump_profile (3, cmu_out_buf[i]);

   return 0;
}

int STORE_TS (unsigned int address, unsigned int data, 
	      int nbytes, DISV_DATA_TYPE disv, char code)
{

   int i, j, word_index, byte_index;
   struct MB_ENTRY *mb_entry1, *mb_entry2;

#ifdef DBG_ST
   fprintf (outfile, "Thread %d STORE_TS%d, addr = %x, data = %x, disv=%d\n", 
		pid, nbytes, address, data,disv);
#endif

   word_index = address & 0xFFFFFFFC;
   byte_index = address & 0x3;

   mb_entry1 = get_mb_entry(word_index);
   mb_entry2 = NON_NULL;
   if (byte_index + nbytes > 4)
	mb_entry2 = get_mb_entry (word_index + 4); 

   if (code == FWD_ONLY)
	goto forward;

   if (mb_entry1 == NULL || mb_entry2 == NULL) {
        fprintf (outfile, "MB full in store_ts, spinlocking \n");
	if (disv == 1)
	    MB_WRITE (address, data, nbytes); 
	else if (pid != head_thread_id) {
#ifdef DBG_ST
   fprintf (outfile, "Thread %d mb full in stts, blocked.\n", pid);
#endif
	    block_tu();
	    task_enque (CHECK_HEAD, 0, 0, 0, 0, (int)sim_cycle + 1);
	}
	goto forward;
   } else {
   	    mb_entry1->tag = word_index;
   	    mb_entry1->word_valid = TRUE;

	    if (mb_entry2 != NON_NULL) {
   	        mb_entry1->tag = word_index+4;
   	        mb_entry1->word_valid = TRUE;
	    }
	
	    for (i=byte_index; i< byte_index + nbytes; i++) {
		struct MB_ENTRY *pme;
	        int actual_index;

		if (i<4) {
		    pme = mb_entry1;
		    actual_index = i;
		} else {
		    pme = mb_entry2;
		    actual_index = i - 4;
		}

		/* clear unnecessary address bits */
		if (pme[actual_index].tsdva > disv) {
		    unsigned int tb, mask;
		    for (tb=disv>>1, mask=0; tb > 0; tb = tb >> 1)
			 mask = mask | tb;
		    pme[actual_index].tsdva = pme[actual_index].tsdva & mask;
		}

                pme[actual_index].tsdva &= (~disv);

	 	/* update alias count for local calls */
		if (disv == 1 && pme[actual_index].alias_count)
		    pme[actual_index].alias_count--;

		     /* obsolete data from predecessor */
		if (pme[actual_index].addr_valid && 
			    pme[actual_index].data_valid && 
				pme[actual_index].tsdvd < disv &&
				    disv > 1)
		         continue;

   		pme[actual_index].tag = word_index+i;
		pme[actual_index].addr_valid = TRUE;
		pme[actual_index].data_valid = TRUE;
		pme[actual_index].tsdvd = disv;
		pme[actual_index].access_mode = nbytes;
		pme[actual_index].word_valid = 1;

			/* data is always right-aligned */
		pme[actual_index].data=
			(data >> ((nbytes-i+byte_index-1)*8)) & 0xFF;
	    }
#ifdef DBG_ST
   fprintf (outfile, "Thread %d STORE_TS%d,addr = %x,ac %d, tsdva %x, dvd %x\n",
		pid, nbytes, mb_entry1[byte_index].tag, 
		mb_entry1[byte_index].alias_count,
		mb_entry1[byte_index].tsdva, mb_entry1[byte_index].tsdvd);
#endif

    } /* MB not full */

forward:
   /* always forward received messages, forward local data only if 
      alias count is zero */
   if (next_thread == 0 || (disv==1 && MB_ALIAS_COUNT(mb_entry1[byte_index]) >0)
	 || disv >= MAX_DISV)
        return 0;

  /* protocol for communication unit :
   (thread status ( includes thread ID, num committed , etc, 16 bytes total),
    ( TSD nbytes (1Byte), TS Distance Vector (1Byte), TS Action code(2 bytes),
      TS address(4 bytes), TS data (8 bytes) ),  .......

   */

   i = get_cmu_entry();
   if (i<0) {
	task_enque (FWD_STTS, address, data, nbytes, disv, (int)sim_cycle + 1);
	return 0;
   }

   cmu_out_buf[i].nbytes = nbytes;
   cmu_out_buf[i].disv =  disv << 1;
   cmu_out_buf[i].st_op =  OP_STORE_TS;
   cmu_out_buf[i].addr = address;
   cmu_out_buf[i].data = data;

#ifdef DBG_ST
   fprintf (outfile,"Thread %d forwarding STTS %x,data %x, %dth entry used\n",
		pid,address,data,i);
#endif

   if (dump_prof && disv == 1)
	dump_profile (3, cmu_out_buf[i]);

   return 0;

}

int RELEASE_TS (unsigned int address, DISV_DATA_TYPE disv, int nbytes,char code)
{

   int i, j, word_index, byte_offset;
   struct MB_ENTRY *mb_entry1, *mb_entry2, *pme;

   word_index = address & 0xFFFFFFFC;
   byte_offset = address & 0x3;
   mb_entry1 = get_mb_entry(word_index);

   if (byte_offset + nbytes > 4)
	mb_entry2 = get_mb_entry(word_index + 4);
   else 
	mb_entry2 = NON_NULL; /* non NULL value */

   if (code == FWD_ONLY)
	goto forward;

   if (disv == 1 && dump_prof) 
	dump_local_profile (OP_RELEASE_TS);

   if (mb_entry1 == NULL || mb_entry2 == NULL || !MB_WORD_VALID(mb_entry1[0]) 
		|| mb_entry1[0].tag != word_index) {
	fprintf(outfile, "Thread %d got %x for %x in RLTS\n", pid,
			mb_entry1[0].tag, address);
	fatal("Releasing non-existing entry!\n");
   }

#ifdef DBG_ST
   fprintf (outfile, "Thread %d RLTS %x, disv = %x, nbytes = %d\n", 
	  pid, address, disv, nbytes);   
#endif

   for (i=byte_offset; i< byte_offset + nbytes; i++) {
	int act_ind;

	if (i < 4) {
	    pme = mb_entry1;
	    act_ind = i;
	} else {
	    pme = mb_entry2;
	    act_ind = i - 4;
	}

        pme[act_ind].tsdva &= ~(disv);

	if (disv == 1 && pme[act_ind].alias_count)
	    pme[act_ind].alias_count --;

	/* freeable byte entry */
   	if (!pme[act_ind].alias_count && 
	    !pme[act_ind].tsdva && !pme[act_ind].data_valid) 
	    pme[act_ind].addr_valid = 0;

   }/* end for */

	/* check for freeable entry */
   if (!pme[0].addr_valid && !pme[1].addr_valid && 
	    !pme[2].addr_valid && !pme[3].addr_valid) {
		pme[0].word_valid = 0;
		free_hash_entry (word_index + (i>4)*4);
   }

forward:
  /* this statement may cause a fault if mb_entry == NULL , I assume this
     is a mistake in program, since a release TS must have a matching
     allocate TS */
  if (next_thread == 0 || (disv==1 && 
	MB_ALIAS_COUNT(mb_entry1[byte_offset]) >0)
	|| disv >= MAX_DISV)
        return 0;

   i = get_cmu_entry();
   if (i<0) {
	task_enque (FWD_RLTS, address, disv, nbytes, FWD_ONLY, 
		(int)sim_cycle + 1);
	return 0;
   }

   if (disv == 1 && !MB_ALIAS_COUNT(mb_entry1[byte_offset])
		 && MB_DATA_BITS(mb_entry1[byte_offset]) 
		&& MB_TSDVD(mb_entry1[byte_offset]) < MAX_DISV) {
       cmu_out_buf[i].nbytes = nbytes;
       cmu_out_buf[i].disv =  2;
       cmu_out_buf[i].st_op = OP_STORE_TS;
       cmu_out_buf[i].data = 0;

       for (j=0; j<nbytes; j++) {
		pme = (byte_offset + j < 4 ? mb_entry1 : mb_entry2);
           	cmu_out_buf[i].data |= 
		    pme[(byte_offset+j) & 3].data << (nbytes - 1 -j)*8;
       }
   } else {
        cmu_out_buf[i].nbytes = nbytes;
        cmu_out_buf[i].disv = disv << 1; 
	cmu_out_buf[i].st_op = OP_RELEASE_TS;
   }

   cmu_out_buf[i].addr = address;

   if (dump_prof && disv == 1)
	dump_profile (3, cmu_out_buf[i]);

   return 0;
}

int dbg_info (int val)
{

  int i;

  fprintf (stderr, "Thread %d reached %d\n", pid, val);

  /* for (i=1; i< SS_NUM_REGS; i++) 
	fprintf (outfile, "Thread %d, R%d = %d, F%d = %.2f, FCC=%d\n",
			pid, i, regs_R[i], i, regs_F.f[i], regs_FCC);
  */

}

/* correspond to the instruction called "hfrk", which forks a new thread */
int hfrk(unsigned int label)
{

  if (label > 1 && FORK_LABEL != label) {
	FORK_LABEL = label;
        store_fork_state();
	region_begin_time = sim_cycle; /* not exact, but close */
	if (dump_reg_stat)
	    seq_reg_stat(); 
  } else if (label == 1)
	fatal("PFRK is not supported in this version, sorry.\n");

  HFRK_NOW = 1;
  fork_delay = HFRK_LATENCY;

  stage2_start = sim_cycle; 
  if ((int)stage1_start > 0 && num_active_threads > 1)
	stage1_start = sim_cycle - stage1_start;
  else
	stage1_start = 0;

  if (dump_prof)
	dump_local_profile (OP_FORK);

  return 0;
}

/* correspond to instruction "glock", which acquires a lock for mutex */
int get_lock(int lid)
{
#ifdef DBG_ST
  fprintf (outfile, "Thread %d asking for lock %d\n", pid, lid);
#endif
  if (*(system_lock + lid) == 0) {
      *(system_lock+lid) = pid;
      In_Critical_Region = TRUE;
      tu_continue();

#ifdef DBG_ST
  fprintf (outfile, "Thread %d got lock %d\n", pid, lid);
#endif

  } else {
#ifdef DBG_ST 
   fprintf (outfile, "Thread %d falied to get lock, blocked.\n", pid);
#endif
      block_tu();
      task_enque(GET_LOCK, lid, 0, 0, 0, (int)sim_cycle + 1);
  }
 
  return 0;
}

/* correspond to instruction "rlock", which releases the lock */
int release_lock(int lid)
{

  if (*(system_lock + lid) != pid) {
        printf ("Thread %d trying to release unowned lock. Owner is %d\n",
                pid, *(system_lock + lid));
        fatal("In release lock");
  }

  /* only one process will do it, no need for semaphore */
  *(system_lock + lid) = FALSE;
  In_Critical_Region = FALSE;

#ifdef DBG_ST
  fprintf (outfile, "Thread %d releasing lock %d\n", pid, lid);
#endif

  return 0;
}

/* correspond to instruction "wtsagd", asks current thread to wait */
int wait_tsag_done()
{

#ifdef DBG_ST
  fprintf (outfile, "Thread %d wait-tsag-done\n", pid);
#endif

  if (NEED_TO_WAIT /* && head_thread_id != pid */)
                WAIT_TSAG_DONE = 1;
  if (dump_prof)
	dump_local_profile (OP_WAIT_TSAGD); 
}

#if 0 /* not used now */
int ABORT_FUTURE ()
{

#ifdef DBG_ST
  fprintf (outfile, "Thread %d received SIGQUIT.\n", pid);
#endif

  ABORT_NOW = TRUE;
  HFRK_NOW = FALSE;
  FORK_ALLOWED = FALSE;
  return 0;
}
#endif

int thread_abort()
{
  int i;
#ifdef DBG_ST
  fprintf (outfile, "Thread %d aborting.\n", pid);
#endif

  kill_successor();

  next_cmu = NULL;
 
  for (i=0; i< mem_buf_size; i++) {
	if ((i & 3) == 0 && MB_WORD_VALID(shm_mem_buf[i]))  
		free_hash_entry(MB_TAG(shm_mem_buf[i])); 
	
        clear_entry(i);
  }
  /* squashed instructions do not count */
  total_insn_squashed += NUM_INSN_EXED;
  sim_num_insn -= NUM_INSN_EXED;
  num_active_threads--;
  parent_id = UNIT_FREE;
  num_squashed++;
  stage1_start = stage2_start = stage3_start = stage4_start = 0;

#ifdef DBG_TR
  fprintf (outfile, "Thread %d aborted.\n", pid);
#endif

  if (dump_prof)
	dump_local_profile (OP_ABORT); 
}

/* finished execution of super-threaded region, non-last threads exit */
int ST_DONE()
{
  int i;
  struct Task_Que_Entry *ptqe;

#ifdef DBG_TR
  fprintf (outfile, "Thread %d, ST_DONE!\n",pid);
#endif

  if (next_thread > 0 && (cmu_out_buf[0].st_op != ENTRY_FREE 
		|| st_msg_q[st_msg_q_head].st_op != ENTRY_FREE)) { 
	task_enque (TR_DONE, 0, 0, 0, 0, sim_cycle + 1);
	return 0;
  }


  if (dump_prof)
	dump_local_profile (TR_DONE); 

  total_set_conflict+= set_conflict;
  total_data_delay += data_delay;
  total_wb_delay += wb_delay;
  total_life += ((int)sim_cycle - life_start);
  total_lock_delay += lock_delay;
  next_cmu = NULL;
  tot_stage1 += (int)(stage1_start); 
  tot_stage2 += (int)(stage2_start); 
  tot_stage3 += (int)(stage3_start); 
  tot_stage4 += (int)(sim_cycle - stage4_start); 
  stage1_start = stage2_start = stage3_start = stage4_start = 0;

  for (i=0; i<mem_buf_size; i++) {
	if ((i & 3) == 0 && MB_WORD_VALID(shm_mem_buf[i])) 
	    free_hash_entry (MB_TAG(shm_mem_buf[i]));
	clear_entry(i);
  }

  /* any unfinished taskes are abandoned */
  while (task_que) {
        ptqe = task_que;
        task_que = task_que->next;
        free_tq_entry (ptqe);
  } 

  /* if this is the last thread, continue pass this point */
  if (next_thread == 0 && head_thread_id == pid) {
	TSAG_PASSED = FALSE;
	NEED_TO_WAIT = FALSE;
	WAIT_TSAG_DONE = FALSE;
  	FORK_ALLOWED = TRUE;
	tu_continue();
	FORK_LABEL = 0;    /* reset for the next region */
	accounted_time += (unsigned int)(sim_cycle - region_begin_time);

	/* clear the message buffers */
        for (i = 0; i < cmu_size; i++) {
                shm_comm_unit[i].st_op = ENTRY_FREE;
                st_msg_q[i].st_op = ENTRY_FREE;
                cmu_out_buf[i].st_op = ENTRY_FREE;
        }

  /* obtain register defining statistics */
      if (dump_reg_stat)
          par_reg_stat(); 

#ifdef DBG_ST
  fprintf(outfile, "End of ST_REGION.\n");
#endif

        return 0;
  }

#ifdef DBG_TR
  fprintf (outfile, "Thread %d done! next_thread=%d\n", pid, next_thread);
#endif

  head_thread_id = next_thread;
  /* go to sleep */
  parent_id = UNIT_FREE;
  num_active_threads--;
  next_thread = 0;
  return 0;
}

/* sends successor TSAG_DONE flag */
int TSAG_DONE ()
{

  int i;

  TSAG_PASSED = 1;
  stage3_start = sim_cycle;
  if ((int)stage2_start > 0)
      stage2_start = sim_cycle - stage2_start;

#ifdef DBG_ST
  fprintf (outfile, "Thread %d TSAG_DONE, next_thread = %d.\n", pid, next_thread
);
#endif

  if (dump_prof)
	dump_local_profile (OP_TSAG_DONE);
  
  if (next_thread == 0)
        return 0;

  i = get_cmu_entry();
  if (i < 0) {
	task_enque (FWD_TSAGD, 0, 0, 0, 0, (int)sim_cycle + 1);
  	return 0;
  }

  cmu_out_buf[i].st_op = OP_TSAG_DONE;

   if (dump_prof)
	dump_profile (3, cmu_out_buf[i]);

  return 0;
}

unsigned char __MB_READ_BYTE (unsigned int addr)
{

  struct MB_ENTRY *mb_entry;

  mb_entry = get_mb_entry (addr);

  return mb_entry[addr & 0x3].data;

}

unsigned short __MB_READ_HALF (unsigned int addr)
{

  struct MB_ENTRY *mb_entry1, *mb_entry2;
  int byte_ofs;
  unsigned short output;

  mb_entry1 = get_mb_entry (addr);
  byte_ofs = addr & 0x3;

  if (byte_ofs + 2 > 4) /* cros word boundary */
      mb_entry2 = get_mb_entry(addr - byte_ofs + 4);
  else 
	mb_entry2 = NON_NULL;

  output = (short) (mb_entry1[byte_ofs].data << 8); 

  if (mb_entry2 != NON_NULL)
	output |= mb_entry2[0].data;
  else
  	output |= mb_entry1[byte_ofs+ 1].data;

  return output;
}

unsigned int __MB_READ_WORD (unsigned int addr)
{
  struct MB_ENTRY *mb_entry1, *mb_entry2, *pme;
  int byte_ofs, i;
  unsigned int output;

  mb_entry1 = get_mb_entry (addr);
  byte_ofs = addr & 0x3;
  output = 0;

  if (byte_ofs + 4 > 4)
      mb_entry2 = get_mb_entry ( addr - byte_ofs + 4);
  else
      mb_entry2 = NON_NULL;

  for (i=0; i<4; i++) {
	pme = (byte_ofs + i < 4 ? mb_entry1 : mb_entry2);
	output |= (unsigned int)(pme[(byte_ofs + i) & 3].data << ((3-i) * 8));
  }

#ifdef DBG_ST
  fprintf (outfile, "Thread %d MBread_w %x content %x\n",
                pid,addr, output);
#endif

  return output;

}

int MB_WRITE (unsigned int addr, unsigned int a, int n)
{

  int i, p, word_index, byte_ofs;
  struct MB_ENTRY *mb_entry1, *mb_entry2, *pme;
  static flag;

  mb_entry1 = get_mb_entry (addr);
  word_index = addr & 0xFFFFFFFC;
  byte_ofs = addr & 0x3;
  if (byte_ofs + n > 4) 
	mb_entry2 = get_mb_entry(word_index + 4);
  else
	mb_entry2 = NON_NULL;

  if (mb_entry1 == NULL || mb_entry2 == NULL) { /* memory buffer full */
	if (!flag) {
                fprintf (outfile, "MB_WRITE, mb full.\n");
		flag =1 ;
	}
#ifdef DBG_ST
                fprintf (outfile, "MB_WRITE, mb full.\n");
#endif
        if (head_thread_id != pid) {
		block_tu();
		task_enque(MBUF_WRITE, addr, a, n, 0, (int)sim_cycle + 1);
		return 0;
        }

        switch (n) {
            case 1:
                __MEM_WRITE_BYTE(addr, (unsigned char)a);
                break;
            case 2:
                __MEM_WRITE_HALF(addr, (unsigned short)a);
                break;
            case 4:
                __MEM_WRITE_WORD(addr, (unsigned int)a);
                break;
        }

	tu_continue();
        return 0;
  }

#ifdef DBG_ST
        fprintf (outfile, "Thread %d MB_WRITE, addr %x data %x.\n", pid,addr,a);
#endif

/*  mb_port[p].busy = sim_cycle + MB_LATENCY; */

  mb_entry1->word_valid = TRUE;
  mb_entry1->tag = word_index;

  if (mb_entry2 != NON_NULL) {
      mb_entry2->word_valid = TRUE;
      mb_entry2->tag = word_index+4;
  }

  /* do not change the alias count and tsdva */
  for (i=byte_ofs; i < n + byte_ofs; i++) {
      pme = (i < 4 ? mb_entry1 : mb_entry2);
      pme[i&3].addr_valid = TRUE;
      pme[i&3].data_valid = TRUE;
      pme[i&3].tag = word_index + i;
      pme[i&3].tsdvd = 0; /* locally buffered write, not STTS */ 

/*    local non-stts writes to allocate-ts'ed entry is allowed. 
      But it does not change the dependence infor. Any local 
      writes that overwrites all previous writes would need to 
      be done by explicit stts. */

      pme[i&3].data = (a >> ((n - i + byte_ofs-1)*8)) & 0xFF;
      pme[i&3].access_mode = n;
      pme[i&3].word_valid = 1;
  }

  return 0;

}

int bstr()
{
  int i, j, n;

  store_fork_state();

  /* assumes the starting point for the child threads is the
     next instruction. Although it shows no benefit in the current
     version of the simulator, future optimizations including
     early initialization of the local register file, and early
     start of instruction fetch will improve the performance */

  FORK_LABEL = regs_PC[TRI] + sizeof (SS_INST_TYPE);
  region_begin_time = sim_cycle;

  if (dump_reg_stat)
      seq_reg_stat();

  if (dump_prof)
      dump_local_profile (OP_BEGIN_TR);
}

void seq_reg_stat () 
{
  int i, j, n;

  fprintf (profile_fp, "DEF in SEQ: ");
  for (i=0, n=0; i<3; i++) {
      for (j=0; j<32; j++)
	 if ((reg_masks[TRI][i] >> j) & 1) {
		n++;
		if (dump_prof)
		    fprintf (profile_fp, "%d, ", i*32+j);
	 }
  }

  if (dump_prof)
      fprintf (profile_fp, "\n");

  if (n > max_seq_reg_def)
	max_seq_reg_def = n;
  if (n < min_seq_reg_def)
	min_seq_reg_def = n;
  
  total_seq_reg_def += n;
  total_seq_region++;

  /* reset the masks */
  reg_masks[TRI][0] = reg_masks[TRI][1] = reg_masks[TRI][2] = 0;
} 

void par_reg_stat () 
{
  int i, j, n;

  fprintf (profile_fp, "DEF in PAR: ");
  for (i=0, n=0; i<3; i++) {
      for (j=0; j<32; j++)
	 if ((reg_masks[TRI][i] >> j) & 1) {
		n++;
		if (dump_prof)
		fprintf (profile_fp, "%d, ", i*32+j);
	 }
  }

  if (dump_prof)
      fprintf (profile_fp, "\n");
  if (n > max_par_reg_def)
	max_par_reg_def = n;
  if (n < min_par_reg_def)
	min_par_reg_def = n;
  
  total_par_reg_def += n;
  total_par_region++;

  /* reset the masks */
  reg_masks[TRI][0] = reg_masks[TRI][1] = reg_masks[TRI][2] = 0;

}

#endif
