#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include "regs.h"
#include "misc.h"
#include "memory.h"

#define FALSE 0
extern void kill_proc();
extern SS_TIME_TYPE sim_cycle;

struct Task_Que_Entry *free_tq_list;

/* initialize the queue that stores unfinished tasks of current thread */
int task_queue_init()
{
  int i;

  free_tq_list = (struct Task_Que_Entry *) malloc(sizeof(struct Task_Que_Entry)
                * TASK_Q_SIZE);

  for (i=0; i< TASK_Q_SIZE - 1; i++) {
      free_tq_list[i].next = &free_tq_list[i+1];
  }

  free_tq_list[i].next = NULL;

  return 0;
}

struct Task_Que_Entry * new_tq_entry()
{
  struct Task_Que_Entry *ptqe;

  if (free_tq_list == NULL)
        task_queue_init();

  ptqe = free_tq_list;

  free_tq_list = free_tq_list->next;
  return ptqe;
}

int free_tq_entry (struct Task_Que_Entry *tqe)
{
  struct Task_Que_Entry *ptqe;

  if (free_tq_list == NULL)
	task_queue_init();

  for (ptqe = free_tq_list; ptqe->next != NULL; ptqe=ptqe->next);
  ptqe->next = tqe;
  tqe->next = NULL;
  return 0;

}

int task_enque(char func,unsigned int arg1,unsigned int arg2,
                unsigned int arg3, unsigned int arg4, unsigned int t)
{
  struct Task_Que_Entry *ptqe, *temp;

  if (func > 16 || func < 5)
	fatal ("Unknown task code\n");

  ptqe = new_tq_entry ();

  if (task_que == NULL) {
        task_que = ptqe;
  } else {

        for (temp = task_que; temp->next != NULL; temp = temp->next);
        temp->next = ptqe;
  }

  ptqe->next = NULL;
  ptqe->func = func;
  ptqe->arg1 = arg1;
  ptqe->arg2 = arg2;
  ptqe->arg3 = arg3;
  ptqe->arg4 = arg4;
  ptqe->t    = t;
  return 0;
}

int process_task_que ()
{

  struct Task_Que_Entry *ptqe, *ptqe2;

  ptqe = task_que;
  task_que = NULL;

  if (ptqe == NULL)
        return 0;

  for (; ptqe != NULL; ) {
      switch (ptqe->func) {
	  case FWD_MB:
		forward_mbuf (ptqe->arg1);
		break;
	  case FWD_ALTS:
	 	ALLOCATE_TS (ptqe->arg1, ptqe->arg2, ptqe->arg3, ptqe->arg4);
		break;
	  case FWD_RLTS:
	  	RELEASE_TS (ptqe->arg1, ptqe->arg2, ptqe->arg3, ptqe->arg4);
		break;
	  case FWD_STTS:
		STORE_TS (ptqe->arg1, ptqe->arg2, ptqe->arg3, ptqe->arg4, 
			FWD_ONLY);
		break;
	  case FWD_TSAGD:
		TSAG_DONE ();
		break;
          case WRITE_BACK:
                writeback_stage(ptqe->arg1);
                break;
          case GET_LOCK:
		lock_delay++;
                if (*(system_lock + ptqe->arg1) == FALSE)
		    tu_continue();
		else
		    task_enque(GET_LOCK, ptqe->arg1, 0,0,0, (int)sim_cycle+1);
                break;
          case CHECK_HEAD:
                if (head_thread_id != pid) {
		    block_tu ();
		    task_enque(CHECK_HEAD, 0, 0,0,0, (int)sim_cycle+1);
		} else
		    tu_continue();
                break;
          case MBUF_WRITE:
                MB_WRITE (ptqe->arg1, ptqe->arg2, ptqe->arg3);
                break;
          case CHECK_WTSAGD:
		data_delay++;
		if (!WAIT_TSAG_DONE)
		    tu_continue();
		else 
		    task_enque(CHECK_WTSAGD, 0, 0, 0, 0, (int)sim_cycle + 1);
                break;
	  case CHECK_MB:
		data_delay++;
		if (in_mb (ptqe->arg1) != ADDR_VALID)
		    tu_continue();
		else 
		    task_enque(CHECK_MB, ptqe->arg1, 0,0,0, (int)sim_cycle + 1);
		break;
	  case TR_DONE:
		ST_DONE();
		break;
          default:
                fatal("Unknown task in task queue!\n");
      }/* end switch */
     ptqe2 = ptqe;
     ptqe = ptqe->next;
     free_tq_entry(ptqe2);
  }
  return 0;
}

/* check the CMU for incoming messages, and put it in the local buffer */
int check_cmu ()
{

  int i, j, k;
  struct CMU_ENTRY *msgq, *cmu;

  msgq = st_msg_q;

  /* is the queue full ? */
  if (st_msg_q_tail == st_msg_q_head && msgq[st_msg_q_head].st_op != ENTRY_FREE)
      return 0;

  cmu = shm_comm_unit;

  for (k=0, i=cmu_head; k<cmu_size; k++) {

  	j = st_msg_q_tail;

        if ( cmu[i].st_op == ENTRY_FREE /* empty cmu */ ||
	     cmu[i].arrival_time > sim_cycle /* immature */ ||
	     msgq[j].st_op != ENTRY_FREE /* full msg que */) {
                break;
	}

	if (dump_prof) /* this queue's ID is 1 */
	    dump_profile (1, cmu[i]);

        msgq[j].nbytes = cmu[i].nbytes;
        msgq[j].addr = cmu[i].addr;
        msgq[j].data = cmu[i].data;
        msgq[j].disv = cmu[i].disv;
        msgq[j].st_op = cmu[i].st_op;
        cmu[i].st_op = ENTRY_FREE;

        st_msg_q_tail++;
        st_msg_q_tail %= cmu_size;
	i++;
	if (i>=cmu_size)
	    i = 0;
  }

  cmu_head = i;
  return 0;
}

int prev_val;

/* to avoid deadlock, the "all or none" scheme is used to acquire ports */
int forward_mbuf (int i)
{

  int  j, k, p, q;
  DISV_DATA_TYPE TU_MASK;
  struct MB_ENTRY *mbuf;
  static int access_mode, alts_fwd_done;

  if (next_thread <= 0) {/* is next thread killed ? */
	forwarding = FALSE;
  	return 0;
  }

  forward_delay++;
  TU_MASK = (1 << (ST_NUM_TU - 1)) - 1;
  mbuf = shm_mem_buf;

  for (; i< mem_buf_size; ) {

        if (MB_ADDR_BIT (mbuf[i])){/* addr valid, a candidate */

	    /* due to non-aligned accesses, we have to forward in bytes */
            /* access_mode = mbuf[i].access_mode; */
	    access_mode = ONE_BYTE;
            /* Two cases:
                   1) Still wait for predecessor's data
                   2) uptodate data present, and not written back yet
		   both may be true at the same time 
            */
            if (!alts_fwd_done && MB_TSDVA(mbuf[i]) && 
		(MB_TSDVA(mbuf[i]) & TU_MASK))  {

		alts_fwd_done = 0;

	    	p = get_mb_port();
		if (p < 0) {
		    task_enque (FWD_MB, i,0,0,0, (int)sim_cycle + 1);
		    return 0;
		}

		/* forward as allocate_TS */
		j = get_cmu_entry();
		if (j < 0) {
		    task_enque (FWD_MB, i,0,0,0, (int)sim_cycle + 1);
		    return 0;
		}	

		alts_fwd_done = 1;
		mb_port[p].busy = sim_cycle + MB_LATENCY;
		cmu_out_buf[j].disv = (mbuf[i].tsdva & TU_MASK) << 1;
		cmu_out_buf[j].st_op = OP_ALLOCATE_TS;
		cmu_out_buf[j].addr = mbuf[i].tag;
	   	cmu_out_buf[j].nbytes = access_mode;
		if (dump_prof)
		    dump_profile (2, cmu_out_buf[j]);
	    } 
		/* may need to forward as both ALTS and STTS */
   	    if (MB_DATA_BITS(mbuf[i]) && (MB_TSDVD(mbuf[i]) & TU_MASK)) {

		/* forward as STORE_TS */
		p = get_mb_port();
		if (p < 0) {
		        task_enque (FWD_MB, i,0,0,0, (int)sim_cycle + 1);
		        return 0;
	        }

		j = get_cmu_entry();
		if (j < 0) {
		    task_enque (FWD_MB, i,0,0,0, (int)sim_cycle + 1);
		    return 0;
		}	

		mb_port[p].busy = sim_cycle + MB_LATENCY;
		cmu_out_buf[j].disv = (mbuf[i].tsdvd & TU_MASK) << 1;
		cmu_out_buf[j].st_op = OP_STORE_TS;
		cmu_out_buf[j].addr = mbuf[i].tag;
		cmu_out_buf[j].nbytes = access_mode;

		switch (access_mode) {
		    case FULL_WORD:
			cmu_out_buf[j].data = (mbuf[i].data << 24) | 
				(mbuf[i+1].data << 16) | 
				(mbuf[i+2].data << 8) | mbuf[i+3].data;
			if (i == 0) {
			    if (prev_val - cmu_out_buf[j].data != 3)
			    prev_val = cmu_out_buf[j].data;
			}
			break;
		    case HALF_WORD:
			cmu_out_buf[j].data = (mbuf[i].data << 8) | 
					      mbuf[i+1].data; 
			break;
		    case ONE_BYTE:
			cmu_out_buf[j].data = mbuf[i].data;
			break;
		}

		if (dump_prof)
		    dump_profile (2, cmu_out_buf[j]);
	    }
	} else 
	    access_mode = ONE_BYTE; /* entry is not used,default is byte mode */

	alts_fwd_done = 0;
	i += access_mode;

  } /* end for */

  j = get_cmu_entry();
  if (j<0) {
      task_enque (FWD_MB, i,0,0,0, (int)sim_cycle + 1);
      return 0;
  }	

  cmu_out_buf[j].st_op = OP_TU_GO;
  if (dump_prof)
      dump_profile (3, cmu_out_buf[j]);

  return 0;
}

/* allocate an entry in the local memory buffer */
struct MB_ENTRY * get_mb_entry (unsigned int addr)
{

 int index, i, hind;
 struct MB_HTBL_ENTRY *phte;

 addr = addr & (~0x3);

 hind = (addr >> 2) % MB_HT_SIZE;
 if (mb_htable[hind].addr != 0)  { /* possible existing entry */
	for (phte = &mb_htable[hind]; phte != NULL; phte = phte->next) {
		if (phte->addr == addr) {
			if (shm_mem_buf[phte->index].tag != addr)
				fatal ("Poluted hash table\n");
			return 	shm_mem_buf + phte->index;
		}
	}
 }

 index = addr % num_mb_sets;
 index = index * mb_set_size;

 for (i=0; i< mb_set_size; i+=MB_ENTRY_SIZE) {
      if (!MB_WORD_VALID(shm_mem_buf[index + i])) {
	 shm_mem_buf[index+i].word_valid = 1;
	 shm_mem_buf[index+i].tag = addr;
	 hash_entry(addr, index + i);
	 if (i>0)
	    set_conflict++;
         return shm_mem_buf + index + i;
      }
 }

 return NULL;

}

/* is the wanted address in the memory buffer? If yes, what is the required
   action? */
int in_mb(unsigned int addr)
{
  int j, k, f;
  unsigned int word_index, byte_offset;
  struct MB_ENTRY *mb_entry;
  struct MB_HTBL_ENTRY *phte;

  word_index = addr & 0xFFFFFFFC;
  byte_offset = addr & 0x3;

  j = (word_index >> 2) % MB_HT_SIZE;

  if (mb_htable[j].addr == 0) {
	return FALSE; 
  }

  for (phte = & mb_htable[j]; phte != NULL; phte = phte->next) {
	if (phte->addr == word_index)
		break;
  }

  if (phte == NULL) {
	return FALSE;
  }

  mb_entry =  shm_mem_buf + phte->index;

  if (!MB_ADDR_BIT(mb_entry[byte_offset])) {
      return FALSE;
  }
          /* for read, we only need compre the MSB */
  if (MB_DATA_BITS(mb_entry[byte_offset])) {

	 /* locally buffered write, not STTS */
	  if (!MB_TSDVD(mb_entry[byte_offset]))
		return DATA_VALID;

          for ( j=MB_TSDVA(mb_entry[byte_offset]), f=1,
                k=MB_TSDVD(mb_entry[byte_offset]);
                k>0; j=j>>1, k=k>>1) {

                if ((k & 1) == 1) {
                    return DATA_VALID;
		}
                if ((j & 1) == 1 && f!=1) 
		  	goto head_id;
                f++;
          }

          /* fatal ("Data valid but tsdvd not set."); */

  } else if (MB_TSDVA(mb_entry[byte_offset]) == 1) {
        return DONT_WAIT;
  }

head_id:
  if (pid == head_thread_id) {/* somehow previous thread forgot to
			         store or release it */
	mb_entry[byte_offset].addr_valid = 0;
	if (!mb_entry[0].addr_valid && !mb_entry[1].addr_valid &&
		!mb_entry[2].addr_valid &&!mb_entry[3].addr_valid ) {
	    mb_entry[0].word_valid = 0;
	    free_hash_entry(word_index);
	}
	return DONT_WAIT;
  }
  return ADDR_VALID;

}

int st_init()
{

  int i, j;

  /* macro-variable initialization */
  num_active_threads = 1; 	/* number of active threads */
  head_thread_id = 1; 		/* head thread pid */
  num_squashed = 0;		/* total number of squashed threads */
  mb_entry_used = 0;		/* total memory buffer entries used */
  total_insn_squashed = 0;	
  total_wb_delay = 0;		/* total waiting time to write back */
  total_data_delay = 0;		/* total waiting time for data */
  total_set_conflict = 0;	/* total memory buffer set conflicts */
  tu_instance = 1;		/* number of threads initiated */
  total_lock_delay = 0;		/* total delay waiting to acquire lock */
  total_life = 0;		/* total thread life time */
  min_seq_reg_def = 100;	/* set them out of bound for init value */
  min_par_reg_def = 100;


  st_msg_q_head = st_msg_q_tail = 0;/* message queue head and tail pointer */

  for (i=0; i< 32; i++) 	/* 32 system wide locks */
      system_lock[i] = FALSE;

				/* accounting for stage length */
  tot_stage1 = tot_stage2 = tot_stage3 = tot_stage4 = 0;
  
  /* global variable init */
  HFRK_LATENCY = 4;
  st_num_mb_ports = NUM_MB_PORTS;
  st_num_ring_ports = NUM_RING_PORTS;
  st_ring_delay = RING_LATENCY;
  st_cmu_speed = CMU_SPEED;

  outfile = stdout;
  mem_buf_size = MEM_BUF_SIZE;
  MB_ASSOC = MEM_BUF_SIZE / MB_ENTRY_SIZE; 	/* fully mapped */
  cmu_size = MAX_CMU_ENTRY;
  dump_prof = 0; 	 /* default is not to dump profile info */
  dump_reg_stat = 0; 	 /* default is not to dump reg update info */

  /* reading configuration file */
  { FILE *fptr;
    char line[80], param[20];

    fptr = fopen ("simca.config", "r");
    if (fptr != NULL) {
	while (fgets(line, 79, fptr) != NULL) {
	    sscanf(line, "%s", param);
	    if (strcmp ("MB_PORTS:", param) == 0) 
		 sscanf (line, "%s %d", param, &st_num_mb_ports); 
	    if (strcmp ("RING_PORTS:", param) == 0) 
		 sscanf (line, "%s %d", param, &st_num_ring_ports); 
	    if (strcmp ("MB_ASSOC:", param) == 0) 
		 sscanf (line, "%s %d", param, &MB_ASSOC); 
	    if (strcmp ("FORK_DELAY:", param) == 0) 
		 sscanf (line, "%s %d", param, &HFRK_LATENCY); 
	    if (strcmp ("MB_SIZE:", param) == 0) 
		 sscanf (line, "%s %d", param, &mem_buf_size); 
	    if (strcmp ("CU_SIZE:", param) == 0) 
		 sscanf (line, "%s %d", param, &cmu_size); 
	    if (strcmp ("CU_SPEED:", param) == 0) 
		 sscanf (line, "%s %d", param, &st_cmu_speed); 
	    if (strcmp ("RING_DELAY:", param) == 0) 
		 sscanf (line, "%s %d", param, &st_ring_delay); 
	    if (strcmp ("DUMP_PROF:", param) == 0) 
		 sscanf (line, "%s %d", param, &dump_prof); 
	    if (strcmp ("DUMP_REG_STAT:", param) == 0) 
		 sscanf (line, "%s %d", param, &dump_reg_stat); 
	 }
         fclose (fptr);
	 if (MB_ASSOC == 0)
  	     MB_ASSOC = MEM_BUF_SIZE / MB_ENTRY_SIZE; 	/* fully mapped */
    }
  }

  num_mb_sets = mem_buf_size / MB_ENTRY_SIZE / MB_ASSOC;
  mb_set_size = MB_ENTRY_SIZE * MB_ASSOC;

  /* initialize communication unit */
  for (i = 0 ; i < ST_NUM_TU; i++) {
      comm_unit[i] = (struct CMU_ENTRY *) malloc (sizeof(struct CMU_ENTRY) * 
					cmu_size);
      cmu_out_buffers[i] =(struct CMU_ENTRY *)malloc(sizeof(struct CMU_ENTRY) * 
					cmu_size);
      for (j=0; j<cmu_size; j++) {
	  comm_unit[i][j].st_op = ENTRY_FREE;
	  cmu_out_buffers[i][j].st_op = ENTRY_FREE;
      }

	/* thread unit state : 
	   parent id, pid, parent ri/parent tsag passed, start pc, ...*/
      tu_state[i] = (int *)malloc (sizeof(int) * 8);
      mem_buf_ports[i] = (struct Port *)malloc(sizeof (struct Port) *
					st_num_mb_ports);	
      ring_ports[i] = (struct Port *)malloc(sizeof (struct Port) *
					st_num_ring_ports);	
      for (j = 0; j < st_num_mb_ports; j++)
	  mem_buf_ports[i][j].busy = 0;
      for (j = 0; j < st_num_ring_ports; j++)
	  ring_ports[i][j].busy = 0;

      tu_state[i][0] = UNIT_FREE;

      /* initialize cu_task_que */
      cu_task_que[i]=(struct CMU_ENTRY *)malloc (sizeof(struct CMU_ENTRY) 
			* cmu_size);
      for (j=0; j<cmu_size; j++)
	  cu_task_que[i][j].st_op = ENTRY_FREE;

      taskque[i] = NULL;
      stage1_start = stage2_start = stage3_start = stage4_start = 0;
      life_start = 0;
  }

  /* now create memory buffers */
  for (i = 0 ; i< ST_NUM_TU; i++) {
      memory_buf[i] = (struct MB_ENTRY *) malloc (sizeof (struct MB_ENTRY) *
						  mem_buf_size);
      /* create hash table for memory buffer lookup */
      mbuf_htable[i] = (struct MB_HTBL_ENTRY *) malloc ( 
				sizeof (struct MB_HTBL_ENTRY) * MB_HT_SIZE );
      for (j =0; j< MB_HT_SIZE; j++)
          mbuf_htable[i][j].addr = 0;
      for (j=0; j < MB_ENTRY_SIZE; j++)
	  memory_buf[i][j].word_valid = memory_buf[i][j].addr_valid = 0;
  }

  MAX_DISV = 1 << (ST_NUM_TU - 1);
  TRI = 0;

  if (dump_prof) {
	profile_fp = fopen ("simca.profile", "w");
  	fprintf (profile_fp, "***** profile starts *****\n");
/* For Vikram only 
  	fprintf (profile_fp, "Que ID 1: message from predecessor \n");
  	fprintf (profile_fp, "Que ID 2: message from forwarding MB \n");
  	fprintf (profile_fp, "Que ID 3: message from execution unit \n");
  	fprintf (profile_fp, "Que ID 4: locally executed ST-instructions\n");
  	fprintf (profile_fp, "Format: Que ID, Thread Instance No., Thread Unit ID, sim_cycle, Mesg ID, DISV, addr, data, # bytes\n");
*/
  }
  return 0;
}

/* return the index of the first available entry in the child's CMU */
int get_next_cmu_entry()
{
  int i, j;

  i = cmu_head_index[child_ri];
  for (j=0; j < cmu_size; j++) {
	if (next_cmu[i].st_op == ENTRY_FREE)
	    return i;
	i = i+1;
	if (i>= cmu_size)
	    i = 0;
  }

  return -1;
}

/* return the index of the first available entry in the local outgoing buffer */
int get_cmu_entry()
{
  int i;

  for (i=0; i < cmu_size; i++) {
	if (cmu_out_buf[i].st_op == ENTRY_FREE)
	    return i;
  }

  return -1;
}

/* clear a memory buffer entry */
int clear_entry (int i)
{

    shm_mem_buf[i].addr_valid = 0;
    shm_mem_buf[i].data_valid = 0;
    shm_mem_buf[i].alias_count = 0;
    shm_mem_buf[i].tsdva=0;
    shm_mem_buf[i].tsdvd=0;
    shm_mem_buf[i].word_valid = 0;
    shm_mem_buf[i].data = 0;

    return 0;
}

/* update the hash table with a new entry in memory buffer */
int hash_entry(unsigned int addr, int index)
{
    int i;
    struct MB_HTBL_ENTRY *phte;

    addr = (addr & (~0x3)); 
    /* assume addr is word address */
    i = (addr>>2) % MB_HT_SIZE;

	/* common case, no conflicting item */
    if (mb_htable[i].addr == 0) {
	mb_htable[i].addr = addr;
	mb_htable[i].index = index;
	mb_htable[i].next = NULL;
	return 0;
    }

    for (phte = & mb_htable[i]; phte->next != NULL; phte = phte->next); 

    phte->next = (struct MB_HTBL_ENTRY *)malloc(sizeof(struct MB_HTBL_ENTRY));
    phte = phte->next;
    phte->next = NULL;
    phte->addr = addr;
    phte->index = index;

    return 0;
}

/* free the hash table entry occupied by this address */
int free_hash_entry (unsigned int addr)
{
    
    int i;
    struct MB_HTBL_ENTRY *phte, *pe;
   
    addr = (addr & (~0x3)); 
    i = (addr >> 2) % MB_HT_SIZE;
	/* handle the most common case */
    if (mb_htable[i].addr == addr) {
	phte = mb_htable[i].next;
	if (phte == NULL) {
	    mb_htable[i].addr = 0;
	    return 0;
	} else {
	    mb_htable[i].addr = phte->addr;
	    mb_htable[i].index = phte->index;
	    mb_htable[i].next = phte->next;
	    free (phte);
	    return 0;
	}
    }

	/* rare case */
    for (pe = NULL, phte = &mb_htable[i]; phte->next != NULL; phte = phte->next)
    {
	if (phte->next->addr == addr) {
		pe = phte->next;
		phte->next = phte->next->next;
		break;
	}
    }

    if (pe)
        free(pe);

    return 0;
}

/* initialize the thread unit */
int tu_init(int i) 
{ char name[10];
  int j;

  TRI = i;
  outfile = stdout; /* fopen (name, "w"); */
  resource_index = i;
  FORK_ALLOWED = TRUE;
  signal(SIGFPE, SIG_IGN);
  pid = i+1;

  for (j=0; j<mem_buf_size; j++) {
	clear_entry(j); 
  }

  regs_init();

  TRI = 0;
  return 0;
}

/* cmu_out_buf is a FIFO */
int process_cmu_out_buf() 
{
  int i, j, p;

  if (next_thread <= 0)
	return 0; 

  for (i = 0; i< cmu_size; i++) {
	if (cmu_out_buf[i].st_op == ENTRY_FREE) 
	    return 0;

  	p = get_ring_port();
	if (p < 0)
	    break;

	j = get_next_cmu_entry();
	if (j < 0)
	    break;

	ring_port[p].busy = sim_cycle + 1; /* pipelined ring */
	next_cmu[j] = cmu_out_buf[i];
	next_cmu[j].arrival_time = sim_cycle + st_ring_delay;

	/* only until now are we done with forwarding */
	if (cmu_out_buf[i].st_op == OP_TU_GO)
	    forwarding = FALSE;

	cmu_out_buf[i].st_op = ENTRY_FREE;
   }

   /* shift the queue */
   if (i>0)
     for (j=0; j< cmu_size - i; j++) {
	cmu_out_buf[j] = cmu_out_buf[i+j];
	cmu_out_buf[i+j].st_op = ENTRY_FREE;
     }

   return 0;   
}

int clear_cmu_out_buf()
{
   int i;

   for (i = 0; i<cmu_size; i++) 
	cmu_out_buf[i].st_op = ENTRY_FREE;

}

int block_tu() 
{
   TU_BLOCKED = TRUE;
}

int tu_continue() 
{
   TU_BLOCKED = FALSE;
}

int dump_profile (int id, struct CMU_ENTRY  cmu_entry)
{

   /* Que ID 1:  any message from  predecessor 
	     2:  any message from execution unit 
	     3:  any message for forwarding memory buffer
   */

   fprintf (profile_fp, "%d, %d, %d, %d, %d, %d, %x, %d, %d\n",
	    id, pid, TRI, (int)sim_cycle, cmu_entry.st_op, cmu_entry.disv, 
	    cmu_entry.addr, cmu_entry.data, cmu_entry.nbytes);
   return 0;
}

int dump_local_profile (int op) {
   /* Que ID 4:  locally executed ST instructions that are not
	necessariliy turned into messages */  

   fprintf (profile_fp, "%d, %d, %d, %d, %d, %d, %x, %d, %d\n",
	    4, pid, TRI, (int)sim_cycle, op, 0,0,0,0);
   return 0;
}

