#ifndef EXT
#define EXT extern
#endif

#define ST_NUM_TU 8		/* change this value requires recompiling */
#define DISV_DATA_TYPE unsigned short	/* data type for distance vector */
#define ST_MAX_TU 16		/* type short is used for distance vector*/
#define MB_LATENCY 1		/* Default Memory Buffer access latency */
#define RING_LATENCY 1		/* Default latency for communication ring */
#define MB_HT_SIZE 1397		/* a random prime number for hash table */
#define NUM_MB_PORTS 2		/* default number of ports for mem buf */
#define NUM_RING_PORTS 2	/* default number of ports for comm. ring */
#define CMU_SPEED	2	/* default cmu speed */
#define TASK_Q_SIZE 64		/* blocked task que size */
#define DEF_MB_ASSOC 0		/* default memory buffer associativity: full*/
				/* 1 stands for direct mapped */
				/* > 1 stands for set associative */
#define NON_NULL (struct MB_ENTRY *)0xf

EXT int MB_ASSOC;		/* memory buffer associativity */
EXT int *tu_state[ST_NUM_TU];	/* thread-unit state variables */
EXT int num_insn_execed[ST_NUM_TU]; /* number of instructions executed since 
				       the beginning of the thread */
EXT long long total_insn_squashed;	    /* number of instructions squashed 
				       due to squashed threads */
EXT int mem_buf_size, cmu_size, num_mb_sets, mb_set_size;
				    /* memory buffer size, 
				       communication unit size,
				       number of sets in memory buffer 
				       set size of memory buffer */
				
EXT int TRI; 			/* Current Thread Unit Resource Index */
#define TU_STATE tu_state[TRI]
#define NUM_INSN_EXED num_insn_execed[TRI]

/* thread unit state vars */
EXT char fork_now[ST_NUM_TU], wait_for_tsagd[ST_NUM_TU]; 
#define HFRK_NOW fork_now[TRI]
#define WAIT_TSAG_DONE wait_for_tsagd[TRI]

EXT char tsag_passed[ST_NUM_TU], abort_now[ST_NUM_TU];
#define TSAG_PASSED tsag_passed[TRI]
#define ABORT_NOW abort_now[TRI]

EXT char need_to_wait[ST_NUM_TU], tu_blocked[ST_NUM_TU];
#define NEED_TO_WAIT need_to_wait[TRI]
#define TU_BLOCKED tu_blocked[TRI]

EXT int next_tu[ST_NUM_TU], next_tu[ST_NUM_TU]; 
#define next_thread next_tu[TRI]

EXT int ri[ST_NUM_TU], child_rind[ST_NUM_TU];
#define resource_index ri[TRI]
#define child_ri child_rind[TRI]

EXT char in_critical_region[ST_NUM_TU], fork_allowed[ST_NUM_TU];
#define In_Critical_Region in_critical_region[TRI]
#define FORK_ALLOWED fork_allowed[TRI]

EXT struct Task_Que_Entry {
    unsigned int t;	/* the due time for the scheduled task, not used now */
    char func;		/* operation code, corresponding to a function name  */
    unsigned int arg1, arg2, arg3, arg4; /* arguements for the function */
    struct Task_Que_Entry *next;
} *(taskque[ST_NUM_TU]);

#define task_que taskque[TRI]

/* ST shared state vars */
EXT unsigned int FORK_LABEL; /* the address to start next thread exec. */
EXT int fork_delay, HFRK_LATENCY;	/* Time taken to do st_fork */
EXT FILE *outfile, *profile_fp;		/* output file pointers */
EXT int st_num_mb_ports;		/* number of ports for mem buffer*/
EXT int st_num_ring_ports;		/* number of cmu ring ports */
EXT int st_ring_delay;			/* communication ring delay */
EXT int st_cmu_speed;			/* number of items processed per cycle*/
EXT DISV_DATA_TYPE MAX_DISV; 		/* maximal distant vector */ 
EXT int mb_set_conflict[ST_NUM_TU];	/* MB set conflicts */
EXT int dump_prof;			/* flag for dumping profile info */
EXT int dump_reg_stat;			/* flag for dumping reg update info */

EXT unsigned int data_dep_delay[ST_NUM_TU];/* delay caused by data dependence */
EXT unsigned int wait_to_wb_delay[ST_NUM_TU]; /* delay by serializatoin of wb*/
EXT long long thread_start[ST_NUM_TU];	/* the time a thread starts */
EXT unsigned int wait_lock_delay[ST_NUM_TU];/* delay by unsuccessful lock */
EXT int reg_masks[ST_NUM_TU][3];	/* register masks for reg writes */

#define set_conflict mb_set_conflict[TRI]
#define data_delay data_dep_delay[TRI]
#define wb_delay wait_to_wb_delay[TRI]
#define life_start thread_start[TRI]
#define lock_delay wait_lock_delay[TRI]

EXT int forwarding_mb_entry[ST_NUM_TU]; /* is the TU forwarding MB entries?  */
EXT long long forward_delay;			/* time to forward MB entries */
#define forwarding forwarding_mb_entry[TRI]

/* stage accounting, these are not accurate */
EXT long long cont_stage_start[ST_NUM_TU], tsag_stage_start[ST_NUM_TU]; 
EXT long long exec_stage_start[ST_NUM_TU], wb_stage_start[ST_NUM_TU];
EXT long long tot_stage1, tot_stage2, tot_stage3, tot_stage4;

/* the total execution time for the program section that we parallelized */
EXT long long accounted_time;
EXT long long region_begin_time;

#define stage1_start cont_stage_start[TRI]
#define stage2_start tsag_stage_start[TRI]
#define stage3_start exec_stage_start[TRI]
#define stage4_start wb_stage_start[TRI]

/* thread management */
EXT int st_init();
EXT int tu_init(int i);
EXT int thread_init();
EXT int writeback_stage (int);
EXT int st_fork_next();
EXT int kill_successor();

/* CMU utilities */
EXT int check_cmu();
EXT int process_st_msg_q();
EXT int get_next_cmu_entry();
EXT int get_cmu_entry();


/* ST instructions */

EXT int ABORT_FUTURE();
 	/* ST_END_REGION corresponds to the "estr" instruction, which is
   	   the combination of writeback_stage, and ST_DONE */
EXT int ST_DONE();		
EXT int STORE_TS (unsigned int a,unsigned int d,int nb, DISV_DATA_TYPE disv, 
	char);
EXT int ALLOCATE_TS (unsigned int addr, DISV_DATA_TYPE disv, int nbytes, char);
EXT int RELEASE_TS (unsigned int addr, DISV_DATA_TYPE disv, int nbytes, char);
EXT int hfrk(unsigned int label);	/* similar to ST_FORK */
EXT int get_lock(int);
EXT int release_lock(int);
EXT int dbg_info (int);			/* added debugging instruction */
EXT int wait_tsag_done();

/* memory buffer access */
EXT int in_mb (unsigned int addr);	/* check the existence of an entry */
EXT unsigned char __MB_READ_BYTE ( unsigned int addr );
EXT unsigned short __MB_READ_HALF ( unsigned int addr );
EXT unsigned int __MB_READ_WORD ( unsigned int addr );
EXT int MB_WRITE (unsigned int addr , unsigned int data, int n);
EXT int forward_mbuf(int);
EXT struct MB_ENTRY *get_mb_entry(unsigned int addr);
EXT int clear_entry(int i);
EXT int hash_entry(unsigned int addr, int index);
EXT int free_hash_entry (unsigned int addr);

/* taske que handling */
EXT int task_queue_init();
EXT struct Task_Que_Entry * new_tq_entry();
EXT int free_tq_entry (struct Task_Que_Entry *tqe);
EXT int task_enque(char func,unsigned int arg1,unsigned int arg2,
                unsigned int arg3, unsigned int arg4, unsigned int t);
EXT int process_task_que();


#define __MB_WRITE_BYTE(X, Y)  MB_WRITE ((X), (int)(Y), 1)
#define __MB_WRITE_HALF(X, Y)  MB_WRITE ((X), (int)(Y), 2)
#define __MB_WRITE_WORD(X, Y)  MB_WRITE ((X), (Y), 4)

EXT struct Port {
    int port_id;
    long long busy;
}*mem_buf_ports[ST_NUM_TU], *ring_ports[ST_NUM_TU];

#define mb_port mem_buf_ports[TRI]
#define ring_port ring_ports[TRI]

EXT struct CMU_ENTRY {
    short nbytes;
    DISV_DATA_TYPE disv;
    char st_op;
    unsigned int addr;
    unsigned int data;
    long long arrival_time;
} *comm_unit[ST_NUM_TU], *next_cu[ST_NUM_TU], *cu_task_que[ST_NUM_TU], 
  *cmu_out_buffers[ST_NUM_TU];

#define shm_comm_unit comm_unit[TRI]
#define next_cmu next_cu[TRI]
#define st_msg_q cu_task_que[TRI]
#define cmu_out_buf cmu_out_buffers[TRI]

EXT struct MB_ENTRY {
    char word_valid;
    char addr_valid;
    char data_valid;
    unsigned char alias_count;
    DISV_DATA_TYPE tsdva;
    DISV_DATA_TYPE tsdvd;
    unsigned int tag;
    unsigned char data;
    /* unsigned int data;  * Here we use byte addressing, which means we 
			   can treat each entry as a single unit, but
			   this also means wasting some storage, if the 
			   programs mostly access by shorts or bytes. 
			   In addition, this technique can not deal with
			   pointer conversion case, so we did not choose it */
    char access_mode;	/* 4 for word, 2 for half_word, 1 for byte */
} *memory_buf[ST_NUM_TU];

#define shm_mem_buf memory_buf[TRI]

EXT struct MB_HTBL_ENTRY {
    unsigned int addr; 	/* word index of entry */
    int index;		/* the index to access memory buffer */
    struct MB_HTBL_ENTRY *next;
} *mbuf_htable[ST_NUM_TU];

#define mb_htable mbuf_htable[TRI]

/* implement this message queue in round-robin style */
EXT int cu_task_que_head[ST_NUM_TU], cu_task_que_tail[ST_NUM_TU];
EXT int cmu_head_index[ST_NUM_TU]; 
#define cmu_head cmu_head_index[TRI]
#define st_msg_q_head cu_task_que_head[TRI]
#define st_msg_q_tail cu_task_que_tail[TRI]

EXT int system_lock[32];  

/* global state and thread accounting */
EXT int num_active_threads, head_thread_id, tu_instance; 
EXT int num_squashed, mb_entry_used, max_mb_entry_used;
EXT long long total_set_conflict;
EXT long long total_data_delay, total_wb_delay, total_life;
EXT long long total_lock_delay;

/* statistics for register definig actions */
EXT int total_par_region, total_seq_region, max_seq_reg_def, min_seq_reg_def;
EXT int max_par_reg_def, min_par_reg_def, total_seq_reg_def, total_par_reg_def;

/* parent thread's ID, -1 if no parent, also means unit idle */
#define parent_id		tu_state[TRI][0]
/* current thread's thread ID */
#define pid			tu_state[TRI][1]
/* parent thread resource index */
#define parent_rid		(tu_state[TRI][2] >> 16)
/* indicate whether parent thread has passed TSAG stage */
#define p_tsag_passed		(tu_state[TRI][2] & 0xFF)
/* the address to start execution */
#define start_pc		tu_state[TRI][3]

#define UNIT_FREE       -1
#define ENTRY_FREE      0

#define MAX_CMU_ENTRY   8	/* number of entries to hold input message */
				/* another 8 entries is used to hold output
				   messages */	

#define MB_ADDR_BIT(X)  (X.addr_valid )
#define MB_DATA_BITS(X)	(X.data_valid )
#define MB_TSDVA(X)     (X.tsdva)
#define MB_TSDVD(X)     (X.tsdvd)
#define MB_TAG(X)	(X.tag)
#define MB_BOFS(X)	(X.byte_ofs)
#define MB_CONT(X)	(X.data)
#define MB_ALIAS_COUNT(X) 	(X.alias_count)
#define MB_WORD_VALID(X) 	(X.word_valid)

#define MEM_BUF_SIZE	512 		/* 1024 Bytes, 128 entries  */
#define MB_ENTRY_SIZE	4 		/* one word */

#define OP_ALLOCATE_TS  0x29
#define OP_STORE_TS     0x2a
#define OP_RELEASE_TS   0x2b
#define OP_TSAG_DONE	0x2c
#define OP_ABORT	0x2d
#define OP_TU_GO	0x2e
#define OP_WAIT_TSAGD	0x2f
#define OP_FORK		0x30
#define OP_STOP		0x31
#define OP_BEGIN_TR	0x32

#define DATA_VALID	4
#define ADDR_VALID	2
#define DONT_WAIT	3

#define ONE_BYTE	0x1
#define HALF_WORD	0x2
#define FULL_WORD	0x4

#define WRITE_BACK	5
#define GET_LOCK	6
#define CHECK_HEAD	7
#define MBUF_WRITE	8
#define CHECK_WTSAGD	9
#define CHECK_MB	10
#define FWD_MB		11
#define TR_DONE 	12
#define FWD_ALTS	13
#define FWD_RLTS	14
#define FWD_STTS	15
#define FWD_TSAGD	16

#define FWD_ONLY	1

#define REG_HI		65
#define REG_LO		66
#define REG_FCC		67

#define SET_DREG_MASK(X)	reg_masks[TRI][(X)/32] |= (1 << ((X) & 31))
void par_reg_stat();
void seq_reg_stat();
void reset_reg_masks();

